/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <dirent.h>

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#include "pwmd-error.h"
#include "util-misc.h"
#include "mem.h"
#include "common.h"
#include "commands.h"
#include "agent.h"
#include "util-string.h"
#include "mutex.h"
#include "rcfile.h"
#include "cache.h"
#include "cipher.h"
#include "crypto.h"

static gpg_error_t
mem_realloc_cb (void *data, const void *buffer, size_t len)
{
  membuf_t *mem = (membuf_t *) data;
  void *p;

  if (!buffer)
    return 0;

  if ((p = xrealloc (mem->buf, mem->len + len)) == NULL)
    return 1;

  mem->buf = p;
  memcpy ((char *) mem->buf + mem->len, buffer, len);
  mem->len += len;
  return 0;
}

/* Forward status messages from gpg-agent to the client. */
static gpg_error_t
status_cb (void *data, const char *line)
{
  struct agent_s *agent = data;

  agent->inquire_maxlen = 0;

  if (!strncmp (line, "INQUIRE_MAXLEN ", 15))
    {
      agent->inquire_maxlen = atoi (line + 15);
      if (!agent->client_ctx)
	return 0;
    }

  return send_status (agent->client_ctx, STATUS_AGENT, "%s", line);
}

static gpg_error_t
assuan_command (struct agent_s *a, char **result,
		size_t * len, const char *cmd)
{
  gpg_error_t rc;

  a->data.len = 0;
  a->data.buf = NULL;
  if (result)
    *result = NULL;
  if (len)
    *len = 0;

  rc = assuan_transact (a->ctx, cmd, mem_realloc_cb, &a->data,
			a->inquire_cb, a->inquire_data, status_cb, a);
  if (rc)
    xfree (a->data.buf);
  else
    {
      if (a->data.buf)
	{
	  mem_realloc_cb (&a->data, "", 1);
	  if (result)
	    *result = (char *) a->data.buf;
	  else
	    xfree (a->data.buf);

	  if (len)
	    *len = a->data.len;
	}
    }

  return rc;
}

/* This commands are sent from connect_agent() after reconnecting to the agent
 * and also from the initial client connection. */
gpg_error_t
send_agent_common_options (struct agent_s *agent)
{
  gpg_error_t rc;

  rc = send_to_agent (agent, NULL, NULL, "OPTION cache-ttl-opt-preset=-1");
  return rc;
}

static gpg_error_t
connect_agent (struct agent_s *agent)
{
  gpg_error_t rc;
  assuan_context_t ctx = NULL;
  static struct assuan_malloc_hooks mhooks = { xmalloc, xrealloc, xfree };
  char *s;

  agent->did_restart = 0;
  rc = assuan_new_ext (&ctx, GPG_ERR_SOURCE_DEFAULT, &mhooks,
		       debug_level ? assuan_log_cb : NULL, "AGENT ");
  if (rc)
    return rc;

  s = config_get_string ("global", "gpg_agent_socket");
  rc = assuan_socket_connect (ctx, s, ASSUAN_INVALID_PID, 0);
  xfree (s);
  if (rc)
    {
      if (ctx)
	assuan_release (ctx);
      return rc;
    }

  if (agent->ctx)
    assuan_release (agent->ctx);

  agent->ctx = ctx;
  rc = send_agent_common_options (agent);
  if (!rc && agent->pinentry_opts.display)
    rc = set_agent_option (agent, "display", agent->pinentry_opts.display);

  if (!rc && agent->pinentry_opts.ttyname)
    rc = set_agent_option (agent, "ttyname", agent->pinentry_opts.ttyname);

  if (!rc && agent->pinentry_opts.ttytype)
    rc = set_agent_option (agent, "ttytype", agent->pinentry_opts.ttytype);

  if (!rc && agent->pinentry_opts.lc_messages)
    rc = set_agent_option (agent, "lc-messages",
			   agent->pinentry_opts.lc_messages);

  if (!rc && agent->pinentry_opts.lc_ctype)
    rc = set_agent_option (agent, "lc-ctype", agent->pinentry_opts.lc_ctype);

  return rc;
}

gpg_error_t
send_to_agent (struct agent_s * agent, char **result, size_t * len,
	       const char *fmt, ...)
{
  gpg_error_t rc = 0;
  va_list ap;
  char *cmd;

  if (!agent->ctx)
    {
      rc = connect_agent (agent);
      if (rc)
	return rc;
    }

  va_start (ap, fmt);
  if (str_vasprintf (&cmd, fmt, ap) > 0)
    {
      rc = assuan_command (agent, result, len, cmd);
      if (!agent->restart && gpg_err_source (rc) == GPG_ERR_SOURCE_DEFAULT
	  && (gpg_err_code (rc) == GPG_ERR_ASS_CONNECT_FAILED
	      || gpg_err_code (rc) == GPG_ERR_EPIPE))
	{
	  log_write (_ ("gpg-agent connection died (rc=%i), reconnecting"), rc);
	  agent->restart = 1;
	  rc = connect_agent (agent);
	  if (!rc)
	    {
	      agent->did_restart = 1;
	      rc = assuan_command (agent, result, len, cmd);
	    }
	}

      agent->restart = 0;
      xfree (cmd);
    }
  else
    rc = GPG_ERR_ENOMEM;

  va_end (ap);
  return rc;
}

static void
agent_disconnect (struct agent_s *agent)
{
  if (!agent)
    return;

  if (agent->ctx)
    assuan_release (agent->ctx);

  agent->ctx = NULL;
}

void
cleanup_agent (struct agent_s *agent)
{
  if (!agent)
    return;

  pinentry_free_opts (&agent->pinentry_opts);
  agent_disconnect (agent);
  xfree (agent);
}

gpg_error_t
agent_init (struct agent_s **agent)
{
  struct agent_s *new = xcalloc (1, sizeof (struct agent_s));

  if (!new)
    return GPG_ERR_ENOMEM;

  *agent = new;
  return 0;
}

static gpg_error_t
inquire_cb (void *user, const char *keyword)
{
  struct inquire_data_s *idata = user;

  if (!idata->preset && (!strcmp (keyword, "PASSPHRASE")
			 || !strcmp (keyword, "NEW_PASSPHRASE")))
    {
      return agent_loopback_cb (idata->crypto, keyword);
    }
  // SAVE --inquire-keyparam
  else if (idata->preset && !strcmp (keyword, "KEYPARAM"))
    {
      idata->preset = 0;
      return agent_loopback_cb (idata->crypto, keyword);
    }

  if (idata->crypto->agent->inquire_maxlen
      && idata->len > idata->crypto->agent->inquire_maxlen)
    {
      log_write (_("Inquired data too large: have=%u, max=%u"), idata->len,
		 idata->crypto->agent->inquire_maxlen);
      return GPG_ERR_TOO_LARGE;
    }

  idata->crypto->agent->inquire_maxlen = 0;
  return assuan_send_data (idata->crypto->agent->ctx, idata->line,
			   idata->len);
}

/* Retrieve the passphrase needed to decrypt the XML from the secret
 * key using gpg-agent. */
gpg_error_t
agent_extract_key (struct crypto_s *crypto, unsigned char **result,
		   size_t * result_len)
{
  gpg_error_t rc;
  gcry_sexp_t enc_sexp = NULL, tmp_sexp;
  struct inquire_data_s idata = { 0 };
  char *hexgrip = NULL;
  char *key;
  size_t keylen, keysize;
  char *tmp;
  int algo = cipher_to_gcrypt (crypto->hdr.flags);
  int shadowed;

  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL, &keysize);
  if (rc)
    return rc;

  tmp_sexp = gcry_sexp_find_token (crypto->ciphertext_sexp, "enc-val", 0);
  if (!tmp_sexp)
    return gpg_error (GPG_ERR_BAD_DATA);

  hexgrip = bin2hex (crypto->grip, sizeof (crypto->grip));
  if (!hexgrip)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  rc = cache_is_shadowed (hexgrip);
  if (rc && rc != GPG_ERR_NO_DATA)
    goto fail;

  shadowed = !rc ? 1 : 0;
  if (!shadowed)
    {
      /* The key is not on a smartcard. */
      gcry_sexp_t tmp2_sexp = gcry_sexp_cdr (tmp_sexp);
      gcry_sexp_release (tmp_sexp);
      tmp_sexp = gcry_sexp_nth (tmp2_sexp, 0);
      gcry_sexp_release (tmp2_sexp);
      rc = gcry_sexp_build (&enc_sexp, NULL, "(enc-val (flags pkcs1) %S)",
			    tmp_sexp);
    }
  else
    rc = gcry_sexp_build (&enc_sexp, NULL, "%S", tmp_sexp);

  gcry_sexp_release (tmp_sexp);
  if (rc)
    return rc;

  crypto->agent->inquire_cb = inquire_cb;
  idata.crypto = crypto;
  idata.len = gcry_sexp_sprint (enc_sexp, GCRYSEXP_FMT_CANON, NULL, 0);
  idata.line = xmalloc (idata.len);
  if (!idata.line)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  idata.len = gcry_sexp_sprint (enc_sexp, GCRYSEXP_FMT_CANON, idata.line,
				idata.len);
  crypto->agent->inquire_data = &idata;
  gcry_sexp_release (enc_sexp);
  enc_sexp = NULL;
  log_write1 (_("Keygrip is %s, bits=%i"), hexgrip,
	      gcry_pk_get_nbits (crypto->pkey_sexp));
  rc = send_to_agent (crypto->agent, NULL, NULL, "SETKEY %s", hexgrip);
  if (rc)
    goto fail;

  if (!crypto->agent->pinentry_opts.desc)
    {
      tmp =
	plus_escape (_
		     ("A %s is required to unlock the secret key for the "
		      "encrypted data file \"%s\". Please enter the %s "
		      "below."), shadowed ? "PIN" : _("passphrase"),
		     crypto->filename, shadowed ? "PIN" : _("passphrase"));
    }
  else
    tmp = plus_escape (crypto->agent->pinentry_opts.desc);

  rc = send_to_agent (crypto->agent, NULL, NULL, "SETKEYDESC %s", tmp);
  xfree (tmp);
  if (rc)
    goto fail;

  assuan_begin_confidential (crypto->agent->ctx);
  rc = send_to_agent (crypto->agent, &key, &keylen, "PKDECRYPT");
  assuan_end_confidential (crypto->agent->ctx);
  if (rc)
    goto fail;

  rc = gcry_sexp_new (&tmp_sexp, key, keylen, 1);
  xfree (key);
  if (rc)
    goto fail;

  key = (char *) gcry_sexp_nth_data (tmp_sexp, 1, result_len);
  if (key)
    {
      *result = gcry_malloc (*result_len);
      if (!*result)
	rc = GPG_ERR_ENOMEM;
      else
	memcpy (*result, key, *result_len);
    }
  else
    rc = GPG_ERR_BAD_DATA;

  gcry_sexp_release (tmp_sexp);

fail:
  xfree (idata.line);
  xfree (hexgrip);

  if (enc_sexp)
    gcry_sexp_release (enc_sexp);

  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

// FIXME
/* Return an error for public key algorithms other than RSA or DSA or ELG.
 * Others are supported by libgcrypt but I dont know what to do for ECDA. Key
 * generation works as does signing but decryption fails for unsupported
 * algorithms so disable them. */
static gpg_error_t
get_key_info (gcry_sexp_t pubkey, char **hash_str, int *hash_type,
	       size_t *len, int *pk_algo)
{
  unsigned nbits = gcry_pk_get_nbits (pubkey);
  gcry_sexp_t sexp, tmp;
  char *algo;
  gpg_error_t rc = 0;

  tmp = gcry_sexp_find_token (pubkey, "public-key", 0);
  if(!tmp)
    return GPG_ERR_NO_PUBKEY;

  sexp = gcry_sexp_nth (tmp, 1);
  gcry_sexp_release (tmp);
  algo = gcry_sexp_nth_string (sexp, 0);
  gcry_sexp_release (sexp);

  if (!algo)
    {
      return GPG_ERR_BAD_PUBKEY;
    }

  if (!strcmp (algo, "rsa"))
    {
      *pk_algo = GCRY_PK_RSA;
      *hash_str = "sha256";
      *hash_type = GCRY_MD_SHA256;
      *len = gcry_md_get_algo_dlen (GCRY_MD_SHA256);
    }
  else if (!strcmp (algo, "elg"))
    {
      *pk_algo = GCRY_PK_ELG;
      *hash_str = "sha256";
      *hash_type = GCRY_MD_SHA256;
      *len = gcry_md_get_algo_dlen (GCRY_MD_SHA256);
    }
  else if (!strcmp (algo, "dsa"))
    {
      *pk_algo = GCRY_PK_DSA;
      switch (nbits)
	{
	default:
	case 512 ... 1024:
	  *hash_str = "sha1";
	  *hash_type = GCRY_MD_SHA1;
	  *len = gcry_md_get_algo_dlen (GCRY_MD_SHA1);
	  break;
	case 2048:
	  *hash_str = "sha224";
	  *hash_type = GCRY_MD_SHA224;
	  *len = gcry_md_get_algo_dlen (GCRY_MD_SHA224);
	  break;
	case 3072:
	  *hash_str = "sha256";
	  *hash_type = GCRY_MD_SHA256;
	  *len = gcry_md_get_algo_dlen (GCRY_MD_SHA256);
	  break;
	case 7680:
	  *hash_str = "sha384";
	  *hash_type = GCRY_MD_SHA384;
	  *len = gcry_md_get_algo_dlen (GCRY_MD_SHA384);
	  break;
        case 15360:
	  *hash_str = "sha512";
	  *hash_type = GCRY_MD_SHA512;
	  *len = gcry_md_get_algo_dlen (GCRY_MD_SHA512);
	  break;
	}
    }
  else
    rc = GPG_ERR_UNSUPPORTED_ALGORITHM;

  gcry_free (algo);
  return rc;
}

gpg_error_t
agent_verify (gcry_sexp_t pkey, gcry_sexp_t sig_sexp, const void *data,
	      size_t len)
{
  size_t hashlen;
  int pk_algo;
  unsigned char *hash;
  gcry_sexp_t data_sexp;
  char *hash_str = NULL;
  int hash_type;
  gpg_error_t rc = get_key_info (pkey, &hash_str, &hash_type, &hashlen,
				 &pk_algo);

  if (rc)
    return rc;

  hash = gcry_malloc (hashlen);
  if (!hash)
    return GPG_ERR_ENOMEM;

  gcry_md_hash_buffer (hash_type, hash, data, len);
  if (pk_algo == GCRY_PK_RSA || pk_algo == GCRY_PK_ELG)
    rc = gcry_sexp_build (&data_sexp, NULL,
			  "(data (flags pkcs1) (hash %s %b))", hash_str,
			  hashlen, hash);
  else
    rc = gcry_sexp_build (&data_sexp, NULL,
			  "(data (flags raw) (value %b))", hashlen, hash);

  gcry_free (hash);
  if (!rc)
    {
      rc = gcry_pk_verify (sig_sexp, data_sexp, pkey);
      gcry_sexp_release (data_sexp);
    }

  return rc;
}

/* Inquire data from the client. The inquire initiated from gpg-agent. */
gpg_error_t
agent_loopback_cb (void *user, const char *keyword)
{
  struct crypto_s *crypto = user;
  gpg_error_t rc;
  unsigned char *result;
  size_t len;
  int keyparam = 0;

  if (!strcmp (keyword, "KEYPARAM"))
    {
      keyparam = 1;
      rc = assuan_inquire (crypto->client_ctx, keyword, &result, &len, 0);
    }
  else
    {				// PASSPHRASE or NEW_PASSPHRASE
      assuan_begin_confidential (crypto->client_ctx);
      rc = assuan_inquire (crypto->client_ctx, keyword, &result, &len, 0);
      assuan_end_confidential (crypto->client_ctx);

      if (!rc && len == 1 && *result == 0)
	len = 0;
    }

  if (!rc)
    {
      if (keyparam && !len)
	{
	  char *tmp = default_key_params (crypto);

	  if (!tmp)
	    return gpg_error (GPG_ERR_ENOMEM);

	  len = strlen (tmp);
	  xfree (result);
	  result = xmalloc (len);
	  memcpy (result, tmp, len);
	  xfree (tmp);
	}

      pthread_cleanup_push (xfree, result);

      if (keyparam)
	rc = assuan_send_data (crypto->agent->ctx, result, len);
      else
	{
	  assuan_begin_confidential (crypto->agent->ctx);
	  rc = assuan_send_data (crypto->agent->ctx, result, len);
	  assuan_end_confidential (crypto->agent->ctx);
	}

      pthread_cleanup_pop (1);
    }
  else if (gpg_err_code (rc) == GPG_ERR_ASS_CANCELED)
    {
      gpg_error_t arc = assuan_write_line (crypto->agent->ctx, "CAN");

      if (!arc)
	{
	  char *line;
	  size_t len;

	  arc = assuan_read_line (crypto->agent->ctx, &line, &len);
	  if (arc)
	    rc = arc;
	}
      else
	rc = arc;
    }

  return rc;
}

static gpg_error_t
sign (gcry_sexp_t * rsexp, gcry_sexp_t sigpkey, struct crypto_s *crypto,
      const void *data, size_t len)
{
  gpg_error_t rc;
  char *result;
  unsigned char grip[20];
  char *tmp;

  gcry_pk_get_keygrip (sigpkey, grip);
  tmp = bin2hex (grip, sizeof(grip));
  pthread_cleanup_push (xfree, tmp);
  log_write1 (_("Signed with keygrip %s"), tmp);
  rc = send_to_agent (crypto->agent, NULL, NULL, "SIGKEY %s", tmp);
  pthread_cleanup_pop (1);
  if (!rc)
    {
      size_t hashlen;
      char *hash_str = NULL;
      int hash_type;
      int pk_algo;

      rc = get_key_info (sigpkey, &hash_str, &hash_type, &hashlen, &pk_algo);
      if (!rc)
	{
          unsigned char *hash = gcry_malloc (hashlen);

	  if (!hash)
	    rc = gpg_error (GPG_ERR_ENOMEM);

	  if (!rc)
	    {
	      gcry_md_hash_buffer (hash_type, hash, data, len);
	      tmp = bin2hex (hash, hashlen);
	      gcry_free (hash);
	      pthread_cleanup_push (xfree, tmp);
	      rc = send_to_agent (crypto->agent, NULL, NULL,
				  "SETHASH --hash=%s %s", hash_str, tmp);
	      pthread_cleanup_pop (1);
	    }
	}
    }

  if (rc)
    return rc;

  if (!rc)
    {
      struct inquire_data_s idata = { 0 };

      idata.crypto = crypto;
      crypto->agent->inquire_data = &idata;
      crypto->agent->inquire_cb = inquire_cb;
      rc = send_to_agent (crypto->agent, &result, &len, "PKSIGN");
      if (!rc)
	{
	  rc = gcry_sexp_sscan (rsexp, NULL, result, len);
	  xfree (result);
	}
    }

  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

/* Encrypt the data file to 'pubkey' and sign with 'sigpkey'. */
gpg_error_t
encrypt_data_file (assuan_context_t ctx, struct crypto_s * crypto,
		   gcry_sexp_t pubkey, gcry_sexp_t sigpkey,
		   const char *filename, const void *xml, size_t len)
{
  gpg_error_t rc;
  void *data = NULL;
  size_t data_len = 0;
  void *enc_xml = NULL;
  size_t enc_xml_len = 0;
  unsigned char *iv = NULL;
  size_t iv_len = 0;
  int algo = cipher_to_gcrypt (crypto->save.hdr.flags);
  void *key = NULL;
  size_t keysize;
  unsigned char sig_grip[20];
  unsigned char grip[20];

  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL, &keysize);
  if (rc)
    return rc;

  pthread_cleanup_push (gcry_free, key);
  key = gcry_random_bytes_secure (keysize, GCRY_STRONG_RANDOM);
#ifdef HAVE_PTHREAD_TESTCANCEL
  pthread_testcancel ();	// may have been a long operation
#endif
  pthread_cleanup_pop (0);
  if (!key)
    return gpg_error (GPG_ERR_ENOMEM);

  gcry_pk_get_keygrip (pubkey, grip);
  gcry_pk_get_keygrip (sigpkey, sig_grip);
  pthread_cleanup_push (gcry_free, key);
  rc = encrypt_xml (ctx, key, keysize, algo, xml, len, &enc_xml, &enc_xml_len,
		    &iv, &iv_len);
  if (!rc)
    {
      gcry_sexp_t sig_sexp = NULL;
      char *hexgrip = bin2hex (grip, 20);

      pthread_cleanup_push (xfree, iv);
      log_write1 (_("Encrypted with keygrip %s"), hexgrip);
      xfree (hexgrip);
      hexgrip = bin2hex (sig_grip, 20);
      pthread_cleanup_push (gcry_free, enc_xml);
      rc = sign (&sig_sexp, sigpkey, crypto, enc_xml, enc_xml_len);
      xfree (hexgrip);

      if (!rc)
	{
	  rc = agent_verify (sigpkey, sig_sexp, enc_xml, enc_xml_len);
	  if (!rc)
	    {
	      gcry_sexp_t tmp_sexp;

	      rc = gcry_sexp_build (&tmp_sexp, NULL,
				    "(data (flags pkcs1) (value %b))",
				    keysize, key);
	      if (!rc)
		{
		  gcry_sexp_t key_sexp;

		  pthread_cleanup_push ((void (*)(void *)) gcry_sexp_release,
					(void *) sig_sexp);
		  rc = gcry_pk_encrypt (&key_sexp, tmp_sexp, pubkey);
		  pthread_cleanup_pop (0);
		  gcry_sexp_release (tmp_sexp);

		  if (!rc)
		    {
		      memcpy (crypto->save.hdr.iv, iv, iv_len);
		      crypto->save.hdr.datalen = enc_xml_len;
		      rc = gcry_sexp_build (&tmp_sexp, NULL, "%S%S", key_sexp,
					    sig_sexp);
		      gcry_sexp_release (key_sexp);

		      if (!rc)
			{
			  data_len = gcry_sexp_sprint (tmp_sexp,
						       GCRYSEXP_FMT_CANON,
						       NULL, 0);
			  data = xmalloc (data_len);
			  if (data)
			    gcry_sexp_sprint (tmp_sexp, GCRYSEXP_FMT_CANON,
					      data, data_len);
			  else
			    rc = GPG_ERR_ENOMEM;

			  gcry_sexp_release (tmp_sexp);
			}
		    }
		}
	    }
	}

      pthread_cleanup_pop (0);	// enc_xml
      pthread_cleanup_pop (1);	// iv

      if (sig_sexp)
	gcry_sexp_release (sig_sexp);
    }

  pthread_cleanup_pop (1);	// key

  if (!rc)
    {
      pthread_cleanup_push (gcry_free, enc_xml);
      rc = write_file (crypto, filename, enc_xml, enc_xml_len, data, data_len,
		       pubkey, sigpkey);
      pthread_cleanup_pop (1);	// enc_xml
      if (!rc)
	memcpy (&crypto->hdr, &crypto->save.hdr, sizeof (file_header_t));
    }

  xfree (data);
  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

/* Generated a new keypair from the key parameters in 'sexpstr'. */
gpg_error_t
generate_key (struct crypto_s * crypto, char *sexpstr, int empty, int preset)
{
  gpg_error_t rc;
  char *pkey;
  size_t plen;

  if (crypto->save.hdr.s2k_count)
    {
      rc = send_to_agent (crypto->agent, NULL, NULL,
			  "OPTION s2k-count=%llu", crypto->save.hdr.s2k_count);
      if (rc)
	return rc;
    }

  if (!crypto->agent->inquire_cb)
    crypto->agent->inquire_cb = inquire_cb;

  rc = send_to_agent (crypto->agent, &pkey, &plen, "GENKEY %s%s",
		      preset ? "--preset " : "",
		      empty ? "--no-protection" : "");
  if (rc)
    return rc;

  if (crypto->save.pkey)
    gcry_sexp_release (crypto->save.pkey);

  crypto->save.pkey = NULL;
  rc = gcry_sexp_new (&crypto->save.pkey, pkey, plen, 1);
  if (!rc)
    {
      unsigned char grip[20];

      gcry_pk_get_keygrip (crypto->save.pkey, grip);
      char *hexgrip = bin2hex (grip, sizeof (grip));
      log_write (_("Keygrip is %s"), hexgrip);
      xfree (hexgrip);

      if (!crypto->save.sigpkey)
	{
	  gcry_sexp_build ((gcry_sexp_t *) & crypto->save.sigpkey, NULL, "%S",
			   crypto->save.pkey);
	}
    }

  xfree (pkey);
  return rc;
}

gpg_error_t
set_agent_option (struct agent_s * agent, const char *name, const char *value)
{
  return send_to_agent (agent, NULL, NULL, "OPTION %s=%s", name, value);
}

gpg_error_t
set_agent_passphrase (struct crypto_s * crypto, const char *key, size_t len)
{
  char *hexgrip;
  struct inquire_data_s idata;
  gpg_error_t rc;
  int i;

  /* This is for use with key files or passphrases obtained from an inquire.
   * gpg-agent uses strings as passphrases and will truncate the passphrase
   * at the first encountered null byte.  It's only a warning because the
   * passphrase may belong to a key shared outside of pwmd. */
  for (i = 0; i < len; i++)
    {
      if (key[i] == 0)
	{
	  log_write (_("WARNING: keylen=%i, truncated to %i."), len, i);
	  break;
	}
    }

  hexgrip = bin2hex (crypto->grip, 20);
  crypto->agent->inquire_cb = inquire_cb;
  crypto->agent->inquire_data = &idata;
  idata.crypto = crypto;
  idata.line = (char *) key, idata.len = len;
  idata.preset = 1;
  assuan_begin_confidential (crypto->agent->ctx);
  rc = send_to_agent (crypto->agent, NULL, NULL,
		      "PRESET_PASSPHRASE --inquire %s -1", hexgrip);
  assuan_end_confidential (crypto->agent->ctx);
  idata.preset = 0;
  xfree (hexgrip);
  return rc;
}

gpg_error_t
set_pinentry_mode (struct agent_s * agent, const char *mode)
{
  return set_agent_option (agent, "pinentry-mode", mode);
}

gpg_error_t
get_pubkey_bin (struct crypto_s * crypto, const unsigned char *grip,
		gcry_sexp_t * result)
{
  char *hexgrip = bin2hex (grip, 20);
  gpg_error_t rc;

  if (!hexgrip)
    return GPG_ERR_ENOMEM;

  rc = get_pubkey (crypto, hexgrip, result);
  xfree (hexgrip);
  return rc;
}

gpg_error_t
get_pubkey (struct crypto_s * crypto, const char *grip, gcry_sexp_t * result)
{
  char *pkey = NULL;
  size_t plen;
  gpg_error_t rc;

  rc = send_to_agent (crypto->agent, &pkey, &plen, "READKEY %s", grip);
  if (!rc)
    rc = gcry_sexp_new (result, pkey, plen, 1);

  xfree (pkey);
  return rc;
}

gpg_error_t
agent_set_pinentry_options (struct agent_s * agent)
{
  gpg_error_t rc = 0;
  struct pinentry_option_s *opts = &agent->pinentry_opts;

  if (!opts->display && getenv ("DISPLAY"))
    {
      rc = set_agent_option (agent, "display", getenv ("DISPLAY"));
      if (!rc)
	{
	  opts->display = str_dup (getenv ("DISPLAY"));
	}
    }
  else if (!opts->ttyname && ttyname (STDOUT_FILENO))
    {
      rc = set_agent_option (agent, "ttyname", ttyname (STDOUT_FILENO));
      if (!rc)
	{
	  rc = set_agent_option (agent, "ttytype",
				 opts->ttytype ? opts->ttytype : getenv ("TERM"));
	  if (!rc)
	    {
	      xfree (opts->ttyname);
	      xfree (opts->ttytype);
	      opts->ttyname = str_dup (ttyname (STDOUT_FILENO));
	      opts->ttytype = str_dup (getenv ("TERM"));
	    }
	}
    }

  return rc;
}

static gpg_error_t
inquire_keyfile (void *user, const char *keyword)
{
  struct crypto_s *crypto = user;
  char *filename = crypto->agent->inquire_data2;
  char *params = crypto->agent->inquire_data3;
  int fd;
  struct stat st;
  unsigned char *buf;
  gpg_error_t rc;

  if (!strcmp (keyword, "KEYPARAM"))
    return assuan_send_data (crypto->agent->ctx, params, strlen (params));

  // This function is only used when generating a new keypair.
  if (strcmp (keyword, "NEW_PASSPHRASE"))
    return gpg_error (GPG_ERR_ASS_UNKNOWN_INQUIRE);

  if (stat (filename, &st) == -1)
    return gpg_error_from_errno (errno);

  if (crypto->agent->inquire_maxlen
      && st.st_size > crypto->agent->inquire_maxlen)
    {
      log_write (_("The passphrase is too large: have=%u, max=%u."),
		 (unsigned) st.st_size, crypto->agent->inquire_maxlen);
      return GPG_ERR_TOO_LARGE;
    }

  buf = gcry_malloc_secure (st.st_size);
  if (!buf)
    return GPG_ERR_ENOMEM;

  fd = open (filename, O_RDONLY);
  if (fd == -1)
    rc = gpg_error_from_errno (errno);
  else
    {
      size_t len = read (fd, buf, st.st_size);

      if (len == st.st_size)
	{
	  assuan_begin_confidential (crypto->agent->ctx);
	  rc = assuan_send_data (crypto->agent->ctx, buf, len);
	  assuan_end_confidential (crypto->agent->ctx);
	}
      else if (len == -1)
	rc = gpg_error_from_errno (errno);
      else
	rc = GPG_ERR_BUFFER_TOO_SHORT;
    }

  if (fd != -1)
    close (fd);

  gcry_free (buf);
  return rc;
}

gpg_error_t
agent_export_common (struct crypto_s * crypto, const char *hexgrip,
		     const char *sign_hexgrip, int no_passphrase,
		     const void *data, size_t datalen, const char *outfile,
		     const char *keyparams, const char *keyfile)
{
  gpg_error_t rc = 0;

  if (!sign_hexgrip && hexgrip)
    sign_hexgrip = hexgrip;

  if (sign_hexgrip)
    {
      if (crypto->sigpkey_sexp)
	gcry_sexp_release (crypto->sigpkey_sexp);

      crypto->sigpkey_sexp = NULL;
      rc = get_pubkey (crypto, sign_hexgrip, &crypto->save.sigpkey);
      if (rc)
	return rc;

      gcry_pk_get_keygrip (crypto->save.sigpkey, crypto->sign_grip);
    }

  if (hexgrip)
    {
      rc = get_pubkey (crypto, hexgrip, &crypto->save.pkey);
      if (!rc)
	gcry_pk_get_keygrip (crypto->save.pkey, crypto->grip);
    }
  else
    {
      struct inquire_data_s idata = { 0 };
      char *params = keyparams ? str_dup (keyparams)
	: default_key_params (crypto);

      pthread_cleanup_push (xfree, params);
      log_write (_("Generating a new keypair ..."));
      if (keyfile)
	{
	  log_write (_("Using passphrase obtained from file '%s'"), keyfile);
	  rc = set_pinentry_mode (crypto->agent, "loopback");
	  crypto->agent->inquire_cb = inquire_keyfile;
	  crypto->agent->inquire_data = crypto;
	  crypto->agent->inquire_data2 = (char *) keyfile;
	  crypto->agent->inquire_data3 = params;
	}
      else
	{
	  idata.line = params;
	  idata.len = strlen (params);
	  idata.crypto = crypto;
	  crypto->agent->inquire_cb = inquire_cb;
	  crypto->agent->inquire_data = &idata;
	}

      if (!rc)
	{
	  rc = generate_key (crypto, params, no_passphrase, 1);
	  if (!rc)
	    gcry_pk_get_keygrip (crypto->save.pkey, crypto->grip);
	}

      (void) set_pinentry_mode (crypto->agent, "ask");
      pthread_cleanup_pop (1);
    }

  if (!rc)
    {
      rc = encrypt_data_file (NULL, crypto, crypto->save.pkey,
			      crypto->save.sigpkey, outfile, data, datalen);
      if (!rc)
	{
	  char *tmp = bin2hex (crypto->grip, sizeof (crypto->grip));

	  log_write (_("Success! Keygrip is %s."), tmp);
	  rc = send_to_agent (crypto->agent, NULL, NULL,
			      "CLEAR_PASSPHRASE --mode=normal %s", tmp);
	  xfree (tmp);

	  if (sign_hexgrip)
	    {
	      tmp = bin2hex (crypto->sign_grip, sizeof (crypto->sign_grip));
	      log_write (_("Signed with keygrip %s."), tmp);
	      xfree (tmp);
	    }
	}
    }

  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

char *
default_key_params (struct crypto_s *crypto)
{
  return config_get_string (NULL, "keyparam");
}

gpg_error_t
agent_passwd (struct crypto_s * crypto)
{
  struct inquire_data_s idata = { 0 };
  gpg_error_t rc;
  char *tmp = bin2hex (crypto->grip, 20);

  idata.crypto = crypto;
  crypto->agent->inquire_cb = inquire_cb;
  crypto->agent->inquire_data = &idata;
  rc = send_to_agent (crypto->agent, NULL, NULL, "PASSWD --preset %s", tmp);
  xfree (tmp);
  return rc;
}

gpg_error_t
kill_scd (struct agent_s * agent)
{
  gpg_error_t rc = 0;

  if (!use_agent)
    return 0;

  if (config_get_boolean (NULL, "kill_scd"))
    {
      rc = send_to_agent (agent, NULL, NULL, "SCD KILLSCD");
      if (rc && gpg_err_code (rc) != GPG_ERR_NO_SCDAEMON
          && gpg_err_code (rc) != GPG_ERR_EPIPE)
	log_write ("%s: ERR %u: %s", __FUNCTION__, rc, pwmd_strerror (rc));
    }

  return rc;
}
