/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <ctype.h>
#include <errno.h>

#include "pwmd-error.h"
#include <gcrypt.h>
#include "mutex.h"
#include "cache.h"
#include "status.h"
#include "mem.h"
#include "util-misc.h"
#include "util-string.h"
#include "crypto.h"

#ifdef HAVE_TIME_H
#include <time.h>
#endif

static pthread_mutex_t cache_mutex;
static struct slist_s *key_cache;
#ifdef WITH_AGENT
static struct agent_s *cache_agent;
#endif

extern void log_write (const char *fmt, ...);

typedef gpg_error_t (*free_data_fn_t) (file_cache_t *);
static free_data_fn_t free_data_fn;
static gpg_error_t clear_once (file_cache_t * p);
static int remove_entry (const unsigned char *md5file);
static void free_entry (file_cache_t * p);

static file_cache_t *
get_entry (const unsigned char *md5file)
{
  int t = slist_length (key_cache);
  int i;

  if (!md5file)
    return NULL;

  for (i = 0; i < t; i++)
    {
      file_cache_t *p = slist_nth_data (key_cache, i);

      if (!memcmp (p->filename, md5file, sizeof (p->filename)))
	return p;
    }

  return NULL;
}

gpg_error_t
cache_lock_mutex (void *ctx, const unsigned char *md5file,
		  long lock_timeout, int add, int timeout)
{
  MUTEX_LOCK (&cache_mutex);
  gpg_error_t rc = 0;
  file_cache_t *p = get_entry (md5file);

  if (p)
    {
      MUTEX_UNLOCK (&cache_mutex);
      MUTEX_TRYLOCK (ctx, p->mutex, rc, lock_timeout);
    }
  else if (add)
    {
      rc = cache_add_file (md5file, NULL, NULL, timeout);
      if (!rc)
	{
	  p = get_entry (md5file);
	  MUTEX_UNLOCK (&cache_mutex);
	  MUTEX_TRYLOCK (ctx, p->mutex, rc, lock_timeout);
	}
      else
	{
	  MUTEX_UNLOCK (&cache_mutex);
	}
    }
  else
    {
      MUTEX_UNLOCK (&cache_mutex);
    }

  return !p && !rc ? GPG_ERR_NO_DATA : rc;
}

static int
remove_entry (const unsigned char *md5file)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);

  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
  if (p)
    {
      /* Keep a refcount because another client maybe editing this same new
       * file. The entry needs to be kept in case the other client SAVE's.
       */
      p->refcount--;
      if (!p->refcount)
	free_entry (p);
    }

  pthread_cleanup_pop (1);
  return p ? 1 : 0;
}

gpg_error_t
cache_unlock_mutex (const unsigned char *md5file, int remove)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);

  if (p)
    {
      MUTEX_UNLOCK (p->mutex);
      if (remove)
	remove_entry (md5file);
    }

  MUTEX_UNLOCK (&cache_mutex);
  return p ? 0 : GPG_ERR_NO_DATA;
}

#ifdef WITH_AGENT
static int
valid_agent_grip (file_cache_t *e)
{
  size_t c;

  for (c = 0; c < sizeof (e->grip); c++)
    {
      if (e->grip[c] && e->grip[c] != '0')
	return 1;
    }

  return 0;
}
#endif

static int
valid_grip (file_cache_t * e)
{
  if (e->data && e->data->key)
    return 1;

#ifdef WITH_AGENT
  return valid_agent_grip (e);
#else
  return 0;
#endif
}

static gpg_error_t
iscached (const unsigned char *md5file, int *defer)
{
  gpg_error_t rc = 0;
  file_cache_t *p = get_entry (md5file);

  if (!p || !p->data || !valid_grip (p))
    return GPG_ERR_NO_DATA;

  if (defer)
    *defer = p->defer_clear;

  if (p->data && p->data->key)
    return 0;

#ifdef WITH_AGENT
  char *line;

  rc = send_to_agent (cache_agent, &line, NULL, "KEYINFO --data %s", p->grip);
  if (!rc)
    {
      char **fields = str_split (line, " ", 0);

      /* Smartcard with a cached document. */
      if (*fields[1] == 'T' && p->data->doc)
	rc = 0;
      else
	rc = isdigit (*fields[4]) || *fields[5] == 'C' ? 0
          : GPG_ERR_KEY_EXPIRED;

      strv_free (fields);
      xfree (line);
    }
#endif

  return rc;
}

unsigned
cache_file_count ()
{
  MUTEX_LOCK (&cache_mutex);
  unsigned total = 0, i, n;

  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
  n = slist_length (key_cache);
  for (i = 0; i < n; i++)
    {
      file_cache_t *p = slist_nth_data (key_cache, i);

      if (!iscached (p->filename, NULL))
	total++;
    }

  pthread_cleanup_pop (1);
  return total;
}

void
cache_adjust_timeout ()
{
  MUTEX_LOCK (&cache_mutex);
  int t = slist_length (key_cache);
  int i;

  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);

  for (i = 0; i < t; i++)
    {
      file_cache_t *p = slist_nth_data (key_cache, i);

      if (p->timeout > 0)
	p->timeout--;

      if (!p->timeout || p->defer_clear)
	{
	  /* The file associated with this cache entry may be locked by a
	   * client. Defer freeing this cache entry until the next timeout
	   * check. */
	  if (!clear_once (p))
	    send_status_all (STATUS_CACHE, NULL);
	}
    }

  pthread_cleanup_pop (1);
}

static void
setgrip (file_cache_t * p, const unsigned char *grip)
{
  if (grip)
    {
      char *tmp = bin2hex (grip, 20);

      memcpy (p->grip, tmp, sizeof (p->grip));
      xfree (tmp);
    }
  else
    memset (p->grip, 0, sizeof(p->grip));
}

static gpg_error_t
set_timeout (const unsigned char *md5file, int timeout, int defer, int force)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);
  gpg_error_t rc = 0;

  if (!p)
    {
      MUTEX_UNLOCK (&cache_mutex);
      return GPG_ERR_NOT_FOUND;
    }

  p->reset = timeout;
  if (!p->defer_clear && (p->timeout == -1 || force))
    p->timeout = timeout;

  if (!timeout || defer)
    {
      rc = clear_once (p);
      if (!rc)
	send_status_all (STATUS_CACHE, NULL);
    }

  MUTEX_UNLOCK (&cache_mutex);
  return rc;
}

gpg_error_t
cache_set_data (const unsigned char *md5file, struct cache_data_s * data,
		const unsigned char *grip)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);

  if (p)
    {
      setgrip (p, grip);
      p->data = data;
      set_timeout (md5file, p->reset, p->defer_clear, 0);
      if (!p->reset)
	{
	  clear_once (p);
	  send_status_all (STATUS_CACHE, NULL);
	}
    }

  MUTEX_UNLOCK (&cache_mutex);
  return p ? 0 : GPG_ERR_NO_DATA;
}

struct cache_data_s *
cache_get_data (const unsigned char *md5file)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);

  MUTEX_UNLOCK (&cache_mutex);
  return p ? p->data : NULL;
}


struct cache_data_s *
cache_get_data_filename (const char *filename)
{
  unsigned char md5file[16];

  gcry_md_hash_buffer (GCRY_MD_MD5, md5file, filename, strlen (filename));
  return cache_get_data (md5file);
}

gpg_error_t
cache_add_file (const unsigned char *md5file, const unsigned char *grip,
		struct cache_data_s *data, int timeout)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);
  struct slist_s *new;
  gpg_error_t rc;

  if (p)
    {
      setgrip (p, grip);
      p->data = data;
      p->refcount++;
      rc = set_timeout (md5file, timeout, p->defer_clear, 0);
      MUTEX_UNLOCK (&cache_mutex);
      send_status_all (STATUS_CACHE, NULL);
      return rc;
    }

  p = xcalloc (1, sizeof (file_cache_t));
  if (!p)
    {
      MUTEX_UNLOCK (&cache_mutex);
      return GPG_ERR_ENOMEM;
    }

  p->mutex = (pthread_mutex_t *) xmalloc (sizeof (pthread_mutex_t));
  if (!p->mutex)
    {
      xfree (p);
      MUTEX_UNLOCK (&cache_mutex);
      return GPG_ERR_ENOMEM;
    }

  pthread_mutexattr_t attr;
  pthread_mutexattr_init (&attr);
  pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init (p->mutex, &attr);
  pthread_mutexattr_destroy (&attr);
  memcpy (p->filename, md5file, sizeof (p->filename));
  setgrip (p, grip);
  p->data = data;
  p->refcount++;
  new = slist_append (key_cache, p);
  if (!new)
    {
      pthread_mutex_destroy (p->mutex);
      xfree (p->mutex);
      xfree (p);
      MUTEX_UNLOCK (&cache_mutex);
      return GPG_ERR_ENOMEM;
    }

  key_cache = new;
  rc = cache_set_timeout(md5file, timeout);
  MUTEX_UNLOCK (&cache_mutex);
  send_status_all (STATUS_CACHE, NULL);
  return rc;
}

static gpg_error_t
clear_once (file_cache_t * p)
{
  gpg_error_t rc = free_data_fn (p);

  if (!rc && valid_grip (p))
    {
#ifdef WITH_AGENT
      if (use_agent && valid_agent_grip (p))
	{
	  rc = send_to_agent (cache_agent, NULL, NULL,
			      "CLEAR_PASSPHRASE --mode=normal %s", p->grip);
	  if (rc)
	    log_write ("%s(): %s", __FUNCTION__, pwmd_strerror (rc));
	}
#endif
      memset (p->grip, 0, sizeof (p->grip));
    }

  return rc;
}

static void
free_entry (file_cache_t * p)
{
  gpg_error_t rc;

  if (!p)
    return;

  rc = clear_once (p);
  if (rc)
    log_write ("%s(): %s", __FUNCTION__, pwmd_strerror (rc));

  if (p->mutex)
    {
      pthread_mutex_destroy (p->mutex);
      xfree (p->mutex);
    }

  key_cache = slist_remove (key_cache, p);
  xfree (p);
}

gpg_error_t
cache_clear (const unsigned char *md5file)
{
  file_cache_t *p;
  gpg_error_t rc = 0;
  int i, t;

  MUTEX_LOCK (&cache_mutex);

  if (md5file)
    {
      p = get_entry (md5file);
      if (!p)
	{
	  MUTEX_UNLOCK (&cache_mutex);
	  return 0;
	}

      pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
      rc = clear_once (p);
      pthread_cleanup_pop (1);
      if (rc)
	log_write ("%s(): %s", __FUNCTION__, pwmd_strerror (rc));

      return rc;
    }

  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
  t = slist_length (key_cache);
  for (i = 0; i < t; i++)
    {
      p = slist_nth_data (key_cache, i);
      clear_once (p);
    }

  pthread_cleanup_pop (1);
  return 0;
}

gpg_error_t
cache_iscached (const char *filename, int *defer)
{
  gpg_error_t rc;
  unsigned char md5file[16];

  if (!filename)
    return GPG_ERR_INV_PARAMETER;

  MUTEX_LOCK (&cache_mutex);
  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
  gcry_md_hash_buffer (GCRY_MD_MD5, md5file, filename, strlen (filename));
  rc = iscached (md5file, defer);
  pthread_cleanup_pop (1);

  /* Test if the data file disappeared from the filesystem. */
  if ((!rc || gpg_err_code (rc) == GPG_ERR_NO_DATA)
      && access (filename, R_OK) == -1)
    {
      rc = gpg_error_from_errno (errno);
      if (gpg_err_code (rc) == GPG_ERR_ENOENT)
	{
	  /* When -1, don't clear the cache entry. Fixes clearing the
	   * cache entry when saving a new file. */
	  if ((defer && *defer != -1) || !defer)
	    {
	      rc = cache_defer_clear (md5file);
	      if (!rc && defer)
		*defer = 1;
	    }

	  rc = GPG_ERR_ENOENT;
	}
    }

  return rc;
}

gpg_error_t
cache_defer_clear (const unsigned char *md5file)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);
  gpg_error_t rc = 0;

  if (!p)
    rc = GPG_ERR_NOT_FOUND;
  else
    p->defer_clear = 1;

  MUTEX_UNLOCK (&cache_mutex);
  return rc;
}

gpg_error_t
cache_set_timeout (const unsigned char *md5file, int timeout)
{
  return set_timeout (md5file, timeout, 0, 1);
}

int
cache_get_grip (const unsigned char *md5file, unsigned char *grip)
{
  MUTEX_LOCK (&cache_mutex);
  file_cache_t *p = get_entry (md5file);

  if (!p || !valid_grip (p))
    {
      MUTEX_UNLOCK (&cache_mutex);
      return 0;
    }

  memcpy (grip, p->grip, sizeof (p->grip));
  MUTEX_UNLOCK (&cache_mutex);
  return 1;
}

/* The atfork parameter is needed to prevent clearing gpg-agent
   cached keys during an atfork callback since the agent is still
   shared between the two processes. */
void
cache_deinit (int atfork)
{
  MUTEX_LOCK (&cache_mutex);
  int i, t = slist_length (key_cache);

  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
  for (i = 0; i < t; i++)
    {
      file_cache_t *p = slist_nth_data (key_cache, i);

#ifdef WITH_AGENT
      if (atfork && valid_agent_grip (p))
	  continue;
#endif

      free_entry (p);
      t = slist_length (key_cache);
      i = -1;
    }

#ifdef WITH_AGENT
  if (use_agent && cache_agent)
    cleanup_agent (cache_agent);

  cache_agent = NULL;
#endif
  gcry_free (cache_key);
  xfree (cache_iv);
  cache_key = NULL;
  cache_iv = NULL;
  pthread_cleanup_pop (1);
  pthread_mutex_destroy (&cache_mutex);
}

void
cache_mutex_init ()
{
  pthread_mutexattr_t attr;

  pthread_mutexattr_init (&attr);
  pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init (&cache_mutex, &attr);
  pthread_mutexattr_destroy (&attr);
}

gpg_error_t
cache_init (free_data_fn_t fn)
{
  gpg_error_t rc = 0;

  if (!cache_key)
    {
      rc = gcry_cipher_algo_info (GCRY_CIPHER_AES, GCRYCTL_GET_BLKLEN, NULL,
				  &cache_blocksize);
      if (rc)
	return rc;

      rc = gcry_cipher_algo_info (GCRY_CIPHER_AES, GCRYCTL_GET_KEYLEN, NULL,
				  &cache_keysize);
      if (rc)
	return rc;

      cache_key = gcry_malloc (cache_keysize);
      if (!cache_key)
	return GPG_ERR_ENOMEM;

      cache_iv = xmalloc (cache_blocksize);
      if (!cache_iv)
	{
	  gcry_free (cache_key);
	  cache_key = NULL;
	  return GPG_ERR_ENOMEM;
	}

      gcry_create_nonce (cache_key, cache_keysize);
      gcry_create_nonce (cache_iv, cache_blocksize);
    }

#ifdef WITH_AGENT
  if (use_agent && !cache_agent)
    {
      rc = agent_init (&cache_agent);
      if (!rc)
	{
	  char *line;

	  rc = send_to_agent (cache_agent, &line, NULL, "GETINFO version");
	  if (!rc)
	    {
	      char **fields = str_split (line, ".", 0);
	      int major, minor;

	      major = atoi (fields[0]);
	      minor = atoi (fields[1]);
	      if (major < 2 || minor < 1)
		rc = GPG_ERR_UNKNOWN_VERSION;

	      strv_free (fields);
	      xfree (line);
	    }
	}
    }
#endif

  cache_mutex_init ();
  free_data_fn = fn;
  return rc;
}

#ifdef WITH_AGENT
gpg_error_t
cache_is_shadowed (const char *grip)
{
  MUTEX_LOCK (&cache_mutex);
  char *line;
  size_t len;
  gpg_error_t rc;

  pthread_cleanup_push (cleanup_mutex_cb, &cache_mutex);
  rc = send_to_agent (cache_agent, &line, &len, "KEYINFO --data %s", grip);
  pthread_cleanup_pop (1);

  if (!rc)
    {
      char **fields = str_split (line, " ", 0);

      rc = (*fields[1] == 'T') ? 0 : GPG_ERR_NO_DATA;
      xfree (line);
      strv_free (fields);
    }

  return rc;
}
#endif

void
cache_lock ()
{
  MUTEX_LOCK (&cache_mutex);
}

void
cache_unlock ()
{
  MUTEX_UNLOCK (&cache_mutex);
}

void
free_cache_data_once (struct cache_data_s *data)
{
  if (!data)
    return;

#ifdef WITH_AGENT
  if (data->pubkey)
    gcry_sexp_release (data->pubkey);

  if (data->sigkey)
    gcry_sexp_release (data->sigkey);
#endif

  gcry_free (data->doc);
  gcry_free (data->key);
  xfree (data->crc);
  xfree (data);
}
