/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>
#include <zlib.h>
#include <errno.h>

#include "pwmd-error.h"
#include <gcrypt.h>
#include "convert.h"
#include "xml.h"
#include "mem.h"
#include "util-misc.h"
#include "cache.h"
#include "crypto.h"
#include "pinentry.h"

#ifndef _
#include "gettext.h"
#define _(msgid) gettext(msgid)
#endif

#define KEYSIZE	32

typedef struct
{
  uint8_t magic[5];
  uint16_t version;
  uint64_t iter;
  uint64_t flags;
  uint8_t iv[16];
  uint8_t salt[8];
} v2_file_header_t;

struct gz_s
{
  z_stream z;
  void *out;
  int done;
};

static unsigned char crypto_magic[5] = "\177PWMD";

static void
gz_cleanup (void *arg)
{
  struct gz_s **gz = (struct gz_s **) arg;

  if (!gz)
    return;

  if (!(*gz)->done && (*gz)->out)
    gcry_free ((*gz)->out);

  if ((*gz)->z.zalloc)
    inflateEnd (&(*gz)->z);

  xfree (*gz);
  *gz = NULL;
}

static void *
z_alloc (void *data, unsigned items, unsigned size)
{
  return gcry_calloc (items, size);
}

static void
z_free (void *data, void *p)
{
  gcry_free (p);
}

#define ZLIB_BUFSIZE	65536

static gpg_error_t
decompress (void *in, unsigned long insize, void * *out,
	    unsigned long *outsize)
{
  struct gz_s *gz;
  gz_header h;
  char buf[17];
  gpg_error_t rc;
  int zrc;

  gz = xcalloc (1, sizeof (struct gz_s));
  if (!gz)
    return GPG_ERR_ENOMEM;

  gz->z.zalloc = z_alloc;
  gz->z.zfree = z_free;
  gz->z.next_in = in;
  gz->z.avail_in = (uInt) insize;
  gz->z.avail_out = ZLIB_BUFSIZE;
  gz->z.next_out = gz->out = gcry_malloc (ZLIB_BUFSIZE);
  if (!gz->out)
    {
      gz_cleanup (&gz);
      return GPG_ERR_ENOMEM;
    }

  zrc = inflateInit2 (&gz->z, 47);
  if (zrc != Z_OK)
    {
      gz_cleanup (&gz);
      return zrc == Z_MEM_ERROR ? GPG_ERR_ENOMEM : GPG_ERR_COMPR_ALGO;
    }

  memset (&h, 0, sizeof (gz_header));
  h.comment = (unsigned char *) buf;
  h.comm_max = sizeof (buf);
  zrc = inflateGetHeader (&gz->z, &h);
  if (zrc != Z_OK)
    {
      gz_cleanup (&gz);
      return zrc == Z_MEM_ERROR ? GPG_ERR_ENOMEM : GPG_ERR_COMPR_ALGO;
    }

  zrc = inflate (&gz->z, Z_BLOCK);
  if (zrc != Z_OK)
    {
      gz_cleanup (&gz);
      return zrc == Z_MEM_ERROR ? GPG_ERR_ENOMEM : GPG_ERR_COMPR_ALGO;
    }

  if (h.comment)
    insize = strtoul ((char *) h.comment, NULL, 10);

  do
    {
      void *p;

      zrc = inflate (&gz->z, Z_FINISH);
      switch (zrc)
	{
	case Z_OK:
	  break;
	case Z_BUF_ERROR:
	  if (!gz->z.avail_out)
	    {
	      p = gcry_realloc (gz->out, gz->z.total_out + ZLIB_BUFSIZE);
	      if (!p)
		{
		  rc = GPG_ERR_ENOMEM;
		  goto fail;
		}

	      gz->out = p;
	      gz->z.next_out = (unsigned char *) gz->out + gz->z.total_out;
	      gz->z.avail_out = ZLIB_BUFSIZE;
	    }
	  else
	    {
	      rc = GPG_ERR_COMPR_ALGO;
	      goto fail;
	    }

	  break;
	case Z_STREAM_END:
	  break;
	default:
	  rc = GPG_ERR_COMPR_ALGO;
	  goto fail;
	  break;
	}
    }
  while (zrc != Z_STREAM_END);

  *out = gz->out;
  *outsize = gz->z.total_out;
  gz->done = 1;
  gz_cleanup (&gz);
  return 0;

fail:
  gz_cleanup (&gz);
  return rc;
}

static gpg_error_t
decrypt (v2_file_header_t * fh, unsigned char *key,
	 unsigned char **ciphertext, size_t datalen, void * *result,
	 size_t * result_len)
{
  gpg_error_t rc;
  size_t blocksize, keysize, len;
  int algo = cipher_to_gcrypt (fh->flags);
  gcry_cipher_hd_t gh = NULL;
  uint64_t iter = 0ULL;
  unsigned char *data = *ciphertext;

  *result = NULL;
  *result_len = 0;

  /* No encryption iterations. This is a plain (gzipped) file. */
  if (!fh->iter)
    goto decompress;

  if (algo == -1)
    return GPG_ERR_CIPHER_ALGO;

  rc = gcry_cipher_algo_info (algo, GCRYCTL_TEST_ALGO, NULL, NULL);
  if (rc)
    return rc;

  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL, &keysize);
  if (rc)
    return rc;

  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_BLKLEN, NULL, &blocksize);
  if (rc)
    return rc;

  rc = gcry_cipher_open (&gh, algo, GCRY_CIPHER_MODE_CBC, 0);
  if (rc)
    return rc;

  if ((rc = gcry_cipher_setiv (gh, fh->iv, blocksize)))
    goto fail;

  if ((rc = gcry_cipher_setkey (gh, key, keysize)))
    goto fail;

  rc = gcry_cipher_decrypt (gh, data, datalen, NULL, 0);
  if (rc)
    goto fail;

  key[0] ^= 1;
  if ((rc = gcry_cipher_setkey (gh, key, keysize)))
    goto fail;

  while (++iter < fh->iter)
    {
      if ((rc = gcry_cipher_setiv (gh, fh->iv, blocksize)))
	goto fail;

      rc = gcry_cipher_decrypt (gh, data, datalen, NULL, 0);
      if (rc)
	goto fail;
    }

decompress:
  len = 0;
  if (fh->version >= 0x218 && fh->iter > 0ULL)
    {
      if (memcmp (data, crypto_magic, sizeof (crypto_magic)))
	{
	  rc = GPG_ERR_BAD_PASSPHRASE;
	  goto fail;
	}

      len = sizeof (crypto_magic);
    }

  rc =
    decompress (data + len, datalen - len, result,
		(unsigned long *) result_len);
  if (rc)
    {
      if (fh->version < 0x218 && rc == GPG_ERR_COMPR_ALGO)
	rc = GPG_ERR_BAD_PASSPHRASE;
      goto fail;
    }

  if (strncasecmp ((char *) *result, "<?xml ", 6) != 0)
    {
      gcry_free (*result);
      *result = NULL;
      *result_len = 0;
      rc = GPG_ERR_BAD_PASSPHRASE;
      goto fail;
    }

fail:
  if (gh)
    gcry_cipher_close (gh);

  return rc;
}

gpg_error_t
read_v2_datafile (const char *filename, const char *keyfile,
		  void * *result, size_t * result_len, uint16_t * ver,
		  int *algo)
{
  gpg_error_t rc;
  int fd;
  v2_file_header_t fh;
  size_t len, passphraselen = 0, datalen;
#ifdef WITH_AGENT
  struct agent_s *agent = NULL;
#endif
  char *passphrase = NULL;
  unsigned char *data = NULL;
  struct stat st;
  unsigned char *key = NULL;
  FILE *fp = NULL;

  if (stat (filename, &st) == -1)
    return gpg_error_from_errno (errno);

  fd = open (filename, O_RDONLY);
  if (fd == -1)
    return gpg_error_from_errno (errno);

  len = read (fd, &fh, sizeof (fh) - 8);
  if (len != sizeof (fh) - 8)
    {
      rc = GPG_ERR_INV_LENGTH;
      goto fail;
    }

  if (fh.version >= 0x221)
    {
      len = read (fd, &fh.salt, 8);
      if (len != 8)
	{
	  rc = GPG_ERR_INV_LENGTH;
	  goto fail;
	}
    }

  *ver = fh.version;
  *algo = cipher_to_gcrypt (fh.flags);

#ifdef WITH_AGENT
  if (use_agent && !keyfile && fh.iter > 0)
    {
      char *desc;

      rc = agent_init (&agent);
      if (rc)
	goto fail;

      desc = plus_escape (_
			  ("A passphrase is required to decrypt the file for "
			   "converting. Please enter the passphrase below."));
      rc = agent_set_pinentry_options (agent);
      if (rc)
	goto fail;

      assuan_begin_confidential (agent->ctx);
      rc = send_to_agent (agent, &passphrase, &passphraselen,
			  "GET_PASSPHRASE --data pwmd:convert + %s %s",
			  _("Passphrase:"), desc);
      assuan_end_confidential (agent->ctx);
      xfree (desc);
      if (rc)
	goto fail;

      passphraselen--;		// null byte
      send_to_agent (agent, NULL, NULL, "CLEAR_PASSPHRASE pwmd:convert");
      cleanup_agent (agent);
      agent = NULL;
    }
    else if (keyfile && fh.iter > 0)
#else
    if (keyfile && fh.iter > 0)
#endif
    {
      struct stat pst;
      int pfd = open (keyfile, O_RDONLY);

      if (pfd == -1)
	{
	  rc = gpg_error_from_errno (errno);
	  goto fail;
	}

      if (stat (keyfile, &pst) == -1)
	{
	  rc = gpg_error_from_errno (errno);
	  close (pfd);
	  goto fail;
	}

      passphrase = xmalloc (pst.st_size);
      if (!passphrase)
	{
	  rc = GPG_ERR_ENOMEM;
	  close (pfd);
	  goto fail;
	}

      passphraselen = read (pfd, passphrase, pst.st_size);
      close (pfd);
      if (passphraselen != pst.st_size)
	{
	  rc = GPG_ERR_INV_LENGTH;
	  goto fail;
	}
    }
    else if (fh.iter > 0)
      {
	rc = getpin_common (NULL, filename, PINENTRY_OPEN, &passphrase,
			    &passphraselen);
	if (rc)
	  goto fail;
      }

  key = gcry_malloc (KEYSIZE);
  if (!key)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  fp = fdopen (fd, "r");
  len = st.st_size - ftell (fp);

  if (fh.version >= 0x221 && fh.version < 0x0300)
    {
      rc = gcry_kdf_derive (passphrase, passphraselen, GCRY_KDF_ITERSALTED_S2K,
			    GCRY_MD_SHA1, fh.salt, 8, 1000, KEYSIZE, key);
      if (rc)
	goto fail;
    }
  else
    gcry_md_hash_buffer (GCRY_MD_SHA256, key, passphrase, passphraselen);

  data = gcry_malloc (len);
  if (!data)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  datalen = read (fd, data, len);
  if (datalen != len)
    {
      rc = GPG_ERR_INV_LENGTH;
      goto fail;
    }

  fprintf (stderr, _("Decrypting ...\n"));
  rc = decrypt (&fh, key, &data, datalen, result, result_len);

fail:
  gcry_free (key);
  gcry_free (data);
  xfree (passphrase);

#ifdef WITH_AGENT
  if (agent)
    cleanup_agent (agent);
#endif

  if (fd != -1)
    {
      if (fp)
        fclose (fp);
      else
        close (fd);
    }

  return rc;
}
