/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef WITH_LIBACL
#include <sys/acl.h>
#endif

#include "pwmd-error.h"
#include "util-misc.h"
#include "common.h"
#include "rcfile.h"
#include "pinentry.h"
#include "crypto.h"
#include "cache.h"
#include "mem.h"
#include "util-string.h"

static uint8_t pwmd_magic[5] = { '\177', 'P', 'W', 'M', 'D' };

/* Sets the default cipher values for new files. */
static void
set_header_defaults (file_header_t * hdr)
{
  char *s = config_get_string (NULL, "cipher");
  int flags = cipher_string_to_cipher (s);

  xfree (s);
  memset (hdr, 0, sizeof (file_header_t));
  memcpy (hdr->magic, pwmd_magic, sizeof (hdr->magic));
  if (flags == -1)
    log_write (_
	       ("Invalid 'cipher' in configuration file. Using a default of aes256."));
  hdr->flags = flags == -1 ? PWMD_CIPHER_AES256 : flags;
  hdr->version = VERSION_HEX;

#ifdef WITH_AGENT
  if (use_agent)
    hdr->flags |= PWMD_FLAG_PKI;
#endif
}

static gpg_error_t
read_header (file_header_t *hdr, int fd)
{
  ssize_t len;
  uint32_t i;
  uint64_t n;

#ifdef WORDS_BIGENDIAN
  len = read (fd, &hdr->magic, sizeof(hdr->magic));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->version, sizeof(hdr->version));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->s2k_count, sizeof(hdr->s2k_count));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->flags, sizeof(hdr->flags));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->iv, sizeof(hdr->iv));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->salt, sizeof(hdr->salt));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->datalen, sizeof(hdr->datalen));
  if (len == -1)
    goto done;
#else
  len = read (fd, &hdr->magic, sizeof(hdr->magic));
  if (len == -1)
    goto done;

  len = read (fd, &i, sizeof(hdr->version));
  if (len == -1)
    goto done;
  hdr->version = ntohl (i);

  len = read (fd, &n, sizeof(hdr->s2k_count));
  if (len == -1)
    goto done;
  hdr->s2k_count = (((uint64_t) ntohl (n)) << 32) + ntohl (n >> 32);

  len = read (fd, &n, sizeof(hdr->flags));
  if (len == -1)
    goto done;
  hdr->flags = (((uint64_t) ntohl (n)) << 32) + ntohl (n >> 32);

  len = read (fd, &hdr->iv, sizeof(hdr->iv));
  if (len == -1)
    goto done;

  len = read (fd, &hdr->salt, sizeof(hdr->salt));
  if (len == -1)
    goto done;

  len = read (fd, &i, sizeof(hdr->datalen));
  if (len == -1)
    goto done;
  hdr->datalen = ntohl (i);
#endif

 done:
  return len == -1 ? gpg_error_from_errno (errno) : 0;
}

/* Read the header of a data file to determine cipher and other. The
 * header is stored big endian in the file and is converted to little
 * endian when needed. */
gpg_error_t
read_data_header (const char *filename, file_header_t * rhdr,
		  struct stat *rst, int *rfd)
{
  gpg_error_t rc = 0;
  struct stat st;
  file_header_t hdr;
  int fd;

  if (lstat (filename, &st) == -1)
    return gpg_error_from_errno (errno);

  if (!S_ISREG (st.st_mode))
    return gpg_error (GPG_ERR_ENOANO);

  fd = open (filename, O_RDONLY);
  if (fd == -1)
    return gpg_error_from_errno (errno);

  rc = read_header (&hdr, fd);
  if (!rc && memcmp (hdr.magic, pwmd_magic, sizeof (hdr.magic)))
    rc = GPG_ERR_BAD_DATA;
  else if (!rc && hdr.version < 0x030000)
    rc = GPG_ERR_UNKNOWN_VERSION;

  if (rc)
    close (fd);
  else
    {
      if (rhdr)
	*rhdr = hdr;
      if (rst)
	*rst = st;
      if (rfd)
	*rfd = fd;
      else
	close (fd);
    }

  return gpg_error (rc);
}

gpg_error_t
read_data_file (const char *filename, struct crypto_s * crypto)
{
  int fd;
  gpg_error_t rc = 0;
  size_t len, rlen;
  char *buf = NULL;
  struct stat st;

  if (!filename)
    return GPG_ERR_INV_PARAMETER;

  cleanup_crypto_stage1 (crypto);
  rc = read_data_header (filename, &crypto->hdr, &st, &fd);
  if (rc)
    return gpg_error (rc);

  crypto->ciphertext_len = crypto->hdr.datalen;
  crypto->ciphertext = xmalloc (crypto->hdr.datalen);
  if (!crypto->ciphertext)
    {
      rc = GPG_ERR_ENOMEM;
      goto done;
    }

  /* The keygrips for PKI files are stored after the header. They are
   * stored in the file to let file(1) magic(5) show the grips. */
  if (crypto->hdr.flags & PWMD_FLAG_PKI)
    {
      rlen = read (fd, crypto->grip, 20);
      if (rlen != 20)
	{
	  rc = rlen == -1 ? gpg_error_from_errno (errno) : GPG_ERR_BAD_DATA;
	  goto done;
	}

      rlen = read (fd, crypto->sign_grip, 20);
      if (rlen != 20)
	{
	  rc = rlen == -1 ? gpg_error_from_errno (errno) : GPG_ERR_BAD_DATA;
	  goto done;
	}
    }

  len = read (fd, crypto->ciphertext, crypto->hdr.datalen);
  if (len != crypto->hdr.datalen)
    {
      rc = len == -1 ? gpg_error_from_errno (errno) : GPG_ERR_BAD_DATA;
      goto done;
    }

  if (!(crypto->hdr.flags & PWMD_FLAG_PKI))
    goto done;

  if (!use_agent)
    {
      rc = GPG_ERR_NOT_IMPLEMENTED;
      goto done;
    }

#ifdef WITH_AGENT
  len = st.st_size - sizeof (file_header_t) - crypto->hdr.datalen - 40;
  buf = xmalloc (len);
  if (!buf)
    {
      rc = GPG_ERR_ENOMEM;
      goto done;
    }

  /* Remaining data file bytes are the encrypted key and XML. */
  rlen = read (fd, buf, len);
  if (rlen != len)
    {
      rc = rlen == -1 ? gpg_error_from_errno (errno) : GPG_ERR_BAD_DATA;
      goto done;
    }

  rc = gcry_sexp_new (&crypto->ciphertext_sexp, buf, rlen, 1);
  if (rc)
    goto done;

  if (crypto->pkey_sexp)
    gcry_sexp_release (crypto->pkey_sexp);

  if (crypto->sigpkey_sexp)
    gcry_sexp_release (crypto->sigpkey_sexp);

  crypto->pkey_sexp = crypto->sigpkey_sexp = NULL;
  rc = get_pubkey_bin (crypto, crypto->grip, &crypto->pkey_sexp);
  if (!rc)
    rc = get_pubkey_bin (crypto, crypto->sign_grip, &crypto->sigpkey_sexp);
#endif

done:
  close (fd);
  xfree (buf);
  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

#define CRYPTO_BLOCKSIZE(c) (blocksize * 1024)

/*
 * Useful for a large amount of data. Rather than doing all of the data in one
 * iteration do it in chunks.  This lets the command be cancelable rather than
 * waiting for it to complete.
 */
static gpg_error_t
iterate_crypto_once (gcry_cipher_hd_t h, unsigned char *inbuf,
		     size_t insize, size_t blocksize, status_msg_t which)
{
  gpg_error_t rc = 0;
  off_t len = CRYPTO_BLOCKSIZE (blocksize);
  void *p = gcry_malloc (len);
  off_t total = 0;

  if (!p)
    return gpg_error (GPG_ERR_ENOMEM);

  if (insize < CRYPTO_BLOCKSIZE (blocksize))
    len = insize;

  pthread_cleanup_push (gcry_free, p);

  for (;;)
    {
      unsigned char *inbuf2 = inbuf + total;
      unsigned char *tmp;

      if (len + total > insize)
	len = blocksize;

      if (which == STATUS_ENCRYPT)
	rc = gcry_cipher_encrypt (h, p, len, inbuf2, len);
      else
	rc = gcry_cipher_decrypt (h, p, len, inbuf2, len);

      if (rc)
	break;

      tmp = inbuf + total;
      memmove (tmp, p, len);
      total += len;
      if (total >= insize)
	break;

#ifdef HAVE_PTHREAD_TESTCANCEL
      pthread_testcancel ();
#endif
    }

  pthread_cleanup_pop (1);
  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

static void
cleanup_cipher (void *arg)
{
  gcry_cipher_close ((gcry_cipher_hd_t) arg);
}

/* Decrypt the XML data. For PKI data files the key is retrieved from
 * gpg-agent and the signature verified. */
gpg_error_t
decrypt_data (assuan_context_t ctx, struct crypto_s *crypto,
	      unsigned char *salted_key, size_t skeylen)
{
  gpg_error_t rc = 0;
  unsigned char *key = salted_key;
  gcry_cipher_hd_t h = NULL;
  size_t blocksize, keysize = 0;
  int algo = cipher_to_gcrypt (crypto->hdr.flags);
  void *outbuf = NULL;
  uint64_t n = crypto->hdr.s2k_count;
  long progress = 0;
#ifdef WITH_AGENT
  size_t keylen = skeylen;
  gcry_sexp_t sig_sexp;

  if (crypto->hdr.flags & PWMD_FLAG_PKI)
    {
      rc = agent_extract_key (crypto, &key, &keylen);
      if (rc)
	return rc;

      sig_sexp = gcry_sexp_find_token (crypto->ciphertext_sexp, "sig-val", 0);
      if (!sig_sexp)
	{
	  gcry_free (key);
	  return GPG_ERR_BAD_DATA;
	}

      rc = agent_verify (crypto->sigpkey_sexp, sig_sexp, crypto->ciphertext,
		   crypto->ciphertext_len);
      gcry_sexp_release (sig_sexp);
    }
#endif

  if (!rc)
    {
      rc = gcry_cipher_open (&h, algo, GCRY_CIPHER_MODE_CBC, 0);
      if (!rc)
	{
	  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL,
				      &keysize);
	  if (!rc)
	    {
	      rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_BLKLEN, NULL,
					  &blocksize);
	      if (!rc)
		{
		  rc = gcry_cipher_setiv (h, crypto->hdr.iv, blocksize);
		  if (!rc)
		    {
		      rc = gcry_cipher_setkey (h, key, keysize);
		    }
		}
	    }
	}
    }

  pthread_cleanup_push (cleanup_cipher, rc ? NULL : h);

  if (!rc)
    {
      outbuf = gcry_malloc (crypto->hdr.datalen);
      if (!outbuf)
	rc = GPG_ERR_ENOMEM;
    }

  pthread_cleanup_push (gcry_free, outbuf);
#ifdef WITH_AGENT
  pthread_cleanup_push (gcry_free, key);
#endif

  if (!rc)
    {
      if (!key)
        rc = GPG_ERR_INV_PARAMETER;
      else
        {
          memcpy (outbuf, crypto->ciphertext, crypto->hdr.datalen);
          rc = iterate_crypto_once (h, outbuf, crypto->hdr.datalen, blocksize,
                                    STATUS_DECRYPT);
        }

      if (!rc && crypto->hdr.version <= 0x03000e)
	{
	  key[0] ^= 1;
	  rc = gcry_cipher_setkey (h, key, keysize);
	}
    }

  if (!rc && crypto->hdr.version <= 0x03000e)
    {
      progress = config_get_long (NULL, "cipher_progress");
      if (progress == -1)
	progress = strtol (DEFAULT_ITERATION_PROGRESS, NULL, 10);
    }

  if (ctx && crypto->hdr.s2k_count)
    rc = send_status (ctx, STATUS_DECRYPT, "%llu %llu", 0,
                      crypto->hdr.version <= 0x03000e ? crypto->hdr.s2k_count : 1);

  if (!rc && crypto->hdr.version <= 0x03000e)
    {
     for (n = 0; !rc && n < crypto->hdr.s2k_count; n++)
       {
         if (ctx && progress && !(n % progress))
           {
             rc = send_status (ctx, STATUS_DECRYPT, "%llu %llu", n,
                               crypto->hdr.s2k_count);
             if (rc)
               break;
           }

         rc = gcry_cipher_setiv (h, crypto->hdr.iv, blocksize);
         if (rc)
           break;

         rc = iterate_crypto_once (h, outbuf, crypto->hdr.datalen, blocksize,
                                   STATUS_DECRYPT);
       }
    }

  if (!rc && ctx && crypto->hdr.s2k_count)
    rc = send_status (ctx, STATUS_DECRYPT, "%llu %llu", 
                      crypto->hdr.version <= 0x03000e ? n : 1,
                      crypto->hdr.version <= 0x03000e ? crypto->hdr.s2k_count : 1);

#ifdef WITH_AGENT
  pthread_cleanup_pop (0);
  if (crypto->hdr.flags & PWMD_FLAG_PKI)
    gcry_free (key);
  else if (key && crypto->hdr.version <= 0x03000e)
    key[0] ^= 1;
#else
  if (crypto->hdr.version <= 0x03000e)
    key[0] ^= 1;
#endif
  pthread_cleanup_pop (rc ? 1 : 0);	/* outbuf */
  pthread_cleanup_pop (1);              /* cipher */
  if (!rc)
    {
      char buf[] = "<?xml ";

      if (memcmp (outbuf, buf, sizeof(buf)-1))
	{
	  gcry_free (outbuf);
	  return gpg_error (GPG_ERR_BAD_PASSPHRASE);
	}

      crypto->plaintext = outbuf;
      crypto->plaintext_len = crypto->hdr.datalen;
    }

  if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_UNKNOWN)
    rc = gpg_error (rc);

  return rc;
}

/* The cached document is encrypted with encrypt_xml(). This decrypts
 * it from the OPEN command. */
gpg_error_t
decrypt_cache (struct crypto_s * crypto, const void *data, size_t len)
{
  gcry_cipher_hd_t h = NULL;
  gpg_error_t rc;

  rc = gcry_cipher_open (&h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_CBC, 0);
  if (rc)
    return rc;

  gcry_free (crypto->plaintext);
  crypto->plaintext = gcry_malloc (len);
  if (!crypto->plaintext)
    {
      rc = GPG_ERR_ENOMEM;
      goto done;
    }

  rc = gcry_cipher_setiv (h, cache_iv, cache_blocksize);
  if (rc)
    goto done;

  rc = gcry_cipher_setkey (h, cache_key, cache_keysize);
  if (rc)
    goto done;

  rc = gcry_cipher_decrypt (h, crypto->plaintext, len, data, len);
  if (rc || strncmp (crypto->plaintext, "<?xml ", 6))
    {
      if (!rc)
	rc = GPG_ERR_BAD_DATA;

      gcry_free (crypto->plaintext);
      crypto->plaintext = NULL;
    }

  crypto->plaintext_len = len;

done:
  if (h)
    gcry_cipher_close (h);

  return rc;
}

/* Encrypt the XML data and set 'result'. At least one encryption
 * iteration is done. */
gpg_error_t
encrypt_xml (assuan_context_t ctx, void *key, size_t keylen,
	     int algo, const void *xml, size_t len, void * *result,
	     size_t *result_len, unsigned char **iv, size_t * iv_len)
{
  gpg_error_t rc;
  gcry_cipher_hd_t h;
  size_t blocksize, keysize;
  void *inbuf = NULL;
  int init_iv = *(iv_len) == 0;
  size_t olen = len;

  if (!key)
    return GPG_ERR_INV_PARAMETER;

  rc = gcry_cipher_open (&h, algo, GCRY_CIPHER_MODE_CBC, 0);
  if (rc)
    return rc;

  pthread_cleanup_push (cleanup_cipher, h);
  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL, &keysize);
  if (!rc)
    rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_BLKLEN, NULL, &blocksize);

  if (!rc && init_iv)
    {
      *(iv) = xmalloc (blocksize);
      if (!*(iv))
	rc = GPG_ERR_ENOMEM;
      else
        gcry_create_nonce (*(iv), blocksize);
    }

  if (!rc)
    {
      if (init_iv)
	*iv_len = blocksize;

      if (len % blocksize)
        len += blocksize - (len % blocksize);

      inbuf = gcry_malloc (len);
      if (!inbuf)
	rc = GPG_ERR_ENOMEM;
      else
        {
          memset (inbuf, 0, len);
          memcpy (inbuf, xml, olen);
        }
    }

  pthread_cleanup_push (gcry_free, inbuf);

  if (!rc && ctx)
    rc = send_status (ctx, STATUS_ENCRYPT, "%llu %llu", 0, 1);

  if (!rc)
    {
      rc = gcry_cipher_setiv (h, *(iv), blocksize);
      if (!rc)
	{
	  rc = gcry_cipher_setkey (h, key, keysize);
	  if (!rc)
            {
	    rc = iterate_crypto_once (h, inbuf, len, blocksize,
				      STATUS_ENCRYPT);
            }
	}
    }

  if (!rc && ctx)
    rc = send_status (ctx, STATUS_ENCRYPT, "%llu %llu", 1, 1);

  pthread_cleanup_pop (rc ? 1 : 0);	        /* inbuf */
  pthread_cleanup_pop (1);	                /* cipher */

  if (rc && init_iv)
    {
      xfree (*(iv));
      *iv = NULL;
    }

  *result = rc ? NULL : inbuf;
  *result_len = len;
  return rc;
}

void
cleanup_save (struct save_s *save)
{
  if (!save)
    return;

#ifdef WITH_AGENT
  if (save->pkey)
    gcry_sexp_release (save->pkey);

  if (save->sigpkey)
    gcry_sexp_release (save->sigpkey);
#endif

  memset (save, 0, sizeof (struct save_s));
}

/* Keep the agent ctx to retain pinentry options which will be freed in
 * cleanup_cb(). Also keep .pubkey since it may be needed for a SAVE. */
void
cleanup_crypto_stage1 (struct crypto_s *cr)
{
  if (!cr)
    return;

  cleanup_save (&cr->save);

#ifdef WITH_AGENT
  if (cr->ciphertext_sexp)
    gcry_sexp_release (cr->ciphertext_sexp);

  cr->ciphertext_sexp = NULL;
#endif

  gcry_free (cr->plaintext);
  xfree (cr->ciphertext);
  xfree (cr->filename);
  cr->filename = NULL;
  cr->ciphertext = NULL;
  cr->ciphertext_len = 0;
  cr->plaintext = NULL;
  cr->plaintext_len = 0;
}

void
cleanup_crypto_stage2 (struct crypto_s *cr)
{
  if (!cr)
    return;

  cleanup_crypto_stage1 (cr);
  set_header_defaults (&cr->hdr);
}

void
cleanup_crypto (struct crypto_s **c)
{
  struct crypto_s *cr = *c;

  if (!cr)
    return;

  cleanup_crypto_stage2 (cr);

#ifdef WITH_AGENT
  if (cr->pkey_sexp)
    gcry_sexp_release (cr->pkey_sexp);

  if (cr->sigpkey_sexp)
    gcry_sexp_release (cr->sigpkey_sexp);

  if (cr->agent)
    cleanup_agent (cr->agent);
#endif

  xfree (cr);
  *c = NULL;
}

gpg_error_t
init_client_crypto (struct crypto_s **crypto)
{
  struct crypto_s *new = xcalloc (1, sizeof (struct crypto_s));
  gpg_error_t rc;

  if (!new)
    {
      rc = GPG_ERR_ENOMEM;
      return rc;
    }

#ifdef WITH_AGENT
  if (use_agent)
    {
      rc = agent_init (&new->agent);
      if (!rc)
	{
	  rc = send_agent_common_options (new->agent);
	  if (!rc)
	    rc = agent_set_pinentry_options (new->agent);
	}

      if (rc)
	{
	  cleanup_agent (new->agent);
	  xfree (new);
	  return rc;
	}
    }
#endif

  set_header_defaults (&new->hdr);
  *crypto = new;
  return 0;
}

static gpg_error_t
try_decrypt (assuan_context_t ctx, int inquire, struct crypto_s *crypto,
	     const char *filename, unsigned char **key, size_t *keysize)
{
  gpg_error_t rc;
  int pin_try = 1;
  int pin_tries = 3;
  char *pin_title = NULL;
  struct client_s *client = ctx ? assuan_get_pointer (ctx) : NULL;

  do
    {
      rc = decrypt_common (ctx, inquire, crypto, filename, NULL, NULL, key,
                           keysize);
      if (gpg_err_code (rc) == GPG_ERR_BAD_PASSPHRASE && ctx)
	{
	  if (pin_try == 1)
	    pin_title = client->pinentry_opts.title;

	  client->pinentry_opts.title = str_asprintf (_ ("Bad passphrase (try %i of %i)"), pin_try+1, pin_tries);
	}
    }
  while (!inquire && gpg_err_code (rc) == GPG_ERR_BAD_PASSPHRASE
	 && ++pin_try <= pin_tries);

  if (ctx && pin_title != client->pinentry_opts.title)
    {
      xfree (client->pinentry_opts.title);
      client->pinentry_opts.title = pin_title;
    }

  return rc;
}

/* Common to both PKI and non-PKI files when saving a data
 * file. Defers to gpg-agent when needed and handles keyfiles. */
gpg_error_t
export_common (assuan_context_t ctx, int inquire, struct crypto_s * crypto,
	       const void *data, size_t datalen, const char *outfile,
	       const char *keyfile, void **rkey, size_t *rkeylen,
	       int use_cache, int force, int no_passphrase)
{
  gpg_error_t rc = 0;
  void *enc_xml = NULL;
  size_t enc_xml_len = 0;
  unsigned char *iv = NULL;
  size_t iv_len = 0;
  int algo = cipher_to_gcrypt (crypto->save.hdr.flags);
  unsigned char *key = NULL;
  void *salted_key = NULL;
  size_t keysize, keylen = 0;
  int cached = 0;

  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL, &keysize);
  if (rc)
    return rc;

  if (keyfile)
    {
      int fd;
      unsigned char *buf;
      struct stat st;

      if (stat (keyfile, &st) == -1)
	  return gpg_error_from_errno (errno);

      buf = gcry_malloc (st.st_size);
      if (!buf)
	return GPG_ERR_ENOMEM;

      fd = open (keyfile, O_RDONLY);
      if (fd == -1)
	rc = gpg_error_from_errno (errno);
      else
	{
	  size_t len = read (fd, buf, st.st_size);

	  if (len != st.st_size)
	    rc = GPG_ERR_TOO_SHORT;
	}

      if (fd != -1)
	close (fd);

      key = buf;
      keylen = st.st_size;
      log_write (_("Using passphrase obtained from file '%s'"), keyfile);

      if (!keylen)
	{
	  unsigned char *tmp;

	  gcry_free (key);
	  tmp = gcry_malloc (sizeof (unsigned char));
	  *tmp = 0;
	  key = tmp;
	  keylen++;
	}
    }
  else
    {
      if (use_cache)
	{
	  int defer = -1;

	  cache_lock ();
	  rc = cache_iscached (outfile, &defer);
	  if (!rc)
	    cached = 1;
	  else if (gpg_err_code (rc) == GPG_ERR_ENOENT
		   || gpg_err_code (rc) == GPG_ERR_NO_DATA)
	    rc = 0;

	  if (cached)
	    {
	      struct cache_data_s *cdata = cache_get_data_filename (outfile);

	      salted_key = gcry_malloc (cdata->keylen);
	      if (salted_key)
		{
		  memcpy (salted_key, cdata->key, cdata->keylen);
		  keysize = cdata->keylen;
		}
	      else
		rc = GPG_ERR_ENOMEM;
	    }

	  cache_unlock ();
	}

      if (!rc && !cached)
	{
	  if (force) // SAVE
	    {
	      struct crypto_s *tmp = NULL;
	      gpg_error_t rc = init_client_crypto (&tmp);

	      if (!rc)
		{
		  rc = read_data_file (outfile, tmp);
		  if (!rc)
		    {
		      rc = try_decrypt (ctx, inquire, tmp, outfile,
                                        &key, &keylen);
		    }
		}

	      cleanup_crypto (&tmp);
	      if (rc && gpg_err_code (rc) == GPG_ERR_ENOENT)
		use_cache = 0;
	      else if (rc)
		return rc;
	    }

	  if (!use_cache && !no_passphrase) // PASSWD or new file
	    {
	      gcry_free (key);
	      if (inquire)
		{
		  rc = inquire_passphrase (ctx, "NEW_PASSPHRASE",
					   (unsigned char **)&key, &keylen);
		}
	      else
		{
		  rc = getpin_common (ctx, outfile, PINENTRY_SAVE,
				      (char **)&key, &keylen);
		}

	      if (rc)
		return rc;
	    }
	  else
	    {
	      if (no_passphrase)
		{
		  keylen = 1;
		  key = gcry_malloc (keylen);
		  memset (key, 0, keylen);
		}
	    }
	}
    }

  if (!rc && !cached)
    {
      gcry_create_nonce (crypto->save.hdr.salt, sizeof(crypto->save.hdr.salt));
      rc = hash_key (algo, crypto->save.hdr.salt,
		     sizeof(crypto->save.hdr.salt), key, keylen, &salted_key,
		     &keysize, crypto->save.hdr.s2k_count);
      gcry_free (key);
    }

  if (!rc)
    {
      rc = encrypt_xml (ctx, salted_key, keysize, algo, data, datalen,
			&enc_xml, &enc_xml_len, &iv, &iv_len);
    }

  if (!rc)
    {
      if (no_passphrase && !(crypto->save.hdr.flags & PWMD_FLAG_PKI))
	crypto->save.hdr.flags |= PWMD_FLAG_NO_PASSPHRASE;

      memcpy (crypto->save.hdr.iv, iv, iv_len);
      xfree (iv);
      crypto->save.hdr.datalen = enc_xml_len;
      rc = write_file (crypto, outfile, enc_xml, enc_xml_len, NULL, 0, NULL,
		       NULL);
      gcry_free (enc_xml);
      if (!rc)
	{
	  *rkey = salted_key;
	  *rkeylen = keysize;
	  memcpy (&crypto->hdr, &crypto->save.hdr, sizeof (file_header_t));
	}
    }

  if (rc)
    gcry_free (salted_key);

  return rc;
}

#ifdef WITH_LIBACL
static void
cleanup_acl (void *arg)
{
  acl_t acl = *(acl_t *) arg;

  if (acl)
    acl_free (acl);
}
#endif

static gpg_error_t
write_header (struct crypto_s *crypto, int fd)
{
  ssize_t len;
  uint64_t n, x;
  uint32_t i;

#ifdef WITH_AGENT
  if (IS_PKI (crypto))
    crypto->save.hdr.s2k_count = 0;
#endif

#ifdef WORDS_BIGENDIAN
  len = write (fd, &crypto->save.hdr.magic, sizeof(crypto->save.hdr.magic));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.version, sizeof(crypto->save.hdr.version));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.s2k_count,
	       sizeof(crypto->save.hdr.s2k_count));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.flags, sizeof(crypto->save.hdr.flags));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.iv, sizeof(crypto->save.hdr.iv));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.salt, sizeof(crypto->save.hdr.salt));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.datalen, sizeof(crypto->save.hdr.datalen));
  if (len == -1)
    goto done;
#else
  len = write (fd, &crypto->save.hdr.magic, sizeof(crypto->save.hdr.magic));
  if (len == -1)
    goto done;

  i = htonl (crypto->save.hdr.version);
  len = write (fd, &i, sizeof(i));
  if (len == -1)
    goto done;

  n = crypto->save.hdr.s2k_count;
  x = (((uint64_t) htonl (n)) << 32) + htonl (n >> 32);
  len = write (fd, &x, sizeof(x));
  if (len == -1)
    goto done;

  n = crypto->save.hdr.flags;
  x = (((uint64_t) htonl (n)) << 32) + htonl (n >> 32);
  len = write (fd, &x, sizeof(x));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.iv, sizeof(crypto->save.hdr.iv));
  if (len == -1)
    goto done;

  len = write (fd, &crypto->save.hdr.salt, sizeof(crypto->save.hdr.salt));
  if (len == -1)
    goto done;

  i = htonl (crypto->save.hdr.datalen);
  len = write (fd, &i, sizeof(i));
  if (len == -1)
    goto done;
#endif

 done:
  return len == -1 ? gpg_error_from_errno (errno) : 0;
}

/* Write the data structures to disk. The advisory lock should be obtained
 * before calling this function. */
gpg_error_t
write_file (struct crypto_s *crypto, const char *filename,
	    void *data, size_t data_len, void *sexp, size_t sexp_len,
	    void *pubkey, void *sigpkey)
{
  char tmp[FILENAME_MAX] = { 0 };
  mode_t mode = 0;
  struct stat st;
  int fd;
  gpg_error_t rc = 0;
#ifdef WITH_LIBACL
  acl_t acl = NULL;
#endif

  if (filename)
    {
      mode_t mask;

      if (lstat (filename, &st) == 0)
	{
	  mode = st.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO);

	  if (!(mode & S_IWUSR))
	    return GPG_ERR_EACCES;
	}
      else if (errno != ENOENT)
	return gpg_error_from_errno (errno);

      snprintf (tmp, sizeof (tmp), "%s.XXXXXX", filename);
      mask = umask (600);
      fd = mkstemp (tmp);
      if (fd == -1)
	{
	  rc = gpg_error_from_errno (errno);
	  log_write ("%s: %s", tmp, pwmd_strerror (rc));
          umask (mask);
	  return rc;
	}

      umask (mask);
    }
  else
    {
      // xml_import() or convert_file() from command line.
      fd = STDOUT_FILENO;
    }

  pthread_cleanup_push (cleanup_unlink_cb, tmp);
  crypto->save.hdr.version = VERSION_HEX;
  rc = write_header (crypto, fd);
  if (!rc)
    {
      size_t len;

      if (crypto->save.hdr.flags & PWMD_FLAG_PKI)
	{
	  unsigned char grip[20];

	  gcry_pk_get_keygrip ((gcry_sexp_t) pubkey, grip);
	  len = write (fd, grip, sizeof (grip));
	  if (len == sizeof (grip))
	    {
	      gcry_pk_get_keygrip ((gcry_sexp_t) sigpkey, grip);
	      len = write (fd, grip, sizeof (grip));
	      if (len == sizeof (grip))
		{
		  len = write (fd, data, data_len);
		  if (len == data_len)
		    {
		      len = write (fd, sexp, sexp_len);
		      if (len != sexp_len)
			rc = gpg_error_from_errno (errno);
		    }
		  else
		    rc = gpg_error_from_errno (errno);
		}
	    }
	  else
	    rc = gpg_error_from_errno (errno);
	}
      else
	{
	  len = write (fd, data, data_len);
	  if (len != data_len)
	    rc = gpg_error_from_errno (errno);
	}
    }

  if (!rc)
    {
      if (fsync (fd) != -1)
	{
	  if (filename && close (fd) != -1)
	    {
#ifdef WITH_LIBACL
	      acl = acl_get_file (filename, ACL_TYPE_ACCESS);
	      if (!acl && errno == ENOENT)
		acl = acl_get_file (".", ACL_TYPE_DEFAULT);
	      if (!acl)
		log_write ("ACL: %s: %s", filename,
			   pwmd_strerror (gpg_error_from_errno (errno)));
              pthread_cleanup_push (cleanup_acl, &acl);
#endif

	      if (mode && config_get_boolean (filename, "backup"))
		{
		  char tmp2[FILENAME_MAX];

		  snprintf (tmp2, sizeof (tmp2), "%s.backup", filename);
		  if (rename (filename, tmp2) == -1)
		    rc = gpg_error_from_errno (errno);
		}

              if (rename (tmp, filename) != -1)
                {
                  if (mode)
                    chmod (filename, mode);

#ifdef WITH_LIBACL
                  if (acl && acl_set_file (filename, ACL_TYPE_ACCESS, acl))
                    log_write ("ACL: %s: %s", filename,
                               pwmd_strerror (gpg_error_from_errno (errno)));
#endif
                }
              else
                rc = gpg_error_from_errno (errno);

#ifdef WITH_LIBACL
              pthread_cleanup_pop (1);
#endif
            }
          else if (filename)
	    rc = gpg_error_from_errno (errno);

          if (!rc && fd != -1)
            {
              char *tmp = str_asprintf ("%s/data", homedir);

              if (tmp)
                {
                  int dfd = open (tmp, O_RDONLY);

                  if (dfd != -1)
                    {
                      if (fsync (dfd) == -1)
                        log_write ("%s %d: %s", __FILE__, __LINE__,
                                   pwmd_strerror (errno));

                      close (dfd);
                    }
                  else
                    log_write ("%s %d: %s", __FILE__, __LINE__,
                               pwmd_strerror (errno));

                  xfree (tmp);
                }
            }
	}
      else
	rc = gpg_error_from_errno (errno);
    }

  pthread_cleanup_pop (rc ? 1 : 0);	// unlink
  return rc;
}

gpg_error_t
hash_key (int algo, unsigned char *salt, size_t salt_len, const void *key,
	  size_t keylen, void **result, size_t *rlen, uint64_t iterations)
{
  gpg_error_t rc;
  void *tmp;

  /* Be sure the algorithm is available. */
  rc = gcry_cipher_algo_info (algo, GCRYCTL_GET_KEYLEN, NULL, rlen);
  if (rc)
    return rc;

  /* Always allocate enough for a 256-bit key although the algorithms
     themselves may use less. Fixes SAVE --cipher with a different
     keylen than the previously saved cipher when cached. */
  *rlen = 32;
  tmp = xmalloc (*rlen);
  if (!tmp)
    return GPG_ERR_ENOMEM;

  if (!iterations)
    iterations = DEFAULT_KDFS2K_ITERATIONS;

  rc = gcry_kdf_derive(key, keylen, GCRY_KDF_ITERSALTED_S2K, GCRY_MD_SHA1,
		       salt, salt_len, iterations, *rlen, tmp);
  if (!rc)
    *result = tmp;
  else
    xfree (tmp);

  return rc;
}

gpg_error_t
crypto_try_decrypt (assuan_context_t ctx, int inquire, const char *filename,
                    struct crypto_s **rcrypto, unsigned char **key,
                    size_t *rlen)
{
  struct crypto_s *crypto = rcrypto ? *rcrypto : NULL;
  gpg_error_t rc = init_client_crypto (&crypto);

  if (rc)
    return rc;

  crypto->client_ctx = ctx;
  rc = read_data_file (filename, crypto);
  if (rc)
    {
      cleanup_crypto (&crypto);
      return rc;
    }

  rc = try_decrypt (ctx, inquire, crypto, filename, key, rlen);
  if (!rcrypto)
    cleanup_crypto (&crypto);
  else *rcrypto = crypto;

  return rc;
}

/* The PASSWD command when not using gpg-agent. */
gpg_error_t
change_passwd (assuan_context_t ctx, const char *filename, int inquire,
	       struct crypto_s **rcrypto, int no_passphrase)
{
  gpg_error_t rc;
  unsigned char *key = NULL;
  size_t keylen = 0;
  struct crypto_s *crypto = NULL;
  int cached;

  rc = crypto_try_decrypt (ctx, inquire, filename, &crypto, NULL, NULL);
  if (rc)
    {
      cleanup_crypto (&crypto);
      return rc;
    }

  if (!rc)
    {
      memcpy (&crypto->save.hdr, &crypto->hdr, sizeof (file_header_t));

      if (!no_passphrase)
	crypto->save.hdr.flags &= ~PWMD_FLAG_NO_PASSPHRASE;

      rc = export_common (ctx, inquire, crypto, crypto->plaintext,
			  crypto->plaintext_len, crypto->filename, NULL,
			  (void **)&key, &keylen, 0, 0, no_passphrase);
    }

  if (!rc)
    {
      rc = save_common (filename, crypto, crypto->plaintext,
			crypto->plaintext_len, key, keylen, &cached, 1);
      if (!rc)
	*rcrypto = crypto;
    }

  gcry_free (key);

  if (rc)
    cleanup_crypto (&crypto);
  else if (!cached)
    send_status_all (STATUS_CACHE, NULL);

  return rc;
}

gpg_error_t
save_common (const char *filename, struct crypto_s *crypto,
	     const unsigned char *data, size_t datalen,
	     const unsigned char *key, size_t keylen, int *cached, int no_agent)
{
  struct cache_data_s *cdata;
  gpg_error_t rc = 0;
  gcry_sexp_t pubkey = NULL;
  gcry_sexp_t sigkey = NULL;

  /* This is safe since it is a (somewhat) fast operation. */
  cache_lock ();
  pthread_cleanup_push (cleanup_cache_mutex, NULL);
  cdata = cache_get_data_filename (filename);
  *cached = cdata != NULL;

  if (!cdata)
    cdata = xcalloc (1, sizeof (struct cache_data_s));

#ifdef WITH_AGENT
  if (use_agent && !no_agent)
    {
      rc = gcry_sexp_build (&pubkey, NULL, "%S", crypto->pkey_sexp);
      if (!rc)
	rc = gcry_sexp_build (&sigkey, NULL, "%S", crypto->sigpkey_sexp);
    }
#endif

  if (!rc)
    {
      gcry_free (cdata->doc);
      cdata->doc = NULL;
      cdata->doclen = 0;
      rc = encrypt_xml (NULL, cache_key, cache_keysize, GCRY_CIPHER_AES, data,
			datalen, &cdata->doc, &cdata->doclen, &cache_iv,
			&cache_blocksize);
    }

  if (!rc)
    {
      unsigned char md5file[16];

      /* Update in case of any --keygrip argument */
      if (cdata->pubkey)
	gcry_sexp_release (cdata->pubkey);

      if (cdata->sigkey)
	gcry_sexp_release (cdata->sigkey);

      cdata->pubkey = pubkey;
      cdata->sigkey = sigkey;
      xfree (cdata->key);
      cdata->key = NULL;
      gcry_md_hash_buffer (GCRY_MD_MD5, md5file, filename, strlen (filename));
      rc = cache_set_data (md5file, cdata, crypto->grip);
      if (!rc)
        {
          cdata->key = xmalloc (keylen);
          if (!cdata->key)
            rc = GPG_ERR_ENOMEM;
          else
            {
              memcpy (cdata->key, key, keylen);
              cdata->keylen = keylen;
            }
        }
    }
  else
    {
      if (pubkey)
	gcry_sexp_release (pubkey);

      if (sigkey)
	gcry_sexp_release (sigkey);
    }

  pthread_cleanup_pop (1);	// mutex unlock
  return rc;
}

gpg_error_t
getpin_common (assuan_context_t ctx, const char *filename, int which,
	       char **rkey, size_t *rkeylen)
{
  gpg_error_t rc;
  struct client_s *client = ctx ? assuan_get_pointer (ctx) : NULL;
  struct pinentry_s *pin = pinentry_init (filename);

  if (!pin)
    return GPG_ERR_ENOMEM;

  if (client)
    pinentry_merge_options (&pin->pinentry_opts, &client->pinentry_opts);

  rc = pinentry_lock (ctx, pin);
  if (!rc)
    rc = pinentry_getpin (pin, rkey, which);

  pinentry_deinit (pin);
  if (rc)
    return rc;

  *rkeylen = strlen (*rkey);
  if (!(*rkeylen))
    (*rkeylen)++;

  return rc;
}

/* Note: 'key' is freed with gcry_free() and not xfree(). Although both call
 * xfree(). The advisory lock should be obtained before calling this function.
 */
gpg_error_t
inquire_passphrase (assuan_context_t ctx, const char *keyword,
		    unsigned char **result, size_t *rlen)
{
  gpg_error_t rc;

  assuan_begin_confidential (ctx);
  rc = assuan_inquire (ctx, keyword, result, rlen, 0);
  assuan_end_confidential (ctx);
  return rc;
}

/* Common to both PKI and non-PKI files. */
gpg_error_t
decrypt_common (assuan_context_t ctx, int inquire, struct crypto_s *crypto,
		const char *filename, void **salted_key, size_t *salted_keysize,
                unsigned char **rkey, size_t *rkeylen)
{
  unsigned char *key = NULL;
  size_t keylen = 0;
  gpg_error_t rc = read_data_file (filename, crypto);
  int algo = cipher_to_gcrypt (crypto->hdr.flags);
  void *skey = NULL;
  size_t skeysize = 0;

  if (salted_key)
    *salted_key = NULL;
  if (salted_keysize)
    *salted_keysize = 0;

  if (rkey)
    *rkey = NULL;
  if (rkeylen)
    *rkeylen = 0;

  if (rc)
    {
      log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
      return rc;
    }

  key = (unsigned char *)config_get_string (filename, "passphrase");
  if (key)
    {
      log_write (_("Trying the passphrase specified in config ..."));
      keylen = strlen ((char *)key);
    }

  if (!key)
    {
      key = (unsigned char *)config_get_string (filename, "passphrase_file");
      if (key)
	{
	  char *tmp = expand_homedir ((char *)key);
	  int fd;
	  struct stat st;

	  xfree (key);
	  key = (unsigned char *)tmp;
	  log_write (_("Trying the passphrase using file '%s' ..."), key);
	  fd = open ((char *)key, O_RDONLY);
	  if (fd == -1)
	    {
	      rc = gpg_error_from_errno (errno);
	      log_write ("%s: %s", key, pwmd_strerror (rc));
	      xfree (key);
	      return rc;
	    }

	  if (stat ((char *)key, &st) == -1)
            {
	      rc = gpg_error_from_errno (errno);
	      log_write ("%s: %s", key, pwmd_strerror (rc));
	      xfree (key);
              close(fd);
	      return rc;
            }

	  xfree (key);
	  key = xmalloc (st.st_size);
	  if (key)
	    {
	      if (read (fd, key, st.st_size) != st.st_size)
		{
		  log_write ("short read() count");
		  rc = GPG_ERR_TOO_SHORT;
		}
	    }
	  else
	    rc = GPG_ERR_ENOMEM;

	  close (fd);

	  if (rc)
	    {
	      log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
	      xfree (key);
	      return rc;
	    }

	  keylen = st.st_size;
	  if (!keylen)
	    {
	      tmp = xmalloc (sizeof (unsigned char));
	      *tmp = 0;
	      xfree (key);
	      key = (unsigned char *)tmp;
	      keylen++;
	    }
	}
    }

  if (!key && !IS_PKI (crypto) && inquire
      && !(crypto->hdr.flags & PWMD_FLAG_NO_PASSPHRASE))
    {
      rc = inquire_passphrase (ctx, "PASSPHRASE", (unsigned char **)&key,
			       &keylen);
      if (!rc)
        if (!keylen)
          keylen++;
    }
  else if (!key && !IS_PKI (crypto))
    {
      if (crypto->hdr.flags & PWMD_FLAG_NO_PASSPHRASE)
	{
	  keylen = 1;
	  key = gcry_malloc (keylen);
	  memset (key, 0, keylen);
	}
      else
        rc = getpin_common (ctx, filename, PINENTRY_OPEN, (char **)&key,
                            &keylen);
    }
  else if (key && IS_PKI (crypto))
    {
#ifdef WITH_AGENT
      rc = set_agent_passphrase (crypto, (char *)key, keylen);
#else
      rc = GPG_ERR_NOT_IMPLEMENTED;
#endif
    }
  else if (!key && IS_PKI (crypto))
    {
#ifdef WITH_AGENT
      /* The cache entry has been cleared in save_command(). Will prompt. */
      rc = set_pinentry_mode (crypto->agent, inquire ? "loopback" : "ask");
#else
      rc = GPG_ERR_NOT_IMPLEMENTED;
#endif
    }

  if (rc)
    {
      xfree (key);
      log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
      return rc;
    }

  if (key && !IS_PKI (crypto))
    {
      rc = hash_key (algo, crypto->hdr.salt, sizeof(crypto->hdr.salt), key,
		     keylen, &skey, &skeysize,
                     crypto->hdr.version <= 0x03000e ? COMPAT_KDFS2K_ITERATIONS : crypto->hdr.s2k_count);
      if (rc)
	{
	  xfree (key);
          log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
	  return rc;
	}

      if (rkey)
        *rkey = key;
      if (rkeylen)
        *rkeylen = keylen;
    }

  xfree (crypto->filename);
  crypto->filename = str_dup (filename);
  rc = decrypt_data (ctx, crypto, skey, skeysize);
  if (rc)
    {
      log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
      xfree (key);
      xfree (skey);
    }
  else
    {
      if (salted_key)
        *salted_key = skey;
      else
        xfree (skey);

      if (salted_keysize)
        *salted_keysize = skeysize;
    }

  return rc;
}
