/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CRYPTO_H
#define CRYPTO_H

#include <assuan.h>
#include <gpg-error.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gcrypt.h>

#ifdef HAVE_STDINT_H
#include <stdint.h>
#elif defined (HAVE_INTTYPES_H)
#include <inttypes.h>
#endif

#ifdef WITH_AGENT
#include "agent.h"
#endif

#include "cipher.h"

#define DEFAULT_KDFS2K_ITERATIONS 5000000
#define COMPAT_KDFS2K_ITERATIONS 1000

typedef struct
{
  uint8_t magic[5];
  uint32_t version;
  uint64_t s2k_count;
  uint64_t flags;
  uint8_t iv[16];
  uint8_t salt[8];
  uint32_t datalen;		/* of the encrypted xml */
} __attribute__ ((packed)) file_header_t;

struct save_s
{
  gcry_sexp_t pkey;		/* SAVE --keygrip */
  gcry_sexp_t sigpkey;		/* SAVE --sign-keygrip */
  file_header_t hdr;
};

struct crypto_s
{
  assuan_context_t client_ctx;
#ifdef WITH_AGENT
  struct agent_s *agent;
#endif
  struct save_s save;
  gcry_sexp_t pkey_sexp;
  unsigned char grip[20];
  gcry_sexp_t sigpkey_sexp;
  unsigned char sign_grip[20];
  gcry_sexp_t ciphertext_sexp;
  void *ciphertext;
  size_t ciphertext_len;
  void *plaintext;
  size_t plaintext_len;
  file_header_t hdr;
  char *filename;		/* the currently opened data file */
};

int use_agent;
#define IS_PKI(crypto) (use_agent && (crypto->hdr.flags & PWMD_FLAG_PKI))

void cleanup_save (struct save_s *save);
gpg_error_t encrypt_xml (assuan_context_t ctx, void *key, size_t keylen, int
			 algo, const void *xml, size_t len, void **result,
			 size_t * result_len, unsigned char **iv,
			 size_t * iv_len);
gpg_error_t decrypt_cache (struct crypto_s *crypto, const void *data,
			   size_t len);
gpg_error_t read_data_file (const char *filename, struct crypto_s *crypto);
gpg_error_t read_data_header (const char *filename, file_header_t * fh,
			      struct stat *st, int *fd);
gpg_error_t decrypt_data (assuan_context_t ctx, struct crypto_s *crypto,
			  unsigned char *salted_key, size_t keylen);
void cleanup_crypto_stage1 (struct crypto_s *cr);
void cleanup_crypto_stage2 (struct crypto_s *cr);
void cleanup_crypto (struct crypto_s **c);
gpg_error_t init_client_crypto (struct crypto_s **crypto);
gpg_error_t write_file (struct crypto_s *crypto, const char *filename,
			void *data, size_t data_len, void *sexp,
			size_t sexp_len, void *pubkey, void *sigpkey);
gpg_error_t export_common (assuan_context_t ctx, int inquire,
			   struct crypto_s * crypto, const void *data,
			   size_t datalen, const char *outfile,
			   const char *keyfile, void **rkey, size_t *rkeylen,
			   int use_cache, int force, int no_passphrase);
gpg_error_t decrypt_common (assuan_context_t ctx, int inquire,
			    struct crypto_s *crypto, const char *filename,
                            void **salted_key, size_t *salted_keysize,
			    unsigned char **rkey, size_t *rkeylen);
gpg_error_t getpin_common (assuan_context_t ctx, const char *filename,
			   int which, char **rkey, size_t *rkeylen);
gpg_error_t save_common (const char *md5file, struct crypto_s *crypto,
			 const unsigned char *data, size_t datalen,
			 const unsigned char *key, size_t keylen, int *cached,
			 int no_agent);
gpg_error_t change_passwd (assuan_context_t ctx, const char *filename,
			   int inquire, struct crypto_s **rcrypto,
			   int no_passphrase);
gpg_error_t inquire_passphrase (assuan_context_t ctx, const char *keyword,
				unsigned char **result, size_t *rlen);
gpg_error_t hash_key (int algo, unsigned char *salt, size_t salt_len,
		      const void *key, size_t keylen, void **result,
		      size_t *rlen, uint64_t iterations);
gpg_error_t crypto_try_decrypt (assuan_context_t ctx, int inquire, const char *filename,
                                struct crypto_s **rcrypto, unsigned char **key,
                                size_t *rlen);

#endif
