/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <err.h>
#include <pthread.h>

#include "memzero.h"

#ifndef _
#include "gettext.h"
#define _(msgid) gettext(msgid)
#endif

#ifdef DEBUG
#ifdef HAVE_BACKTRACE
#include <execinfo.h>

#ifndef BACKTRACE
#define BACKTRACE(fn)	{ \
			    int n, nptrs; \
			    char **strings; \
			    void *buffer[20]; \
			    nptrs = backtrace(buffer, 20); \
			    strings = backtrace_symbols(buffer, nptrs); \
			    for (n = 0; n < nptrs; n++) \
				fprintf(stderr, "BACKTRACE (%s) %i: %s\n", fn, n, strings[n]); \
			    fprintf(stderr, "\n"); \
			    fflush(stderr); \
	                }
#endif
#endif
#else
#define BACKTRACE(fn)
#endif

#include "mem.h"

#ifndef _
#define _ gettext(msgid)
#endif

struct memlist_s
{
  void *ptr;
  size_t size;
  struct memlist_s *next;
};

static struct memlist_s *memlist;
static pthread_mutex_t mem_mutex;
#ifdef DEBUG
static size_t allocations, deallocations;
#endif

void
xmem_init ()
{
  static int init;

  if (!init)
    pthread_mutex_init (&mem_mutex, NULL);

  init = 1;
}

static int
memlist_remove (void *ptr, const char *func)
{
  struct memlist_s *m, *last = NULL, *p;

  pthread_mutex_lock (&mem_mutex);

  for (m = memlist; m; m = m->next)
    {
      if (m->ptr == ptr)
	{
#ifdef DEBUG
	  fprintf (stderr, "%s: %p %zu\n", func, ptr, m->size);
#endif
	  memzero (m->ptr, m->size);
	  free (m->ptr);

	  p = m->next;
	  free (m);
#ifdef DEBUG
	  deallocations++;
#endif

	  if (last)
	    last->next = p;
	  else
	    memlist = p;

	  pthread_mutex_unlock (&mem_mutex);
	  return 1;
	}

      last = m;
    }

  pthread_mutex_unlock (&mem_mutex);
  return 0;
}

static void
memlist_prepend (struct memlist_s *new)
{
  pthread_mutex_lock (&mem_mutex);
#ifdef DEBUG
  allocations++;
#endif
  new->next = memlist;
  memlist = new;
  pthread_mutex_unlock (&mem_mutex);
}

void
xfree (void *ptr)
{
  if (!ptr)
    return;

  if (!memlist_remove (ptr, __FUNCTION__))
    {
      warnx (_("%s: %p not found"), __FUNCTION__, ptr);
      assert (0);
    }
}

void *
xmalloc (size_t size)
{
  void *p;
  struct memlist_s *m;

  if ((m = (struct memlist_s *) malloc (sizeof (struct memlist_s))) == NULL)
    return NULL;

  if ((p = (void *) malloc (size)) == NULL)
    {
      free (m);
      return NULL;
    }

  m->ptr = p;
  m->size = size;
  memlist_prepend (m);
#ifdef DEBUG
  fprintf (stderr, "%s: %p %zu\n", __FUNCTION__, p, size);
  BACKTRACE (__FUNCTION__);
#endif
  return m->ptr;
}

void *
xcalloc (size_t nmemb, size_t size)
{
  void *p;
  struct memlist_s *m;

  if ((m = (struct memlist_s *) malloc (sizeof (struct memlist_s))) == NULL)
    return NULL;

  if ((p = calloc (nmemb, size)) == NULL)
    {
      free (m);
      return NULL;
    }

  m->ptr = p;
  m->size = nmemb * size;
  memlist_prepend (m);
#ifdef DEBUG
  fprintf (stderr, "%s: %p %i\n", __FUNCTION__, p, nmemb * size);
  BACKTRACE (__FUNCTION__);
#endif
  return m->ptr;
}

void *
xrealloc (void *ptr, size_t size)
{
  void *p;
  struct memlist_s *m;

  if (!size && ptr)
    {
      xfree (ptr);
      return NULL;
    }

  if (!ptr)
    return xmalloc (size);

  pthread_mutex_lock (&mem_mutex);

  for (m = memlist; m; m = m->next)
    {
      if (m->ptr == ptr)
	{
	  if ((p = (void *) malloc (size)) == NULL)
	    {
	      pthread_mutex_unlock (&mem_mutex);
	      return NULL;
	    }

	  memcpy (p, m->ptr, size < m->size ? size : m->size);
	  memzero (m->ptr, m->size);
	  free (m->ptr);
	  m->ptr = p;
	  m->size = size;
#ifdef DEBUG
	  fprintf (stderr, "%s: %p %zu\n", __FUNCTION__, p, size);
	  BACKTRACE (__FUNCTION__);
#endif
	  pthread_mutex_unlock (&mem_mutex);
	  return m->ptr;
	}
    }

  warnx (_("%s: %p not found"), __FUNCTION__, ptr);
  pthread_mutex_unlock (&mem_mutex);
  assert (0);
  return NULL;
}

void
xpanic (void)
{
  struct memlist_s *m;

  pthread_mutex_lock (&mem_mutex);

  for (m = memlist; m; m = memlist)
    xfree (m->ptr);

  pthread_mutex_unlock (&mem_mutex);
}

#ifdef DEBUG
void
xdump (void)
{
  struct memlist_s *m;
  size_t total = 0;

  pthread_mutex_lock (&mem_mutex);

  for (m = memlist; m; m = m->next)
    {
      fprintf (stderr, "%s: %p %zu\n", __FUNCTION__, m->ptr, m->size);
      total += m->size;
    }

  fprintf (stderr,
	   "Total unfreed: %zu bytes, allocations: %zu, deallocations: %zu\n",
	   total, allocations, deallocations);
  pthread_mutex_unlock (&mem_mutex);
}
#endif
