/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>

#include "pwmd-error.h"
#include <gcrypt.h>
#include "mem.h"
#include "util-misc.h"
#include "common.h"
#include "commands.h"
#include "pinentry.h"
#include "mutex.h"
#include "rcfile.h"
#include "util-string.h"

#ifdef WITH_QUALITY
#include <crack.h>
#endif

void cleanup_all_clients (int atfork);
static gpg_error_t set_pinentry_strings (struct pinentry_s *pin, int which);

static gpg_error_t
mem_realloc_cb (void *data, const void *buffer, size_t len)
{
  membuf_t *mem = (membuf_t *) data;
  void *p;

  if (!buffer)
    return 0;

  p = xrealloc (mem->buf, mem->len + len);
  if (!p)
    return 1;

  mem->buf = p;
  memcpy ((char *) mem->buf + mem->len, buffer, len);
  mem->len += len;
  return 0;
}

#ifdef WITH_QUALITY
static gpg_error_t
quality_cb (void *data, const char *line)
{
  struct pinentry_s *pin = data;
  const char *tmp;
  int score = 0;
  char buf[5];

  if (strncmp (line, "QUALITY ", 8) != 0)
    return GPG_ERR_INV_ARG;

  tmp = FascistCheck (line + 8, CRACKLIB_DICT);
  if (!tmp)
    return assuan_send_data (pin->ctx, "100", 3);

  if (!strcmp (tmp, _("it's WAY too short")))
    score = 10;
  else if (!strcmp (tmp, _("it is too short")))
    score = 20;
  else if (!strcmp (tmp, _("it is all whitespace")))
    score = 25;
  else if (!strcmp (tmp, _("it is based on your username")))
    score = 30;
  else if (!strcmp (tmp, _("it is based on a dictionary word")))
    score = 40;
  else if (!strcmp (tmp, _("it is based upon your password entry")))
    score = 50;
  else if (!strcmp (tmp, _("it's derived from your password entry")))
    score = 50;
  else if (!strcmp (tmp, _("it is based on a (reversed) dictionary word")))
    score = 60;
  else if (!strcmp (tmp, _("it is derivable from your password entry")))
    score = 70;
  else
    if (!strcmp
	(tmp, _("it does not contain enough DIFFERENT characters")))
      score = 80;
    else if (!strcmp (tmp, _("it is too simplistic/systematic")))
      score = 90;
    else
      score = 0;

  snprintf (buf, sizeof (buf), "%i", score);
  tmp = buf;
  return assuan_send_data (pin->ctx, tmp, strlen (tmp));
}
#endif

static gpg_error_t
assuan_command (struct pinentry_s *pin, char **result, const char *cmd)
{
  gpg_error_t rc;

  pin->data.len = 0;
  pin->data.buf = NULL;

  rc = assuan_transact (pin->ctx, cmd, mem_realloc_cb, &pin->data,
			pin->inquire_cb, pin->inquire_data, NULL, NULL);
  if (rc)
    {
      if (pin->data.buf)
	{
	  xfree (pin->data.buf);
	  pin->data.buf = NULL;
	}
    }
  else
    {
      if (pin->data.buf)
	{
	  mem_realloc_cb (&pin->data, "", 1);
	  *result = (char *) pin->data.buf;
	}
    }

  return rc;
}

static gpg_error_t
set_pinentry_options (struct pinentry_s *pin)
{
  char *display = getenv ("DISPLAY");
  int have_display = 0;
  char *tty = NULL, *ttytype = NULL;
  char *opt, *val;
  gpg_error_t rc;
  char *result = NULL;
  char cmd[ASSUAN_LINELENGTH];

  if (pin->pinentry_opts.display || display)
    have_display = 1;
  else
    {
      tty = pin->pinentry_opts.ttyname ? pin->pinentry_opts.ttyname
	: ttyname (STDOUT_FILENO);
      if (!tty)
	return gpg_error_from_errno (errno);
    }

  if (!have_display && !tty)
    return GPG_ERR_CANCELED;

  if (!have_display)
    {
      char *p = getenv ("TERM");

      ttytype = pin->pinentry_opts.ttytype ? pin->pinentry_opts.ttytype : p;
      if (!ttytype)
	return GPG_ERR_ENOTTY;
    }

  opt = have_display ? str_dup ("DISPLAY") : str_dup ("TTYNAME");
  val = have_display ? pin->pinentry_opts.display ? pin->pinentry_opts.display
    : display : tty;
  snprintf (cmd, sizeof (cmd), "OPTION %s=%s", str_down (opt), val);
  xfree (opt);
  rc = assuan_command (pin, &result, cmd);
  if (rc)
    return rc;

  if (!have_display)
    {
      snprintf (cmd, sizeof (cmd), "OPTION ttytype=%s", ttytype);
      rc = assuan_command (pin, &result, cmd);
    }

  if (!rc)
    {
      snprintf (cmd, sizeof(cmd), "SETTIMEOUT %i", pin->pinentry_opts.timeout);
      rc = assuan_command (pin, NULL, cmd);
      /* Not all pinentrys support the SETTIMEOUT command. */
      if (rc && gpg_err_source (rc) == GPG_ERR_SOURCE_PINENTRY
	  && gpg_err_code (rc) == GPG_ERR_ASS_UNKNOWN_CMD)
	rc = 0;
    }

  return rc;
}

#ifdef HAVE_PTHREAD_ATFORK
static void
atfork_cb ()
{
  cleanup_all_clients (1);
}
#else
static void
atfork_cb (void *data, int n)
{
  (void)data;
  if (!n)
    cleanup_all_clients (1);
}
#endif

static gpg_error_t
launch_pinentry (struct pinentry_s *pin)
{
  gpg_error_t rc;
  assuan_context_t ctx;
  static struct assuan_malloc_hooks mhooks = { xmalloc, xrealloc, xfree };
  int child_list[] = { -1 };
  const char *argv[8];
  const char **p = argv;
  char *tmp;

  *p++ = "pinentry";

  if (pin->pinentry_opts.display)
    {
      *p++ = "--display";
      *p++ = pin->pinentry_opts.display;
    }

  if (pin->pinentry_opts.lc_ctype)
    {
      *p++ = "--lc-ctype";
      *p++ = pin->pinentry_opts.lc_ctype;
    }

  if (pin->pinentry_opts.lc_messages)
    {
      *p++ = "--lc-messages";
      *p++ = pin->pinentry_opts.lc_messages;
    }

  *p = NULL;
  rc = assuan_new_ext (&ctx, GPG_ERR_SOURCE_DEFAULT, &mhooks, NULL, NULL);
  if (rc)
    return rc;

  tmp = config_get_string ("global", "pinentry_path");
#ifdef HAVE_PTHREAD_ATFORK
  pthread_atfork (NULL, NULL, atfork_cb);
  rc = assuan_pipe_connect (ctx, tmp, argv, child_list, NULL, NULL, 0);
#else
  rc = assuan_pipe_connect (ctx, tmp, argv, child_list, atfork_cb, NULL, 0);
#endif
  xfree (tmp);
  if (rc)
    return rc;

  pin->ctx = ctx;
  rc = set_pinentry_options (pin);
  return rc ? rc : set_pinentry_strings (pin, pin->which);
}

static gpg_error_t
pinentry_command (struct pinentry_s *pin, char **result, const char *cmd)
{
  gpg_error_t rc = 0;

  if (!pin->ctx)
    rc = launch_pinentry (pin);

  return rc ? rc : assuan_command (pin, result, cmd);
}

static gpg_error_t
set_pinentry_strings (struct pinentry_s *pin, int which)
{
  char *buf;
  gpg_error_t rc;
  char *title = NULL;

  /* The initial prompt from launch_pinentry(). */
  pin->which = which;

  if (which == PINENTRY_FAIL_CONFIRM)
    {
      title = str_dup (_("Passphrase mismatch, please try again."));
      which = PINENTRY_SAVE;
    }
  else if (!pin->pinentry_opts.title)
    title = str_asprintf (_("Password Manager Daemon%s%s"),
			  pin->name ? ": " : "",
			  pin->name ? pin->name : "");
  else
    title = str_dup (pin->pinentry_opts.title);

#ifdef WITH_QUALITY
  if (which == PINENTRY_SAVE)
    {
      rc = pinentry_command (pin, NULL, "SETQUALITYBAR");
      if (rc)
	goto done;

      pin->inquire_cb = quality_cb;
      pin->inquire_data = pin;
    }
#endif

  if (!pin->pinentry_opts.prompt)
    pin->pinentry_opts.prompt = str_dup (_("Passphrase:"));

  if (!pin->pinentry_opts.desc)
    pin->pinentry_opts.desc = str_asprintf (which == PINENTRY_OPEN ?
			      _("A passphrase is required to open the file "
				"\"%s\". Please enter the passphrase below.")
			      :
			      _("A passphrase is required to save to the file "
				"\"%s\". Please enter the passphrase below."),
			      pin->filename);

  if (which == PINENTRY_SAVE_CONFIRM)
    buf = str_asprintf ("SETDESC %s",
			_("Please enter the passphrase again for confirmation."));
  else
    buf = str_asprintf ("SETDESC %s", pin->pinentry_opts.desc);

  rc = pinentry_command (pin, NULL, buf);
  xfree (buf);
  if (rc)
    goto done;

  buf = str_asprintf ("SETPROMPT %s", pin->pinentry_opts.prompt);
  rc = pinentry_command (pin, NULL, buf);
  xfree (buf);
  if (rc)
    goto done;

  buf = str_asprintf ("SETERROR %s", title);
  rc = pinentry_command (pin, NULL, buf);
  xfree (buf);

 done:
  free (title);
  return rc;
}

static void
pinentry_disconnect (struct pinentry_s *pin)
{
  if (!pin)
    return;

  if (pin->ctx)
    assuan_release (pin->ctx);

  pin->ctx = NULL;
}

static gpg_error_t
do_getpin (struct pinentry_s *pin, char **result)
{
  gpg_error_t rc;

  *result = NULL;
  rc = pinentry_command (pin, result, "GETPIN");
  if (rc)
    return rc;

  if (!*result)
    *result = str_dup ("");

  return rc;
}

gpg_error_t
pinentry_getpin (struct pinentry_s * pin, char **result, pinentry_cmd_t which)
{
  gpg_error_t rc = set_pinentry_strings (pin, which);
  char *result1 = NULL;

  if (rc)
    goto done;

 again:
  rc = do_getpin (pin, result);
  if (rc)
    goto done;

  if (which == PINENTRY_SAVE)
    {
      if (!result1)
	{
	  rc = set_pinentry_strings (pin, PINENTRY_SAVE_CONFIRM);
	  if (rc)
	    goto done;

	  result1 = *result;
	  goto again;
	}

      if (strcmp (result1, *result))
	{
	  xfree (result1);
	  xfree (*result);
	  result1 = *result = NULL;
	  rc = set_pinentry_strings (pin, PINENTRY_FAIL_CONFIRM);
	  if (rc)
	    goto done;

	  goto again;
	}
    }

 done:
  xfree (result1);
  pinentry_disconnect (pin);
  return rc;
}

gpg_error_t
pinentry_lock (assuan_context_t ctx, struct pinentry_s *pin)
{
  gpg_error_t rc = 0;

  MUTEX_TRYLOCK (ctx, &pin_mutex, rc, pin->pinentry_opts.timeout*10);

  if (!rc)
    pin->has_lock = 1;

  return rc;
}

void
pinentry_unlock (struct pinentry_s *pin)
{
  if (pin->has_lock == 0)
    return;

  MUTEX_UNLOCK (&pin_mutex);
  pin->has_lock = 0;
}

void
pinentry_deinit (struct pinentry_s *pin)
{
  if (!pin)
    return;

  pinentry_unlock (pin);

  if (pin->ctx)
    pinentry_disconnect (pin);

  xfree (pin->name);
  xfree (pin->filename);
  pinentry_free_opts (&pin->pinentry_opts);
  xfree (pin);
}

struct pinentry_s *
pinentry_init (const char *filename)
{
  struct pinentry_s *pin = xcalloc (1, sizeof (struct pinentry_s));

  if (!pin)
    return NULL;

  pin->filename = filename ? str_dup (filename) : str_dup (_("*unknown*"));
  if (!pin->filename)
    {
      xfree (pin);
      return NULL;
    }

  return pin;
}

void
pinentry_free_opts (struct pinentry_option_s *opts)
{
  xfree (opts->title);
  xfree (opts->desc);
  xfree (opts->prompt);
  xfree (opts->ttyname);
  xfree (opts->ttytype);
  xfree (opts->display);
  xfree (opts->lc_messages);
  xfree (opts->lc_ctype);
}

gpg_error_t
pinentry_merge_options (struct pinentry_option_s *dst,
			struct pinentry_option_s *src)
{
  dst->timeout = src->timeout;

  if (src->title)
    {
      xfree (dst->title);
      dst->title = str_dup (src->title);
    }

  if (src->desc)
    {
      xfree (dst->desc);
      dst->desc = str_dup (src->desc);
    }

  if (src->prompt)
    {
      xfree (dst->prompt);
      dst->prompt = str_dup (src->prompt);
    }

  if (src->display)
    {
      xfree (dst->display);
      dst->display = str_dup (src->display);
    }

  if (src->ttyname)
    {
      xfree (dst->ttyname);
      dst->ttyname = str_dup (src->ttyname);
    }

  if (src->ttytype)
    {
      xfree (dst->ttytype);
      dst->ttytype = str_dup (src->ttytype);
    }

  if (src->lc_ctype)
    {
      xfree (dst->lc_ctype);
      dst->lc_ctype = str_dup (src->lc_ctype);
    }

  if (src->lc_messages)
    {
      xfree (dst->lc_messages);
      dst->lc_messages = str_dup (src->lc_messages);
    }

  return 0;
}
