/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <stdarg.h>
#include <string.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <pthread.h>
#include <sys/mman.h>
#include <termios.h>
#include <assert.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <setjmp.h>
#include <errno.h>
#include <poll.h>

#ifdef TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#ifdef HAVE_GETOPT_LONG
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#else
#include "getopt_long.h"
#endif

#ifdef HAVE_PR_SET_NAME
#include <sys/prctl.h>
#endif

#include "pwmd-error.h"
#include <gcrypt.h>

#include "util-misc.h"
#include "mem.h"
#include "xml.h"
#include "common.h"
#include "commands.h"
#include "cache.h"
#include "util-string.h"
#include "mutex.h"
#include "rcfile.h"
#include "crypto.h"
#include "convert.h"
#include "pinentry.h"

/* In tenths of a second. */
#define SIG_TIMEOUT 1

/* For (tcp_)accept_thread (usec). */
#define ACCEPT_TIMEOUT 30000

static int quit;
static int exiting;
static int cmdline;
static jmp_buf jmp;
static int nofork;
static pthread_cond_t quit_cond;
static pthread_mutex_t quit_mutex;
static int no_passphrase_file = 0;
static pthread_t keepalive_tid;
static int log_fd;

#ifndef HAVE_PTHREAD_CANCEL
static pthread_key_t signal_thread_key;
#endif

#ifdef WITH_GNUTLS
static int tls_fd;
static int tls6_fd;
static pthread_t tls_tid;
static pthread_t tls6_tid;
static int spawned_tls;
static int spawned_tls6;

static int start_stop_tls (int term);
#endif

static int do_cache_push (const char *filename, struct crypto_s *crypto);
static int signal_loop (sigset_t sigset);

#ifndef HAVE_PTHREAD_CANCEL
#define INIT_THREAD_SIGNAL do {				\
    struct sigaction act;				\
    sigset_t sigset;					\
    sigemptyset (&sigset);				\
    sigaddset (&sigset, SIGUSR2);			\
    pthread_sigmask (SIG_UNBLOCK, &sigset, NULL);	\
    memset (&act, 0, sizeof(act));			\
    act.sa_flags = SA_SIGINFO;				\
    act.sa_mask = sigset;				\
    act.sa_sigaction = catch_thread_signal;		\
    sigaction (SIGUSR2, &act, NULL);			\
  }  while (0)

static void
catch_thread_signal (int sig, siginfo_t *info, void *ctx)
{
  int *n = (int *) pthread_getspecific (signal_thread_key);

  *n = 1;
  pthread_setspecific (signal_thread_key, n);
}
#endif

static void
cache_push_from_rcfile ()
{
  struct crypto_s *crypto = NULL;
  char **cache_push;
  gpg_error_t rc = init_client_crypto (&crypto);

  if (rc)
    {
      log_write ("%s: %s", __FUNCTION__, pwmd_strerror (rc));
      return;
    }

#ifdef WITH_AGENT
  if (use_agent)
    {
      rc = set_agent_option (crypto->agent, "pinentry-mode", "error");
      if (rc)
	{
	  log_write ("%s: %s", __FUNCTION__, pwmd_strerror (rc));
	  return;
	}
    }
#endif

  cache_push = config_get_list ("global", "cache_push");
  if (cache_push)
    {
      char **p;

      for (p = cache_push; *p; p++)
	{
	  (void) do_cache_push (*p, crypto);
	  cleanup_crypto_stage1 (crypto);
	}

      strv_free (cache_push);
    }

#ifdef WITH_AGENT
  (void) kill_scd (crypto->agent);
#endif
  cleanup_crypto (&crypto);
}

static void
setup_logging ()
{
  int n = config_get_boolean ("global", "enable_logging");

  if (n)
    {
      char *p = config_get_string ("global", "log_path");

      if (!p || (logfile && p && log_fd != -1 && strcmp(p, logfile)))
        {
          if (log_fd != -1)
            close (log_fd);

          log_fd = -1;
        }

      xfree (logfile);
      logfile = NULL;
      if (p)
        logfile = expand_homedir (p);
      xfree (p);
    }
  else
    {
      xfree (logfile);
      logfile = NULL;
      if (log_fd != -1)
        close(log_fd);

      log_fd = -1;
    }

  log_syslog = config_get_boolean ("global", "syslog");
  if (log_syslog == 1)
    openlog ("pwmd", LOG_NDELAY | LOG_PID, LOG_DAEMON);
}

static void *
reload_rcfile_thread (void *arg)
{
#ifndef HAVE_PTHREAD_CANCEL
  int *n = xmalloc (sizeof (int));

  *n = 0;
  pthread_setspecific (signal_thread_key, n);
  INIT_THREAD_SIGNAL;
#endif

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "reload rcfile");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));
  MUTEX_LOCK (&rcfile_mutex);
  pthread_cleanup_push (cleanup_mutex_cb, &rcfile_mutex);

  for (;;)
    {
      struct slist_s *keep = NULL;
      struct slist_s *config;
      int b = disable_list_and_dump;

      pthread_cond_wait (&rcfile_cond, &rcfile_mutex);
#ifndef HAVE_PTHREAD_CANCEL
      int *cancel = (int *) pthread_getspecific (signal_thread_key);
      if (*cancel)
	break;
#endif

      pthread_cleanup_push (cleanup_mutex_cb, &rcfile_mutex);
      keep = config_keep_save ();
      log_write (_("reloading configuration file '%s'"), rcfile);

      config = config_parse (rcfile, 1);
      if (config)
	{
	  config_free (global_config);
	  global_config = config;
	  setup_logging ();
	  cache_push_from_rcfile ();
	  config_clear_keys ();
	}

      config_keep_restore (keep);
      disable_list_and_dump = !disable_list_and_dump ? b : 1;

#ifdef WITH_GNUTLS
      /* Kill existing listening threads since the configured listening
       * protocols may have changed. */
      start_stop_tls (1);
      start_stop_tls (0);
#endif
      pthread_cleanup_pop (0);
    }

  pthread_cleanup_pop (1);
  return NULL;
}

gpg_error_t
send_error (assuan_context_t ctx, gpg_error_t e)
{
  struct client_s *client = assuan_get_pointer (ctx);

  if (gpg_err_source (e) == GPG_ERR_SOURCE_UNKNOWN)
    e = gpg_error (e);

  if (client)
    client->last_rc = e;

  if (!e)
    return assuan_process_done (ctx, 0);

  if (!ctx)
    {
      log_write ("ERR %i: %s", e, pwmd_strerror (e));
      return e;
    }

  if (client && client->xml_error)
    {
      log_write ("%s", client->xml_error->message);
      xfree (client->last_error);
      client->last_error = NULL;
      if (client->xml_error->message)
	client->last_error = str_dup (client->xml_error->message);

      e = assuan_process_done (ctx,
			       assuan_set_error (ctx, e,
						 client->xml_error->message ? client->xml_error->message : NULL));
      xmlResetLastError ();
      xmlResetError (client->xml_error);
      xfree (client->xml_error);
      client->xml_error = NULL;
      return e;
    }

  return assuan_process_done (ctx,
			      assuan_set_error (ctx, e, pwmd_strerror (e)));
}

int
assuan_log_cb (assuan_context_t ctx, void *data, unsigned cat,
	       const char *msg)
{
  static pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;
  int i, t;
  int match = 0;

  pthread_mutex_lock (&m);
  pthread_cleanup_push ((void (*)(void *)) pthread_mutex_unlock, &m);
  t = strv_length (debug_level);

  for (i = 0; i < t; i++)
    {
      if (!strcasecmp (debug_level[i], (char *) "init")
	  && cat == ASSUAN_LOG_INIT)
	{
	  match = 1;
	  break;
	}

      if (!strcasecmp (debug_level[i], (char *) "ctx")
	  && cat == ASSUAN_LOG_CTX)
	{
	  match = 1;
	  break;
	}

      if (!strcasecmp (debug_level[i], (char *) "engine")
	  && cat == ASSUAN_LOG_ENGINE)
	{
	  match = 1;
	  break;
	}

      if (!strcasecmp (debug_level[i], (char *) "data")
	  && cat == ASSUAN_LOG_DATA)
	{
	  match = 1;
	  break;
	}

      if (!strcasecmp (debug_level[i], (char *) "sysio")
	  && cat == ASSUAN_LOG_SYSIO)
	{
	  match = 1;
	  break;
	}

      if (!strcasecmp (debug_level[i], (char *) "control")
	  && cat == ASSUAN_LOG_CONTROL)
	{
	  match = 1;
	  break;
	}
    }

  if (match && msg)
    {
      if (logfile)
	{
	  int fd;

	  if ((fd =
	       open (logfile, O_WRONLY | O_CREAT | O_APPEND, 0600)) == -1)
	    warn ("%s", logfile);
	  else
	    {
	      pthread_cleanup_push (cleanup_fd_cb, &fd);
	      ssize_t ret = write (fd, msg, strlen (msg));
              (void)ret;
	      pthread_cleanup_pop (1);
	    }
	}

      if (nofork)
	{
	  fprintf (stderr, "%s%s", data ? (char *) data : "", msg);
	  fflush (stderr);
	}
    }

  pthread_cleanup_pop (1);
  return match;
}

void
log_write (const char *fmt, ...)
{
  char *args;
  va_list ap;
  time_t now;
  char buf[255];
  pthread_t tid = pthread_self ();
  static pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;

  if ((!logfile && !isatty (STDERR_FILENO) && !log_syslog) || !fmt)
    return;

  pthread_mutex_lock (&m);
  pthread_cleanup_push ((void (*)(void *)) pthread_mutex_unlock, &m);

  if (!cmdline && logfile && log_fd == -1)
    {
      log_fd = open (logfile, O_WRONLY | O_CREAT | O_APPEND, 0600);
      if (log_fd == -1)
	warn ("%s", logfile);
    }

  va_start (ap, fmt);

  if (str_vasprintf (&args, fmt, ap) != -1)
    {
      if (cmdline)
	{
	  pthread_cleanup_push (xfree, args);
	  fprintf (stderr, "pwmd: %s\n", args);
	  fflush (stderr);
	  pthread_cleanup_pop (1);
	}
      else
	{
	  char *name = pthread_getspecific (thread_name_key);
          char *line;

	  pthread_cleanup_push (xfree, args);
	  snprintf (buf, sizeof (buf), "%s(%p): ", name ? name : _("unknown"),
		    (pthread_t *) tid);
	  name = buf;

	  if (!cmdline && log_syslog && !nofork)
	    syslog (LOG_INFO, "%s%s", name, args);

	  time (&now);
          struct tm *tm = localtime (&now);
          char tbuf[21];
	  strftime (tbuf, sizeof (tbuf), "%b %d %Y %H:%M:%S ", tm);
	  tbuf[sizeof (tbuf) - 1] = 0;

	  if (args[strlen (args) - 1] == '\n')
	    args[strlen (args) - 1] = 0;

	  line = str_asprintf ("%s %i %s%s\n", tbuf, getpid (), name,
                                     args);
	  pthread_cleanup_pop (1);
	  if (line)
	    {
	      pthread_cleanup_push (xfree, line);
	      if (logfile && log_fd != -1)
		{
		  ssize_t ret = write (log_fd, line, strlen (line));
                  (void)ret;
		  fsync (log_fd);
		}

	      if (nofork)
		{
		  fprintf (stdout, "%s", line);
		  fflush (stdout);
		}

	      pthread_cleanup_pop (1);
	    }
	}
    }

  va_end (ap);
  pthread_cleanup_pop (0);

  if (log_fd != -1 && log_keepopen <= 0)
    {
      close(log_fd);
      log_fd = -1;
    }

  pthread_mutex_unlock (&m);
}

static gpg_error_t
setup_crypto ()
{
  if (!gcry_check_version (GCRYPT_VERSION))
    {
      fprintf (stderr, _("gcry_check_version(): Incompatible libgcrypt. "
			 "Wanted %s, got %s.\n"), GCRYPT_VERSION,
	       gcry_check_version (NULL));
      return GPG_ERR_UNKNOWN_VERSION;
    }

  gcry_set_allocation_handler (xmalloc, xmalloc, NULL, xrealloc, xfree);
  return 0;
}

gpg_error_t
do_validate_peer (assuan_context_t ctx, const char *section,
		  assuan_peercred_t * peer)
{
  char **users;
  int allowed = 0;
  gpg_error_t rc;
  struct client_s *client = assuan_get_pointer (ctx);

  if (!client)
    return GPG_ERR_EACCES;

#ifdef WITH_GNUTLS
  if (client->thd->remote)
    return tls_validate_access (client, section);
#endif

  rc = assuan_get_peercred (ctx, peer);
  if (rc)
    return rc;

  users = config_get_list (section, "allowed");
  if (users)
    {
      for (char **p = users; !rc && *p; p++)
	{
          rc = acl_check_common(client, *p, (*peer)->uid, (*peer)->gid,
                                &allowed);
        }

      strv_free (users);
    }

  return allowed && !rc ? 0 : rc ? rc : GPG_ERR_EACCES;
}

/* Test if uid is a member of group 'name'. Invert when 'not' is true. */
#ifdef HAVE_GETGRNAM_R
static gpg_error_t
acl_check_group (const char *name, uid_t uid, gid_t gid, int not, int *allowed)
{
  char *buf;
  struct group gr, *gresult;
  size_t len = sysconf (_SC_GETGR_R_SIZE_MAX);
  int err;
  gpg_error_t rc = 0;

  if (len == -1)
    len = 16384;

  buf = xmalloc (len);
  if (!buf)
    return GPG_ERR_ENOMEM;

  err = getgrnam_r (name, &gr, buf, len, &gresult);
  if (!err && gresult)
    {
      if (gresult->gr_gid == gid)
        {
          xfree (buf);
          *allowed = !not;
          return 0;
        }

      for (char **t = gresult->gr_mem; !rc && *t; t++)
        {
          char *tbuf;
          struct passwd pw;
          struct passwd *result = get_pwd_struct (*t, 0, &pw, &tbuf, &rc);

          if (!rc && result && result->pw_uid == uid)
            {
              xfree (tbuf);
              *allowed = !not;
              break;
            }

          xfree (tbuf);
        }

      xfree (buf);
      return rc;
    }
  else if (err)
    rc = gpg_error_from_errno (err);

  xfree (buf);
  return rc ? rc : !gresult ? 0 : GPG_ERR_EACCES;
}
#else
static gpg_error_t
acl_check_group (const char *name, uid_t uid, gid_t gid, int not, int *allowed)
{
  struct group *gresult;
  gpg_error_t rc = 0;

  errno = 0;
  gresult = getgrnam (name);
  if (!errno && gresult && gresult->gr_gid == gid)
    {
      *allowed = !not;
      return 0;
    }
  else if (errno)
    rc = gpg_error_from_syserror ();
  else if (!gresult)
    return 0;

  for (char **t = gresult->gr_mem; !rc && *t; t++)
    {
      char *buf;
      struct passwd pw;
      struct passwd *result = get_pwd_struct (*t, 0, &pw, &buf, &rc);

      if (!rc && result && result->pw_uid == uid)
        {
          xfree (buf);
          *allowed = !not;
          break;
        }

      xfree (buf);
    }

  return rc;
}
#endif

gpg_error_t
peer_is_invoker(struct client_s *client)
{
  struct invoking_user_s *user;
  int allowed = 0;

  if (client->thd->state == CLIENT_STATE_UNKNOWN)
    return GPG_ERR_EACCES;

  for (user = invoking_users; user; user = user->next)
    {
#ifdef WITH_GNUTLS
      if (client->thd->remote)
        {
          if (user->type == INVOKING_TLS
              && !strcmp(client->thd->tls->fp, user->id))
            allowed = user->not ? 0 : 1;

          continue;
        }
#endif

      if (user->type == INVOKING_GID)
        {
          gpg_error_t rc = acl_check_group (user->id,
                                            client->thd->peer->uid,
                                            client->thd->peer->gid,
                                            user->not, &allowed);
          if (rc)
            return rc;
        }
      else if (user->type == INVOKING_UID && client->thd->peer->uid == user->uid)
        allowed = user->not ? 0 : 1;
    }

  return allowed ? 0 : GPG_ERR_EACCES;
}

#ifdef HAVE_GETGRNAM_R
gpg_error_t
acl_check_common (struct client_s *client, const char *user, uid_t uid,
                  gid_t gid, int *allowed)
{
  int not = 0;
  int rw = 0;
  int tls = 0;
  gpg_error_t rc = 0;

  if (!user || !*user)
    return 0;

  if (*user == '-' || *user == '!')
    not = 1;
 
  if (*user == '+') // not implemented yet
    rw = 1;

  if (*user == '#') // TLS fingerprint hash
    tls = 1;

  if (not || rw || tls)
    user++;

  if (tls)
    {
#ifdef WITH_GNUTLS
      if (client->thd->remote)
        {
          if (!strcasecmp (client->thd->tls->fp, user))
            *allowed = !not;
        }

      return 0;
#else
      return 0;
#endif
    }
#ifdef WITH_GNUTLS
  else if (client->thd->remote) // Remote client with no TLS in the ACL
    return 0;
#endif

  if (*user == '@') // all users in group
    return acl_check_group (user+1, uid, gid, not, allowed);
  else
    {
      char *buf;
      struct passwd pw;
      struct passwd *pwd = get_pwd_struct (user, 0, &pw, &buf, &rc);

      if (!rc && pwd && pwd->pw_uid == uid)
        *allowed = !not;

      xfree (buf);
    }

  return rc;
}
#else
gpg_error_t
acl_check_common (struct client_s *client, const char *user, uid_t uid,
                  gid_t gid, int *allowed)
{
  gpg_error_t rc = 0;
  int not = 0;
  int rw = 0;
  int tls = 0;

  if (!user || !*user)
    return 0;

  if (*user == '-' || *user == '!')
    not = 1;
 
  if (*user == '+') // not implemented yet
    rw = 1;

  if (*user == '#') // TLS fingerprint hash
    tls = 1;

  if (not || rw || tls)
    user++;

  if (tls)
    {
#ifdef WITH_GNUTLS
      if (client->thd->remote)
        {
          if (!strcasecmp (client->thd->tls->fp, user))
            *allowed = !not;
        }

      return 0;
#else
      return 0;
#endif
    }

  if (*user == '@') // all users in group
    return acl_check_group (user+1, uid, gid, not, allowed);
  else
    {
      char *buf;
      struct passwd pw;
      struct passwd *result = get_pwd_struct (user, 0, &pw, &buf, &rc);

      if (!rc && result && result->pw_uid == uid)
        *allowed = !not;

      xfree (buf);
    }

  return rc;
}
#endif

static gpg_error_t
validate_peer (struct client_s *cl)
{
  gpg_error_t rc;

#ifdef WITH_GNUTLS
  if (cl->thd->remote)
    return tls_validate_access (cl, NULL);
#endif

  MUTEX_LOCK (&cn_mutex);
  rc = do_validate_peer (cl->ctx, "global", &cl->thd->peer);
  MUTEX_UNLOCK (&cn_mutex);
  log_write ("peer %s: uid=%i, gid=%i, pid=%i, rc=%u",
             !rc ? _("accepted") : _("rejected"), cl->thd->peer->uid,
             cl->thd->peer->gid, cl->thd->peer->pid, rc);
  return rc;
}

static void
xml_error_cb (void *data, xmlErrorPtr e)
{
  struct client_s *client = data;

  /*
   * Keep the first reported error as the one to show in the error
   * description. Reset in send_error().
   */
  if (client->xml_error)
    return;

  client->xml_error = xcalloc (1, sizeof(xmlError));
  xmlCopyError (e, client->xml_error);
}

static pid_t
hook_waitpid (assuan_context_t ctx, pid_t pid, int action,
	      int *status, int options)
{
  return waitpid (pid, status, options);
}

static ssize_t
hook_read (assuan_context_t ctx, assuan_fd_t fd, void *data, size_t len)
{
#ifdef WITH_GNUTLS
  struct client_s *client = assuan_get_pointer (ctx);

  if (client->thd->remote)
    return tls_read_hook (ctx, (int) fd, data, len);
#endif

  return read ((int) fd, data, len);
}

static ssize_t
hook_write (assuan_context_t ctx, assuan_fd_t fd,
	    const void *data, size_t len)
{
#ifdef WITH_GNUTLS
  struct client_s *client = assuan_get_pointer (ctx);

  if (client->thd->remote)
    return tls_write_hook (ctx, (int) fd, data, len);
#endif

  return write ((int) fd, data, len);
}

static int
new_connection (struct client_s *cl)
{
  gpg_error_t rc;
  static struct assuan_malloc_hooks mhooks = { xmalloc, xrealloc, xfree };
  static struct assuan_system_hooks shooks = {
    ASSUAN_SYSTEM_HOOKS_VERSION,
    __assuan_usleep,
    __assuan_pipe,
    __assuan_close,
    hook_read,
    hook_write,
    //FIXME
    NULL,			//recvmsg
    NULL,			//sendmsg both are used for FD passing
    __assuan_spawn,
    hook_waitpid,
    __assuan_socketpair,
    __assuan_socket,
    __assuan_connect
  };

#ifdef WITH_GNUTLS
  if (cl->thd->remote)
    {
      char *prio = config_get_string ("global", "tls_cipher_suite");

      cl->thd->timeout = config_get_integer ("global", "tls_timeout");
      if (fcntl (cl->thd->fd, F_SETFL, O_NONBLOCK) == -1)
        return 0;

      cl->thd->tls = tls_init (cl->thd->fd, cl->thd->timeout, prio);
      xfree (prio);
      if (!cl->thd->tls)
	return 0;
    }
#endif

  rc = assuan_new_ext (&cl->ctx, GPG_ERR_SOURCE_DEFAULT, &mhooks,
		       debug_level ? assuan_log_cb : NULL, NULL);
  if (rc)
    goto fail;

  assuan_ctx_set_system_hooks (cl->ctx, &shooks);
  rc = assuan_init_socket_server (cl->ctx, cl->thd->fd, 2);
  if (rc)
    goto fail;

  assuan_set_pointer (cl->ctx, cl);
  assuan_set_hello_line (cl->ctx, PACKAGE_STRING);
  rc = register_commands (cl->ctx);
  if (rc)
    goto fail;

  rc = assuan_accept (cl->ctx);
  if (rc)
    goto fail;

  rc = validate_peer (cl);
  /* May not be implemented on all platforms. */
  if (rc && gpg_err_code (rc) != GPG_ERR_ASS_GENERAL)
    goto fail;

  MUTEX_LOCK (&cn_mutex);
  cl->thd->state = CLIENT_STATE_INIT;
  MUTEX_UNLOCK (&cn_mutex);
  rc = init_client_crypto (&cl->crypto);
  if (rc)
    goto fail;

#ifdef WITH_AGENT
  if (use_agent)
    cl->crypto->agent->client_ctx = cl->ctx;
#endif

  cl->crypto->client_ctx = cl->ctx;
  cl->lock_timeout = config_get_integer ("global", "lock_timeout");
  xmlSetStructuredErrorFunc (cl, xml_error_cb);
  return 1;

fail:
  log_write ("%s", pwmd_strerror (rc));
  return 0;
}

/*
 * This is called after a client_thread() terminates. Set with
 * pthread_cleanup_push().
 */
static void
cleanup_cb (void *arg)
{
  struct client_thread_s *cn = arg;
  struct client_s *cl = cn->cl;

  MUTEX_LOCK (&cn_mutex);
  cn_thread_list = slist_remove (cn_thread_list, cn);
  MUTEX_UNLOCK (&cn_mutex);

  if (cl)
    {
      unlock_flock (&cl->flock_fd);
      cleanup_client (cl);
      if (cl->xml_error)
	xmlResetError (cl->xml_error);

      xfree (cl->xml_error);

#ifdef WITH_GNUTLS
      if (cn->tls)
	{
	  gnutls_deinit (cn->tls->ses);
	  xfree (cn->tls->fp);
	  xfree (cn->tls);
	}
#endif

      if (!cn->atfork && cl->ctx)
	assuan_release (cl->ctx);
      else if (!cn->atfork && cl->thd && cl->thd->fd != -1)
	close (cl->thd->fd);

      if (cl->crypto)
	cleanup_crypto (&cl->crypto);

      pinentry_free_opts (&cl->pinentry_opts);
      xfree (cl);
    }
  else
    {
      if (cn->fd != -1)
	close (cn->fd);
    }

  while (cn->msg_queue)
    {
      struct status_msg_s *msg = cn->msg_queue;

      cn->msg_queue = msg->next;
      xfree (msg->line);
      xfree (msg);
    }

   if (!cn->atfork && cn->status_msg_pipe[0] != -1)
     close (cn->status_msg_pipe[0]);

   if (!cn->atfork && cn->status_msg_pipe[1] != -1)
     close (cn->status_msg_pipe[1]);

   pthread_mutex_destroy (&cn->status_mutex);

   if (!cn->atfork)
     {
       log_write (_("exiting, fd=%i"), cn->fd);
       send_status_all (STATUS_CLIENTS, NULL);
     }

  xfree (cn->name);
#ifdef WITH_GNUTLS
  xfree (cn->peeraddr);
#endif
  xfree (cn);
  pthread_cond_signal (&quit_cond);
}

void
cleanup_all_clients (int atfork)
{
  /* This function may be called from pthread_atfork() which requires
     reinitialization. */
  if (atfork)
    {
      pthread_mutexattr_t attr;

      pthread_mutexattr_init (&attr);
      pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
      pthread_mutex_init (&cn_mutex, &attr);
      pthread_mutexattr_destroy (&attr);
      cache_mutex_init ();
    }

  MUTEX_LOCK (&cn_mutex);

  while (slist_length (cn_thread_list))
    {
      struct client_thread_s *thd = slist_nth_data (cn_thread_list, 0);

      thd->atfork = atfork;
      cleanup_cb (thd);
    }

  exiting = 1;
  MUTEX_UNLOCK (&cn_mutex);
  cache_deinit (atfork);
}

static gpg_error_t
send_msg_queue (struct client_thread_s *thd)
{
  MUTEX_LOCK (&thd->status_mutex);
  gpg_error_t rc = 0;
  char c;
  ssize_t ret;

  ret = read (thd->status_msg_pipe[0], &c, 1);
  rc = gpg_error_from_syserror ();
  if (ret == -1 && gpg_err_code (rc) != GPG_ERR_EAGAIN)
    log_write ("%s (%i): %s", __FUNCTION__, __LINE__, pwmd_strerror (rc));

  thd->wrote_status = 0;

  while (thd->msg_queue)
    {
      struct status_msg_s *msg = thd->msg_queue;

#ifndef HAVE_PTHREAD_CANCEL
      if (thd->fd == -1)
        break;
#endif

      thd->msg_queue = thd->msg_queue->next;
      MUTEX_UNLOCK (&thd->status_mutex);
      rc = send_status (thd->cl->ctx, msg->s, msg->line);
      MUTEX_LOCK (&thd->status_mutex);
      xfree (msg->line);
      xfree (msg);

      if (rc)
	break;
    }

  MUTEX_UNLOCK (&thd->status_mutex);
  if (gpg_err_code (rc) == GPG_ERR_EAGAIN)
    rc = 0;

  if (rc && gpg_err_code (rc) != GPG_ERR_EPIPE)
    log_write ("%s: %s", __FUNCTION__, pwmd_strerror (rc));

  return rc;
}

static void *
client_thread (void *data)
{
  struct client_thread_s *thd = data;
  struct client_s *cl = xcalloc (1, sizeof (struct client_s));

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "client");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));

  if (!cl)
    {
      log_write ("%s(%i): %s", __FILE__, __LINE__,
		 pwmd_strerror (GPG_ERR_ENOMEM));
      return NULL;
    }

  MUTEX_LOCK (&cn_mutex);
  pthread_cleanup_push (cleanup_cb, thd);
  thd->cl = cl;
  cl->thd = thd;
  cl->flock_fd = -1;
  MUTEX_UNLOCK (&cn_mutex);

  if (new_connection (cl))
    {
      int finished = 0;
      gpg_error_t rc;
      struct pollfd fds[2];

      fds[0].fd = thd->fd;
      fds[0].events = POLLIN;
      fds[1].fd = thd->status_msg_pipe[0];
      fds[1].events = POLLIN;

      send_status_all (STATUS_CLIENTS, NULL);
      rc = send_status (cl->ctx, STATUS_CACHE, NULL);
      if (rc)
	{
	  log_write ("%s(%i): %s", __FILE__, __LINE__, pwmd_strerror (rc));
	  finished = 1;
	}

      while (!finished)
	{
          int n;
	  int eof;

          n = poll (fds, 2, -1);
	  if (n == -1)
	    {
	      log_write ("%s", strerror (errno));
	      break;
	    }

          if (fds[1].revents & POLLIN)
	    {
	      rc = send_msg_queue (thd);
	      if (rc && gpg_err_code (rc) != GPG_ERR_EPIPE)
		break;
	    }

#ifdef HAVE_PTHREAD_CANCEL
          if (!(fds[0].revents & POLLIN))
#else
          if (thd->fd != -1 && !(fds[0].revents & POLLIN))
#endif
	    continue;

	  rc = assuan_process_next (cl->ctx, &eof);
	  if (rc || eof)
	    {
	      if (gpg_err_code (rc) == GPG_ERR_EOF || eof)
		break;

	      log_write ("assuan_process_next(): rc=%i %s", rc,
			 pwmd_strerror (rc));
	      if (rc == gpg_error (GPG_ERR_ETIMEDOUT))
		break;

	      rc = send_error (cl->ctx, rc);
	      if (rc)
		{
		  log_write ("assuan_process_done(): rc=%i %s", rc,
			     pwmd_strerror (rc));
		  break;
		}
	    }

	  /* Since the msg queue pipe fd's are non-blocking, check for
	   * pending status msgs here. GPG_ERR_EPIPE can be seen when the
	   * client has already disconnected and will be converted to
	   * GPG_ERR_EOF during assuan_process_next().
	   */
	  rc = send_msg_queue (thd);
	  if (rc && gpg_err_code (rc) != GPG_ERR_EPIPE)
	    break;
	}
    }

  /* Don't do pthread_exit() here because any set pthread_cleanup_push
   * functions would be called after a command failed but then the client
   * exited normally which may lead to a double free. */
  pthread_cleanup_pop (1);
  return NULL;
}

static int
xml_import (const char *filename, const char *outfile,
	    const char *keygrip, const char *sign_keygrip,
	    const char *keyfile, int no_passphrase, const char *cipher,
	    const char *params, uint64_t iterations)
{
  xmlDocPtr doc;
  int fd;
  struct stat st;
  int len;
  xmlChar *xmlbuf;
  xmlChar *xml;
  gpg_error_t rc;
  struct crypto_s *crypto = NULL;
  void *key = NULL;
  size_t keylen = 0;
  int algo = cipher ? cipher_string_to_gcrypt ((char *) cipher) :
    GCRY_CIPHER_AES256;

  if (algo == -1)
    {
      log_write ("ERR %i: %s", gpg_error (GPG_ERR_CIPHER_ALGO),
		 pwmd_strerror (GPG_ERR_CIPHER_ALGO));
      return 0;
    }

  if (stat (filename, &st) == -1)
    {
      log_write ("%s: %s", filename,
		 pwmd_strerror (gpg_error_from_errno (errno)));
      return 0;
    }

  rc = init_client_crypto (&crypto);
  if (rc)
    return 0;

  memcpy (&crypto->save.hdr, &crypto->hdr, sizeof (file_header_t));
  crypto->save.hdr.flags = set_cipher_flag (crypto->save.hdr.flags, algo);
  log_write (_("Importing XML from '%s'. Output will be written to '%s' ..."),
	     filename, outfile);

  if ((fd = open (filename, O_RDONLY)) == -1)
    {
      log_write ("%s: %s", filename,
		 pwmd_strerror (gpg_error_from_errno (errno)));
      goto fail;
    }

  if ((xmlbuf = xmalloc (st.st_size + 1)) == NULL)
    {
      close (fd);
      log_write ("%s(%i): %s", __FILE__, __LINE__,
		 pwmd_strerror (GPG_ERR_ENOMEM));
      goto fail;
    }

  if (read (fd, xmlbuf, st.st_size) == -1)
    {
      rc = gpg_error_from_errno (errno);
      close (fd);
      log_write ("%s: %s", filename, pwmd_strerror (rc));
      goto fail;
    }

  close (fd);
  xmlbuf[st.st_size] = 0;
  /*
   * Make sure the document validates.
   */
  if ((doc = xmlReadDoc (xmlbuf, NULL, "UTF-8", XML_PARSE_NOBLANKS)) == NULL)
    {
      log_write ("xmlReadDoc() failed");
      xfree (xmlbuf);
      goto fail;
    }

  xfree (xmlbuf);
  xmlNodePtr n = xmlDocGetRootElement (doc);
  if (n && !xmlStrEqual (n->name, (xmlChar *) "pwmd"))
    {
      log_write (_("Could not find root \"pwmd\" element."));
      rc = GPG_ERR_BAD_DATA;
    }

  if (!rc)
    rc = validate_import (NULL, n ? n->children : n);

  if (rc)
    {
      log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
      xmlFreeDoc (doc);
      goto fail;
    }

  xmlDocDumpMemory (doc, &xml, &len);
  xmlFreeDoc (doc);
  crypto->save.hdr.s2k_count = iterations;
  if (!use_agent)
    {
      rc = export_common (NULL, 0, crypto, xml, len, outfile, keyfile, &key,
			  &keylen, 0, 0, no_passphrase);
      if (!rc)
	log_write (_("Success!"));
    }
#ifdef WITH_AGENT
  else
    {
      rc = agent_set_pinentry_options (crypto->agent);
      if (!rc)
	rc = agent_export_common (crypto, keygrip, sign_keygrip, no_passphrase,
				  xml, len, outfile, params, keyfile);
    }
#endif

  gcry_free (key);
  xmlFree (xml);
  if (rc)
    {
      send_error (NULL, rc);
      goto fail;
    }

  cleanup_crypto (&crypto);
  return 1;

fail:
  cleanup_crypto (&crypto);
  return 0;
}

static int
do_cache_push (const char *filename, struct crypto_s *crypto)
{
  unsigned char md5file[16];
  gpg_error_t rc;
  void *key = NULL;
  size_t keylen = 0;
  xmlDocPtr doc;
  struct cache_data_s *cdata;
  unsigned char *crc;
  size_t len;
  int fd = -1;

  log_write (_("Trying to add datafile '%s' to the file cache ..."),
	     filename);

  if (valid_filename (filename) == 0)
    {
      log_write (_("%s: Invalid characters in filename"), filename);
      return 0;
    }

  rc = lock_flock (NULL, filename, LOCK_SH, &fd);
  if (!rc)
    rc = decrypt_common (NULL, 0, crypto, filename, &key, &keylen, NULL, NULL);
  if (rc)
    goto fail;

  rc = parse_doc ((char *) crypto->plaintext, crypto->plaintext_len, &doc);
  if (rc)
    {
      xfree (key);
      goto fail;
    }

  gcry_md_hash_buffer (GCRY_MD_MD5, md5file, filename, strlen (filename));
  cdata = xcalloc (1, sizeof (struct cache_data_s));
  if (!cdata)
    {
      xmlFreeDoc (doc);
      xfree (key);
      goto fail;
    }

  rc = get_checksum (filename, &crc, &len);
  unlock_flock (&fd);
  if (rc)
    {
      xmlFreeDoc (doc);
      free_cache_data_once (cdata);
      goto fail;
    }

  cdata->crc = crc;
  rc = encrypt_xml (NULL, cache_key, cache_keysize, GCRY_CIPHER_AES,
		    crypto->plaintext, crypto->plaintext_len, &cdata->doc,
		    &cdata->doclen, &cache_iv, &cache_blocksize);
  if (!rc && !IS_PKI (crypto))
    {
      cdata->key = key;
      cdata->keylen = keylen;
    }
  else
    xfree (key);

  if (rc)
    {
      xmlFreeDoc (doc);
      free_cache_data_once (cdata);
      goto fail;
    }

#ifdef WITH_AGENT
  if (use_agent && IS_PKI (crypto))
    {
      gcry_sexp_build ((gcry_sexp_t *) & cdata->pubkey, NULL, "%S",
		       crypto->pkey_sexp);
      gcry_sexp_build ((gcry_sexp_t *) & cdata->sigkey, NULL, "%S",
		       crypto->sigpkey_sexp);
    }
#endif

  unlock_flock (&fd);
  int timeout = config_get_integer (filename, "cache_timeout");
  cache_add_file (md5file, crypto->grip, cdata, timeout);
  log_write (_("Successfully added '%s' to the cache."), filename);
  return 1;

fail:
  unlock_flock (&fd);
  log_write ("ERR %u: %s: %s", rc, filename, pwmd_strerror (rc));
  return 0;
}

static gpg_error_t
init_client (int fd, const char *addr)
{
  gpg_error_t rc = 0;
  struct client_thread_s *new = xcalloc (1, sizeof (struct client_thread_s));

  if (!new)
    {
      close (fd);
      return GPG_ERR_ENOMEM;
    }

  MUTEX_LOCK (&cn_mutex);
  pthread_cleanup_push (cleanup_mutex_cb, &cn_mutex);

  if (pipe (new->status_msg_pipe) == -1)
    rc = gpg_error_from_errno (errno);

  if (!rc)
    {
      if (fcntl (new->status_msg_pipe[0], F_SETFL, O_NONBLOCK) == -1)
        rc = gpg_error_from_errno (errno);

      if (!rc)
        if (fcntl (new->status_msg_pipe[1], F_SETFL, O_NONBLOCK) == -1)
          rc = gpg_error_from_errno (errno);

      pthread_mutex_init (&new->status_mutex, NULL);
    }

  if (!rc)
    {
#ifdef WITH_GNUTLS
      new->remote = addr ? 1 : 0;
#endif
      new->fd = fd;
      rc = create_thread (client_thread, new, &new->tid, 1);
      if (rc)
	{
	  close (new->status_msg_pipe[0]);
	  close (new->status_msg_pipe[1]);
	  pthread_mutex_destroy (&new->status_mutex);
	}
    }

  if (!rc)
    {
      struct slist_s *list = slist_append (cn_thread_list, new);

      if (list)
	{
	  cn_thread_list = list;
	  if (addr)
            {
              log_write (_("new connection: tid=%p, fd=%i, addr=%s"),
                         (pthread_t *) new->tid, fd, addr);
#ifdef WITH_GNUTLS
              new->peeraddr = str_dup (addr);
#endif
            }
	  else
	    log_write (_("new connection: tid=%p, fd=%i"),
		       (pthread_t *) new->tid, fd);
	}
      else
	rc = GPG_ERR_ENOMEM;
    }

  pthread_cleanup_pop (1);

  if (rc)
    {
      xfree (new);
      close (fd);
      log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		 pwmd_strerror (rc));
    }
  return rc;
}

static void*
keepalive_thread (void *arg)
{
#ifndef HAVE_PTHREAD_CANCEL
  int *n = xmalloc (sizeof (int));

  *n = 0;
  pthread_setspecific (signal_thread_key, n);
  INIT_THREAD_SIGNAL;
#endif

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "keepalive");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));

  for (;;)
    {
      int n = config_get_integer ("global", "keepalive_interval");
      struct timeval tv = { n, 0 };
#ifndef HAVE_PTHREAD_CANCEL
      int *sigusr2;

      sigusr2 = (int *) pthread_getspecific (signal_thread_key);
      if (*sigusr2)
	break;
#endif

      send_status_all (STATUS_KEEPALIVE, NULL);
      select (0, NULL, NULL, NULL, &tv);
    }

  return NULL;
}

#ifdef WITH_GNUTLS
/* From Beej's Guide to Network Programming. It's a good tutorial. */
static void *
get_in_addr (struct sockaddr *sa)
{
  if (sa->sa_family == AF_INET)
    return &(((struct sockaddr_in *) sa)->sin_addr);

  return &(((struct sockaddr_in6 *) sa)->sin6_addr);
}

static void *
tcp_accept_thread (void *arg)
{
  int sockfd = *(int *) arg;
#ifndef HAVE_PTHREAD_CANCEL
  int *n = xmalloc (sizeof (int));

  *n = 0;
  pthread_setspecific (signal_thread_key, n);
  INIT_THREAD_SIGNAL;
  fcntl (sockfd, F_SETFL, O_NONBLOCK);
#endif

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "tcp_accept");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));

  for (;;)
    {
      struct sockaddr_storage raddr;
      socklen_t slen = sizeof (raddr);
      int fd;
      unsigned long n;
      char s[INET6_ADDRSTRLEN];
      struct timeval tv = { 0, ACCEPT_TIMEOUT };
#ifndef HAVE_PTHREAD_CANCEL
      int *sigusr2;

      sigusr2 = (int *) pthread_getspecific (signal_thread_key);
      if (*sigusr2)
	break;
#endif

      fd = accept (sockfd, (struct sockaddr *) &raddr, &slen);
      if (fd == -1)
	{
	  if (errno == EMFILE || errno == ENFILE)
	    log_write ("accept(): %s",
		       pwmd_strerror (gpg_error_from_errno (errno)));
	  else if (errno != EAGAIN)
	    {
	      if (!quit)	// probably EBADF
		log_write ("accept(): %s", strerror (errno));

	      break;
	    }

#ifndef HAVE_PTHREAD_CANCEL
	  select (0, NULL, NULL, NULL, &tv);
#endif
	  continue;
	}

      if (quit)
        {
          close (fd);
          break;
        }

      inet_ntop (raddr.ss_family, get_in_addr ((struct sockaddr *) &raddr),
		 s, sizeof s);
      (void) init_client (fd, s);
      n = config_get_integer ("global", "tcp_wait");
      if (n > 0)
	{
	  tv.tv_sec = (n * 100000) / 100000;
	  tv.tv_usec = (n * 100000) % 100000;
	  select (0, NULL, NULL, NULL, &tv);
	}
    }

  return NULL;
}

static int
start_stop_tls_with_protocol (int ipv6, int term)
{
  struct addrinfo hints, *servinfo, *p;
  int port = config_get_integer ("global", "tcp_port");
  char buf[7];
  int n;
  gpg_error_t rc;
  int *fd = ipv6 ? &tls6_fd : &tls_fd;

  if (term || config_get_boolean ("global", "enable_tcp") == 0)
    {
      if (tls6_fd != -1)
	{
	  if (spawned_tls6)
	    {
#ifdef HAVE_PTHREAD_CANCEL
	      pthread_cancel (tls6_tid);
#else
	      pthread_kill (tls6_tid, SIGUSR2);
#endif
	      pthread_join (tls6_tid, NULL);
	    }

	  shutdown (tls6_fd, SHUT_RDWR);
	  close (tls6_fd);
	  tls6_fd = -1;
	  spawned_tls6 = 0;
	}

      if (tls_fd != -1)
	{
	  if (spawned_tls)
	    {
#ifdef HAVE_PTHREAD_CANCEL
	      pthread_cancel (tls_tid);
#else
	      pthread_kill (tls_tid, SIGUSR2);
#endif
	      pthread_join (tls_tid, NULL);
	    }

	  shutdown (tls_fd, SHUT_RDWR);
	  close (tls_fd);
	  tls_fd = -1;
	  spawned_tls = 0;
	}

      /* A client may still be connected. */
      if (!quit && x509_cred != NULL)
	tls_deinit_params ();

      return 1;
    }

  if ((ipv6 && tls6_fd != -1) || (!ipv6 && tls_fd != -1))
    return 1;

  memset (&hints, 0, sizeof (hints));
  hints.ai_family = ipv6 ? AF_INET6 : AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;
  snprintf (buf, sizeof (buf), "%i", port);

  if ((n = getaddrinfo (NULL, buf, &hints, &servinfo)) == -1)
    {
      log_write ("getaddrinfo(): %s", gai_strerror (n));
      return 0;
    }

  for (n = 0, p = servinfo; p != NULL; p = p->ai_next)
    {
      int r = 1;

      if ((ipv6 && p->ai_family != AF_INET6)
	  || (!ipv6 && p->ai_family != AF_INET))
	continue;

      if ((*fd = socket (p->ai_family, p->ai_socktype, p->ai_protocol)) == -1)
	{
	  log_write ("socket(): %s", strerror (errno));
	  continue;
	}

      if (setsockopt (*fd, SOL_SOCKET, SO_REUSEADDR, &r, sizeof (int)) == -1)
	{
	  log_write ("setsockopt(): %s",
		     pwmd_strerror (gpg_error_from_errno (errno)));
	  freeaddrinfo (servinfo);
	  goto fail;
	}

      if (bind (*fd, p->ai_addr, p->ai_addrlen) == -1)
	{
	  close (*fd);
	  log_write ("bind(): %s",
		     pwmd_strerror (gpg_error_from_errno (errno)));
	  continue;
	}

      n++;
      break;
    }

  freeaddrinfo (servinfo);

  if (!n)
    goto fail;

#if HAVE_DECL_SO_BINDTODEVICE != 0
  char *tmp = config_get_string ("global", "tcp_interface");
  if (tmp && setsockopt (*fd, SOL_SOCKET, SO_BINDTODEVICE, tmp,
			 strlen (tmp)) == -1)
    {
      log_write ("setsockopt(): %s",
		 pwmd_strerror (gpg_error_from_errno (errno)));
      xfree (tmp);
      goto fail;
    }

  xfree (tmp);
#endif

  if (x509_cred == NULL)
    {
      char *tmp = config_get_string ("global", "tls_dh_level");

      rc = tls_init_params (tmp);
      xfree (tmp);
      if (rc)
	goto fail;
    }

  if (listen (*fd, 0) == -1)
    {
      log_write ("listen(): %s", strerror (errno));
      goto fail;
    }

  if (ipv6)
    rc = create_thread (tcp_accept_thread, fd, &tls6_tid, 0);
  else
    rc = create_thread (tcp_accept_thread, fd, &tls_tid, 0);

  if (rc)
    {
      log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		 pwmd_strerror (rc));
      goto fail;
    }

  if (ipv6)
    spawned_tls6 = 1;
  else
    spawned_tls = 1;

  return 1;

fail:
  start_stop_tls_with_protocol (0, 1);
  if (tls_fd != -1)
    close (tls_fd);

  if (tls6_fd != -1)
    close (tls6_fd);

  tls_fd = -1;
  tls6_fd = -1;
  return 0;
}

static int
start_stop_tls (int term)
{
  char *s = config_get_string ("global", "tcp_bind");
  int b;

  if (!s)
    return 0;

  if (!strcmp (s, "any"))
    {
      b = start_stop_tls_with_protocol (0, term);
      if (b)
	b = start_stop_tls_with_protocol (1, term);
    }
  else if (!strcmp (s, "ipv4"))
    b = start_stop_tls_with_protocol (0, term);
  else if (!strcmp (s, "ipv6"))
    b = start_stop_tls_with_protocol (1, term);
  else
    b = 0;

  xfree (s);
  return b;
}
#endif

static void *
accept_thread (void *arg)
{
  int sockfd = *(int *) arg;
#ifndef HAVE_PTHREAD_CANCEL
  int *n = xmalloc (sizeof (int));

  *n = 0;
  pthread_setspecific (signal_thread_key, n);
  INIT_THREAD_SIGNAL;
  fcntl (sockfd, F_SETFL, O_NONBLOCK);
#endif

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "accept");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));

  for (;;)
    {
      socklen_t slen = sizeof (struct sockaddr_un);
      struct sockaddr_un raddr;
      int fd;
#ifndef HAVE_PTHREAD_CANCEL
      struct timeval tv = { 0, ACCEPT_TIMEOUT };
      int *sigusr2 = (int *) pthread_getspecific (signal_thread_key);

      if (*sigusr2)
	break;
#endif

      fd = accept (sockfd, (struct sockaddr *) &raddr, &slen);
      if (fd == -1)
	{
	  if (errno == EMFILE || errno == ENFILE)
	    log_write ("accept(): %s",
		       pwmd_strerror (gpg_error_from_errno (errno)));
	  else if (errno != EAGAIN)
	    {
	      if (!quit)	// probably EBADF
		log_write ("accept(): %s",
			   pwmd_strerror (gpg_error_from_errno (errno)));

	      break;
	    }

#ifndef HAVE_PTHREAD_CANCEL
	  select (0, NULL, NULL, NULL, &tv);
#endif
	  continue;
	}

      (void) init_client (fd, NULL);
    }

  /* Just in case accept() failed for some reason other than EBADF */
  quit = 1;
  return NULL;
}

static void *
cache_timer_thread (void *arg)
{
#ifndef HAVE_PTHREAD_CANCEL
  int *n = xmalloc (sizeof (int));

  *n = 0;
  pthread_setspecific (signal_thread_key, n);
  INIT_THREAD_SIGNAL;
#endif

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "cache timer");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));

  for (;;)
    {
      struct timeval tv = { 1, 0 };
#ifndef HAVE_PTHREAD_CANCEL
      int *n;

      n = (int *) pthread_getspecific (signal_thread_key);
      if (*n)
	break;
#endif

      select (0, NULL, NULL, NULL, &tv);
      cache_adjust_timeout ();
    }

  return NULL;
}

static int
signal_loop (sigset_t sigset)
{
  int done = 0;
  int siint = 0;

  do
    {
      int sig;

      sigwait (&sigset, &sig);

      if (sig != SIGQUIT)
	log_write (_("caught signal %i (%s)"), sig, strsignal (sig));

      switch (sig)
	{
	case SIGHUP:
	  pthread_cond_signal (&rcfile_cond);
	  break;
	case SIGUSR1:
	  log_write (_("clearing file cache"));
	  cache_clear (NULL);
	  send_status_all (STATUS_CACHE, NULL);
	  break;
	case SIGQUIT:
	  done = 1;
	  break;
	default:
	  siint = 1;
	  done = 1;
	  break;
	}
    }
  while (!done);

  return siint;
}

static void
catchsig (int sig)
{
  log_write (_ ("Caught signal %i (%s). Exiting."), sig, strsignal (sig));
#ifdef HAVE_BACKTRACE
  BACKTRACE (__FUNCTION__);
#endif
  longjmp (jmp, 1);
}

static void *
waiting_for_exit (void *arg)
{
  int last = 0;
#ifndef HAVE_PTHREAD_CANCEL
  int *n = xmalloc (sizeof (int));

  *n = 0;
  pthread_setspecific (signal_thread_key, n);
  INIT_THREAD_SIGNAL;
#endif

#ifdef HAVE_PR_SET_NAME
  prctl (PR_SET_NAME, "exiting");
#endif
  pthread_setspecific (thread_name_key, str_dup (__FUNCTION__));
  log_write (_("waiting for all clients to disconnect"));
  MUTEX_LOCK (&quit_mutex);
  pthread_cleanup_push (cleanup_mutex_cb, &quit_mutex);

  for (;;)
    {
      struct timespec ts;
      int n;

      MUTEX_LOCK (&cn_mutex);
      n = slist_length (cn_thread_list);
      MUTEX_UNLOCK (&cn_mutex);
      if (!n)
	break;

#ifndef HAVE_PTHREAD_CANCEL
      int *s = (int *) pthread_getspecific (signal_thread_key);
      if (*s)
	break;
#endif

      if (last != n)
	{
	  log_write (_("%i clients remain"), n);
	  last = n;
	}

      INIT_TIMESPEC (SIG_TIMEOUT, ts);
      pthread_cond_timedwait (&quit_cond, &quit_mutex, &ts);
    }

  kill (getpid (), SIGQUIT);
  pthread_cleanup_pop (1);
  return NULL;
}

static int
server_loop (int sockfd, char **socketpath)
{
  pthread_t accept_tid;
  pthread_t cache_timeout_tid;
  int cancel_timeout_thread = 0, cancel_accept_thread = 0;
  int cancel_keepalive_thread = 0;
  sigset_t sigset;
  int n;
  int segv = 0;
  gpg_error_t rc;

  init_commands ();
  sigemptyset (&sigset);

  /* Termination */
  sigaddset (&sigset, SIGTERM);
  sigaddset (&sigset, SIGINT);

  /* Clears the file cache. */
  sigaddset (&sigset, SIGUSR1);

  /* Configuration file reloading. */
  sigaddset (&sigset, SIGHUP);

  /* For exiting cleanly. */
  sigaddset (&sigset, SIGQUIT);

#ifndef HAVE_PTHREAD_CANCEL
  /*
    The socket, cache and rcfile threads use this signal when
    pthread_cancel() is unavailable. Prevent the main thread from
    catching this signal from another process.
  */
  sigaddset (&sigset, SIGUSR2);
#endif

  /* When mem.c cannot find a pointer in the list (double free). */
  signal (SIGABRT, catchsig);
  sigaddset (&sigset, SIGABRT);
  sigprocmask (SIG_BLOCK, &sigset, NULL);

#ifndef HAVE_PTHREAD_CANCEL
  /* Remove this signal from the watched signals in signal_loop(). */
  sigdelset (&sigset, SIGUSR2);
#endif

  /* Ignored everywhere. When a client disconnects abnormally this signal
   * gets raised. It isn't needed though because client_thread() will check
   * for rcs even after the client disconnects. */
  signal (SIGPIPE, SIG_IGN);

  /* Can show a backtrace of the stack in the log. */
  signal (SIGSEGV, catchsig);

#ifdef WITH_GNUTLS
  /* Needs to be done after the fork(). */
  if (!start_stop_tls (0))
    {
      segv = 1;
      goto done;
    }
#endif

  pthread_mutex_init (&quit_mutex, NULL);
  pthread_cond_init (&quit_cond, NULL);
  char *p = get_username (getuid());
  log_write (_("%s started for user %s"), PACKAGE_STRING, p);
  xfree (p);

#ifdef WITH_GNUTLS
  if (config_get_boolean ("global", "enable_tcp"))
    log_write (_("Listening on %s and TCP port %i"), *socketpath,
               config_get_integer ("global", "tcp_port"));
  else
    log_write (_("Listening on %s"), *socketpath);
#else
  log_write (_("Listening on %s"), *socketpath);
#endif

  rc = create_thread (keepalive_thread, NULL, &keepalive_tid, 0);
  if (rc)
    {
      log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		 pwmd_strerror (rc));
      goto done;
    }

  cancel_keepalive_thread = 1;
  rc = create_thread (reload_rcfile_thread, NULL, &rcfile_tid, 0);
  if (rc)
    {
      log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		 pwmd_strerror (rc));
      goto done;
    }

  rc = create_thread (cache_timer_thread, NULL, &cache_timeout_tid, 0);
  if (rc)
    {
      log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		 pwmd_strerror (rc));
      goto done;
    }

  cancel_timeout_thread = 1;
  rc = create_thread (accept_thread, &sockfd, &accept_tid, 0);
  if (rc)
    {
      log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		 pwmd_strerror (rc));
      goto done;
    }

  cancel_accept_thread = 1;
  if (!setjmp (jmp))
    signal_loop (sigset);
  else
    segv = 1;

done:
  /*
   * We're out of the main server loop. This happens when a signal was sent
   * to terminate the daemon. We'll wait for all clients to disconnect
   * before exiting but exit immediately if another termination signal is
   * sent.
   */
  if (cancel_accept_thread)
    {
#ifdef HAVE_PTHREAD_CANCEL
      int n = pthread_cancel (accept_tid);
#else
      int n = pthread_kill (accept_tid, SIGUSR2);
#endif
      if (!n)
	pthread_join (accept_tid, NULL);
    }

#ifdef WITH_GNUTLS
  start_stop_tls (1);
#endif
  shutdown (sockfd, SHUT_RDWR);
  close (sockfd);
  unlink (*socketpath);
  xfree (*socketpath);
  *socketpath = NULL;
  MUTEX_LOCK (&cn_mutex);
  n = slist_length (cn_thread_list);
  MUTEX_UNLOCK (&cn_mutex);

  if (n && !segv)
    {
      pthread_t tid;

      rc = create_thread (waiting_for_exit, NULL, &tid, 0);
      if (!rc)
	{
	  if (signal_loop (sigset))
	    {
	      log_write (_("Received second termination request. Exiting."));
#ifdef HAVE_PTHREAD_CANCEL
	      pthread_cancel (tid);
#else
	      pthread_kill (tid, SIGUSR2);
#endif
	      pthread_join (tid, NULL);
	    }
	}
      else
	log_write ("%s(%i): pthread_create(): %s", __FILE__, __LINE__,
		   pwmd_strerror (rc));
    }

  if (cancel_timeout_thread)
    {
#ifdef HAVE_PTHREAD_CANCEL
      pthread_cancel (cache_timeout_tid);
#else
      pthread_kill (cache_timeout_tid, SIGUSR2);
#endif
      pthread_join (cache_timeout_tid, NULL);
    }

  if (cancel_keepalive_thread)
    {
#ifdef HAVE_PTHREAD_CANCEL
      pthread_cancel (keepalive_tid);
#else
      pthread_kill (keepalive_tid, SIGUSR2);
#endif
      pthread_join (keepalive_tid, NULL);
    }

  cleanup_all_clients (0);
#ifdef WITH_GNUTLS
  start_stop_tls (1);
#endif
  deinit_commands ();
  pthread_cond_destroy (&quit_cond);
  pthread_mutex_destroy (&quit_mutex);
  return segv ? EXIT_FAILURE : EXIT_SUCCESS;;
}

static void
startup_failure ()
{
  log_write (_
	     ("Failed to add a file to the cache. Use --force to force startup. Exiting."));
  cache_clear (NULL);
}

/* This is called from cache.c:clear_once().  See
 * command.c:clearcache_command() for details about lock checking.
 */
static gpg_error_t
free_cache_data (file_cache_t * cache)
{
  gpg_error_t rc = GPG_ERR_NO_DATA;
  int i, t;
  struct client_thread_s *found = NULL;
  int self = 0;

  if (!cache->data)
    return 0;

  cache_lock ();
  MUTEX_LOCK (&cn_mutex);
  pthread_cleanup_push (cleanup_mutex_cb, &cn_mutex);
  t = slist_length (cn_thread_list);

  for (i = 0; i < t; i++)
    {
      struct client_thread_s *thd = slist_nth_data (cn_thread_list, i);

      if (!thd->cl)
	continue;

      if (!memcmp (thd->cl->md5file, cache->filename,
		   sizeof (cache->filename)))
	{
	  if (pthread_equal (pthread_self (), thd->tid))
	    {
	      found = thd;
	      self = 1;
	      continue;
	    }

	  /* Continue trying to find a client who has the same file open and
	   * also has a lock. */
	  rc = cache_lock_mutex (thd->cl->ctx, thd->cl->md5file, -1, 0, -1);
	  if (!rc)
	    {
	      self = 0;
	      found = thd;
	      break;
	    }
	}
    }

  if (self && (!rc || rc == GPG_ERR_NO_DATA))
    rc = cache_lock_mutex (found->cl->ctx, found->cl->md5file, -1, 0, -1);

  if (exiting || !rc || rc == GPG_ERR_NO_DATA)
    {
      free_cache_data_once (cache->data);
      cache->data = NULL;
      cache->defer_clear = 0;
      cache->timeout = -1;

      if (found)
	cache_unlock_mutex (found->cl->md5file, 0);

      rc = 0;
    }

  if (rc)
    cache->defer_clear = 1;

  pthread_cleanup_pop (1);
  cache_unlock ();
  return rc;
}

static int
convert_v2_datafile (const char *filename, const char *cipher,
		     const char *keyfile, const char *keygrip,
		     const char *sign_keygrip, int nopass,
		     const char *outfile, const char *keyparam,
		     uint64_t iterations)
{
  gpg_error_t rc;
  void *data = NULL;
  size_t datalen;
  struct crypto_s *crypto = NULL;
  uint16_t ver;
  int algo;
  void *key = NULL;
  size_t keylen = 0;

  if (outfile[0] == '-' && outfile[1] == 0)
    outfile = NULL;

  log_write (_("Converting version 2 data file \"%s\" ..."), filename);
  if (access (filename, R_OK) == -1)
    {
      log_write ("%s: %s", filename,
		 pwmd_strerror (gpg_error_from_errno (errno)));
      return 0;
    }

  if (keyfile)
    {
      log_write (_("Using passphrase file \"%s\" for decryption ..."),
		 keyfile);
      if (access (keyfile, R_OK) == -1)
	{
	  log_write ("%s: %s", keyfile,
		     pwmd_strerror (gpg_error_from_errno (errno)));
	  return 0;
	}
    }

  rc = read_v2_datafile (filename, keyfile, &data, &datalen, &ver, &algo);
  if (rc)
    {
      log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
      return 0;
    }

  if (cipher)
    {
      algo = cipher_string_to_gcrypt (cipher);
      if (algo == -1)
	{
	  rc = GPG_ERR_CIPHER_ALGO;
	  goto fail;
	}
    }

  if (ver < 0x212)
    {
      xmlDocPtr doc;

      rc = parse_doc (data, datalen, &doc);
      if (rc)
	goto fail;

      rc = convert_pre_212_elements (doc);
      gcry_free (data);
      data = NULL;
      if (!rc)
	{
	  xmlDocDumpFormatMemory (doc, (xmlChar **) & data, (int *) &datalen,
				  0);
	  if (!data)
	    rc = GPG_ERR_ENOMEM;
	}

      xmlFreeDoc (doc);
      if (rc)
	goto fail;
    }

  rc = init_client_crypto (&crypto);
  if (!rc)
    {
      memcpy (&crypto->save.hdr, &crypto->hdr, sizeof (file_header_t));
      crypto->save.hdr.flags = set_cipher_flag (crypto->save.hdr.flags, algo);
      crypto->save.hdr.s2k_count = iterations;

      if (!use_agent)
	{
	  rc = export_common (NULL, 0, crypto, data, datalen, outfile, keyfile,
			      &key, &keylen, 0, 0, nopass);
	}
#ifdef WITH_AGENT
      else
	{
          rc = agent_set_pinentry_options (crypto->agent);
          if (!rc)
            rc = agent_export_common (crypto, keygrip, sign_keygrip, nopass,
	                              data, datalen, outfile, keyparam,
				      no_passphrase_file ? NULL : keyfile);
        }
#endif
      if (!rc)
	log_write (_("Output written to \"%s\"."),
                   outfile ? outfile : "<stdout>");
    }

fail:
  if (ver < 0x212)
    xmlFree (data);
  else
    gcry_free (data);

  gcry_free (key);
  cleanup_crypto (&crypto);

  if (rc)
    log_write ("ERR %i: %s", rc, pwmd_strerror (rc));
  return rc ? 0 : 1;
}

static void
usage (const char *pn, int status)
{
  FILE *fp = status == EXIT_FAILURE ? stderr : stdout;

  fprintf (fp, _("Usage: %s [OPTIONS] [file1] [...]\n"
		 "  -f, --rcfile=filename        load the specfied configuration file\n"
		 "                               (~/.pwmd/config)\n"
		 "  --homedir                    alternate pwmd home directory (~/.pwmd)\n"
                 "  --kill                       terminate an existing instance of pwmd\n"
#ifdef WITH_AGENT
		 "  --use-agent[=integer]        enable/disable use of gpg-agent\n"
#endif
		 "  -n, --no-fork                run as a foreground process\n"
		 "  -D, --disable-dump           disable the LIST, XPATH and DUMP commands\n"
		 "  --force                      ignore file errors during startup\n"
		 "  --debug-level=keywords       log protocol output (see manual for details)\n"
		 "  -o, --outfile=filename       output file when importing or converting\n"
		 "  -C, --convert=filename       convert a version 2 data file to version 3\n"
		 "  -I, --import=filename        import a pwmd DTD formatted XML file)\n"
		 "  -k, --passphrase-file=file   for use when importing or converting\n"
		 "  --no-passphrase-file         prompt instead of using --passphrase-file when\n"
		 "                               converting\n"
		 "  --no-passphrase              when importing or converting\n"
		 "  --keygrip=hex                public key to use when encrypting\n"
		 "  --sign-keygrip=hex           private key to use when signing\n"
		 "  --keyparam=s-exp             custom key parameters to use (RSA-2048)\n"
		 "  --cipher=string              encryption cipher (aes256)\n"
		 "  --cipher-iterations=N        cipher iteration count (alias for --s2k-count)\n"
		 "  --s2k-count=N                hash iteration count (>65536, calibrated)\n"
		 "  --help                       this help text\n"
		 "  --version                    show version and compile time features\n"),
	   pn);
  exit (status);
}

static void
unlink_stale_socket (const char *sock, const char *pidfile)
{
  log_write (_ ("removing stale socket %s"), sock);
  unlink (sock);
  unlink (pidfile);
}

static int
test_pidfile (const char *path, const char *sock, char *buf, size_t buflen,
	      char **pidfile, int create, mode_t mode, int terminate)
{
  pid_t pid;
  int fd;
  size_t len;

  if (!create)
    {
      snprintf (buf, buflen, "%s/%s.pid", homedir, sock);
      *pidfile = str_dup (buf);
      fd = open (buf, O_RDONLY);
    }
  else
    fd = open (*pidfile, O_CREAT|O_WRONLY|O_TRUNC, mode);

  if (fd == -1)
    {
      if (!create && errno != ENOENT)
	{
	  log_write ("%s: %s", buf, pwmd_strerror (errno));
	  free (*pidfile);
	  *pidfile = NULL;
	  return -1;
	}
      else if (!create && !terminate)
	return 0;

      log_write ("%s: %s", *pidfile, strerror (errno));
      return -1;
    }

  if (create)
    {
      snprintf (buf, buflen, "%i", getpid ());
      ssize_t ret = write (fd, buf, strlen (buf));
      if (ret == -1)
        log_write ("%s: %s", *pidfile, strerror (errno));

      close (fd);
      return 0;
    }

  len = read (fd, buf, buflen);
  close (fd);
  if (len == 0)
    {
      unlink_stale_socket (path, *pidfile);
      return 0;
    }

  if (sscanf (buf, "%5i", &pid) != 1 || pid == 0)
    {
      if (!terminate)
        {
          unlink_stale_socket (path, *pidfile);
          return 0;
        }
    }

  if (kill (pid, 0) == -1)
    {
      unlink_stale_socket (path, *pidfile);
      return 0;
    }

  if (terminate)
    {
      if (kill (pid, SIGTERM) == -1)
        log_write ("%s: %s", path, pwmd_strerror (errno));
    }
  else
    log_write (_ ("an instance for socket %s is already running"), path);

  xfree (*pidfile);
  *pidfile = NULL;
  return 1;
}

int
main (int argc, char *argv[])
{
  int opt;
  struct sockaddr_un addr;
  char buf[PATH_MAX];
  char *socketpath = NULL, *socketdir, *socketname = NULL;
  char *socketarg = NULL;
  char *datadir = NULL;
  char *pidfile = NULL;
  mode_t mode = 0600;
  int x;
  char *p;
  char **cache_push = NULL;
  char *import = NULL, *keygrip = NULL, *sign_keygrip = NULL;
  char *keyparam = NULL;
  int estatus = EXIT_FAILURE;
  int sockfd;
  char *outfile = NULL;
  int do_unlink = 0;
  int secure = 0;
  int show_version = 0;
  int force = 0;
  int no_passphrase = 0;
  gpg_error_t rc;
  char *convertfile = NULL;
  char *cipher = NULL;
  char *keyfile = NULL;
  uint64_t s2k_count = 0;
  int exists;
  char *debug_level_opt = NULL;
  int optindex;
  int terminate = 0;
  /* Must maintain the same order as longopts[] */
  enum
  {
    OPT_VERSION, OPT_HELP,
#ifdef WITH_AGENT
    OPT_AGENT,
#endif
    OPT_DEBUG_LEVEL, OPT_HOMEDIR, OPT_NO_FORK, OPT_DISABLE_DUMP,
    OPT_IGNORE, OPT_FORCE, OPT_RCFILE, OPT_CONVERT,
    OPT_PASSPHRASE_FILE, OPT_IMPORT, OPT_OUTFILE,
    OPT_NO_PASSPHRASE_FILE, OPT_KEYGRIP, OPT_SIGN_KEYGRIP,
    OPT_KEYPARAM, OPT_CIPHER, OPT_ITERATIONS, OPT_S2K_COUNT,
    OPT_NO_PASSPHRASE, OPT_KILL
  };
  const char *optstring = "nf:C:k:I:o:";
  const struct option longopts[] = {
    {"version", no_argument, 0, 0},
    {"help", no_argument, 0, 0},
#ifdef WITH_AGENT
    {"use-agent", optional_argument, 0, 0},
#endif
    {"debug-level", required_argument, 0, 0},
    {"homedir", required_argument, 0, 0},
    {"no-fork", no_argument, 0, 'n'},
    {"disable_dump", no_argument, 0, 0},
    {"ignore", no_argument, 0, 0},
    {"force", no_argument, 0, 0},
    {"rcfile", required_argument, 0, 'f'},
    {"convert", required_argument, 0, 'C'},
    {"passphrase-file", required_argument, 0, 'k'},
    {"import", required_argument, 0, 'I'},
    {"outfile", required_argument, 0, 'o'},
    {"no-passphrase-file", no_argument, 0, 0},
    {"keygrip", required_argument, 0, 0},
    {"sign-keygrip", required_argument, 0, 0},
    {"keyparam", required_argument, 0, 0},
    {"cipher", required_argument, 0, 0},
    {"cipher-iterations", required_argument, 0, 0},
    {"s2k-count", required_argument, 0, 0},
    {"no-passphrase", no_argument, 0, 0},
    {"kill", no_argument, 0, 0},
    {0, 0, 0, 0}
  };

  log_fd = -1;

#ifndef DEBUG
#ifdef HAVE_SETRLIMIT
  struct rlimit rl;

  rl.rlim_cur = rl.rlim_max = 0;

  if (setrlimit (RLIMIT_CORE, &rl) != 0)
    err (EXIT_FAILURE, "setrlimit()");
#endif

#ifdef HAVE_PR_SET_DUMPABLE
  prctl (PR_SET_DUMPABLE, 0);
#endif
#endif

#ifdef ENABLE_NLS
  setlocale (LC_ALL, "");
  bindtextdomain ("pwmd", LOCALEDIR);
  textdomain ("pwmd");
#endif

#ifndef MEM_DEBUG
  xmem_init ();
#endif
  gpg_err_init ();

  if (setup_crypto ())
    exit (EXIT_FAILURE);

#ifdef WITH_GNUTLS
  gnutls_global_init ();
  gnutls_global_set_log_level (1);
  gnutls_global_set_log_function (tls_log);
  gnutls_global_set_audit_log_function (tls_audit_log);
  tls_fd = -1;
  tls6_fd = -1;
#endif
  xmlMemSetup (xfree, xmalloc, xrealloc, str_dup);
  xmlInitMemory ();
  xmlInitGlobals ();
  xmlInitParser ();
  xmlXPathInit ();
  cmdline = 1;
  use_agent = -1;

  while ((opt = getopt_long (argc, argv, optstring, longopts, &optindex))
	 != -1)
    {
      switch (opt)
	{
	case 'I':
	  import = optarg;
	  break;
	case 'C':
	  convertfile = optarg;
	  break;
	case 'k':
	  keyfile = optarg;
	  break;
	case 'o':
	  outfile = optarg;
	  break;
	case 'D':
	  secure = 1;
	  break;
	case 'n':
	  nofork = 1;
	  break;
	case 'f':
	  rcfile = str_dup (optarg);
	  break;
	default:
	  usage (argv[0], EXIT_FAILURE);
	  break;
	case 0:
	  switch (optindex)
	    {
	    case OPT_VERSION:
	      show_version = 1;
	      break;
	    case OPT_HELP:
	      usage (argv[0], 0);
	      break;
#ifdef WITH_AGENT
	    case OPT_AGENT:
	      use_agent = optarg && *optarg ? atoi (optarg) : 1;
              if (use_agent < 0 || use_agent > 1)
                usage (argv[0], 1);
	      break;
#endif
	    case OPT_DEBUG_LEVEL:
	      debug_level_opt = optarg;
	      break;
	    case OPT_HOMEDIR:
	      homedir = str_dup (optarg);
	      break;
	    case OPT_NO_FORK:
	      nofork = 1;
	      break;
	    case OPT_DISABLE_DUMP:
	      secure = 1;
	      break;
	    case OPT_IGNORE:
	    case OPT_FORCE:
	      force = 1;
	      break;
	    case OPT_RCFILE:
	      rcfile = str_dup (optarg);
	      break;
	    case OPT_CONVERT:
	      convertfile = optarg;
	      break;
	    case OPT_PASSPHRASE_FILE:
	      keyfile = optarg;
	      break;
	    case OPT_IMPORT:
	      import = optarg;
	      break;
	    case OPT_OUTFILE:
	      outfile = optarg;
	      break;
	    case OPT_NO_PASSPHRASE_FILE:
	      no_passphrase_file = 1;
	      break;
	    case OPT_KEYGRIP:
	      keygrip = optarg;
	      break;
	    case OPT_SIGN_KEYGRIP:
	      sign_keygrip = optarg;
	      break;
	    case OPT_KEYPARAM:
	      keyparam = optarg;
	      break;
	    case OPT_CIPHER:
	      cipher = optarg;
	      break;
	    case OPT_ITERATIONS:
	    case OPT_S2K_COUNT:
	      s2k_count = strtoull (optarg, NULL, 10);
	      break;
	    case OPT_NO_PASSPHRASE:
	      no_passphrase = 1;
	      break;
            case OPT_KILL:
              terminate = 1;
              break;
	    default:
	      usage (argv[0], 1);
	    }
	}
    }

  if (show_version)
    {
      printf (_("%s\n\n"
		"Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015\n"
		"%s\n"
		"Released under the terms of the GPL v2. Use at your own risk.\n\n"
		"Compile time features:\n%s"), PACKAGE_STRING,
	      PACKAGE_BUGREPORT,
#ifdef PWMD_HOMEDIR
	      "+PWMD_HOMEDIR=" PWMD_HOMEDIR "\n"
#endif
#ifdef WITH_AGENT
	      "+WITH_AGENT\n"
#else
	      "-WITH_AGENT\n"
#endif
#ifdef WITH_QUALITY
	      "+WITH_QUALITY\n"
#else
	      "-WITH_QUALITY\n"
#endif
#ifdef WITH_GNUTLS
	      "+WITH_GNUTLS\n"
#else
	      "-WITH_GNUTLS\n"
#endif
#ifdef WITH_LIBACL
	      "+WITH_LIBACL\n"
#else
	      "-WITH_LIBACL\n"
#endif
#ifdef DEBUG
	      "+DEBUG\n"
#else
	      "-DEBUG\n"
#endif
#ifdef MEM_DEBUG
	      "+MEM_DEBUG\n"
#else
	      "-MEM_DEBUG\n"
#endif
#ifdef MUTEX_DEBUG
	      "+MUTEX_DEBUG\n"
#else
	      "-MUTEX_DEBUG\n"
#endif
	);
      exit (EXIT_SUCCESS);
    }

  if (!homedir)
#ifdef PWMD_HOMEDIR
    homedir = str_dup(PWMD_HOMEDIR);
#else
    homedir = str_asprintf ("%s/.pwmd", get_home_dir());
#endif

  if (mkdir (homedir, 0700) == -1 && errno != EEXIST)
    err (EXIT_FAILURE, "%s", homedir);

  snprintf (buf, sizeof (buf), "%s/data", homedir);
  if (mkdir (buf, 0700) == -1 && errno != EEXIST)
    err (EXIT_FAILURE, "%s", buf);

  datadir = str_dup (buf);
  pthread_mutexattr_t attr;
  pthread_mutexattr_init (&attr);
  pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init (&rcfile_mutex, &attr);
  pthread_cond_init (&rcfile_cond, NULL);
  pthread_mutex_init (&cn_mutex, &attr);
  pthread_mutexattr_destroy (&attr);
  pthread_key_create (&last_error_key, free_key);
#ifndef HAVE_PTHREAD_CANCEL
  pthread_key_create (&signal_thread_key, free_key);
#endif

  if (!rcfile)
    rcfile = str_asprintf ("%s/config", homedir);

  global_config = config_parse (rcfile, 0);
  if (!global_config)
    exit (EXIT_FAILURE);

#ifdef WITH_AGENT
  if (use_agent == -1)
    use_agent = config_get_boolean ("global", "use_agent");
#endif

  setup_logging ();

  if (debug_level_opt)
    debug_level = str_split (debug_level_opt, ",", 0);

  x = config_get_int_param (global_config, "global", "priority", &exists);
  if (exists && x != atoi(INVALID_PRIORITY))
    {
      errno = 0;
      if (setpriority (PRIO_PROCESS, 0, x) == -1)
	{
	  log_write ("setpriority(): %s",
		     pwmd_strerror (gpg_error_from_errno (errno)));
	  goto do_exit;
	}
    }
#ifdef HAVE_MLOCKALL
  if (disable_mlock == 0 && mlockall (MCL_CURRENT | MCL_FUTURE) == -1)
    {
      log_write ("mlockall(): %s",
		 pwmd_strerror (gpg_error_from_errno (errno)));
      goto do_exit;
    }
#endif

  rc = cache_init (free_cache_data);
  if (rc)
    {
      log_write ("pwmd: ERR %i: %s", rc,
		 gpg_err_code (rc) == GPG_ERR_UNKNOWN_VERSION
		 ? _("incompatible gpg-agent version: 2.1.0 or later required")
		 : pwmd_strerror (rc));
      goto do_exit;
    }

  if (s2k_count == 0)
    s2k_count = config_get_ulonglong (NULL, "s2k_count");

  if (convertfile)
    {
      if (!outfile || !*outfile || argc != optind)
	usage (argv[0], EXIT_FAILURE);

      estatus = convert_v2_datafile (convertfile, cipher, keyfile, keygrip,
				     sign_keygrip, no_passphrase, outfile,
				     keyparam, s2k_count);
      config_free (global_config);
      xfree (rcfile);
      exit (!estatus);
    }

  if (import)
    {
      if (!outfile || !*outfile || argc != optind)
	usage (argv[0], EXIT_FAILURE);

      if (outfile && outfile[0] == '-' && outfile[1] == 0)
	outfile = NULL;

      estatus = xml_import (import, outfile, keygrip, sign_keygrip, keyfile,
			    no_passphrase, cipher, keyparam, s2k_count);
      config_free (global_config);
      xfree (rcfile);
      exit (!estatus);
    }

  p = config_get_string ("global", "socket_path");
  if (!p)
    p = str_asprintf ("%s/socket", homedir);

  socketarg = expand_homedir (p);
  xfree (p);

  if (!secure)
    disable_list_and_dump = config_get_boolean ("global",
						"disable_list_and_dump");
  else
    disable_list_and_dump = secure;

  cache_push = config_get_list ("global", "cache_push");

  while (optind < argc)
    {
      if (strv_printf (&cache_push, "%s", argv[optind++]) == 0)
	errx (EXIT_FAILURE, "%s", pwmd_strerror (GPG_ERR_ENOMEM));
    }

  if (strchr (socketarg, '/') == NULL)
    {
      socketdir = getcwd (buf, sizeof (buf));
      socketname = str_dup (socketarg);
      socketpath = str_asprintf ("%s/%s", socketdir, socketname);
    }
  else
    {
      socketname = str_dup (strrchr (socketarg, '/'));
      socketname++;
      socketarg[strlen (socketarg) - strlen (socketname) - 1] = 0;
      socketdir = str_dup (socketarg);
      socketpath = str_asprintf ("%s/%s", socketdir, socketname);
    }

  if (chdir (datadir))
    {
      log_write ("%s: %s", datadir,
		 pwmd_strerror (gpg_error_from_errno (errno)));
      unlink (socketpath);
      goto do_exit;
    }

  x = test_pidfile (socketpath, socketname, buf, sizeof(buf), &pidfile, 0,
                    mode, terminate);
  if (!terminate && x)
    goto do_exit;
  else if (terminate)
    {
      estatus = x != 1 ? EXIT_FAILURE : EXIT_SUCCESS;
      goto do_exit;
    }

  /*
   * bind() doesn't like the full pathname of the socket or any non alphanum
   * characters so change to the directory where the socket is wanted then
   * create it then change to datadir.
   */
  if (chdir (socketdir))
    {
      log_write ("%s: %s", socketdir,
		 pwmd_strerror (gpg_error_from_errno (errno)));
      goto do_exit;
    }

  xfree (socketdir);

  if ((sockfd = socket (PF_UNIX, SOCK_STREAM, 0)) == -1)
    {
      log_write ("socket(): %s", pwmd_strerror (gpg_error_from_errno (errno)));
      goto do_exit;
    }

  addr.sun_family = AF_UNIX;
  snprintf (addr.sun_path, sizeof (addr.sun_path), "%s", socketname);
  do_unlink = 1;
  if (bind (sockfd, (struct sockaddr *) &addr, sizeof (struct sockaddr)) ==
      -1)
    {
      log_write ("bind(): %s", pwmd_strerror (gpg_error_from_errno (errno)));

      if (errno == EADDRINUSE)
	{
	  do_unlink = 0;
	  log_write (_("Either there is another pwmd running or '%s' is a \n"
		       "stale socket. Please remove it manually."), socketpath);
	}

      goto do_exit;
    }

  {
    char *t = config_get_string ("global", "socket_perms");
    mode_t mask;

    if (t)
      {
	mode = strtol (t, NULL, 8);
	mask = umask (0);
	xfree (t);

	if (chmod (socketname, mode) == -1)
	  {
	    log_write ("%s: %s", socketname,
		       pwmd_strerror (gpg_error_from_errno (errno)));
	    close (sockfd);
	    umask (mask);
	    goto do_exit;
	  }

	umask (mask);
      }
  }

  xfree (--socketname);

  if (chdir (datadir))
    {
      log_write ("%s: %s", datadir,
		 pwmd_strerror (gpg_error_from_errno (errno)));
      close (sockfd);
      goto do_exit;
    }

  xfree (datadir);

  /*
   * Set the cache entry for a file. Prompts for the password.
   */
  if (cache_push)
    {
      struct crypto_s *crypto = NULL;
      gpg_error_t rc = init_client_crypto (&crypto);

      if (rc)
	{
	  estatus = EXIT_FAILURE;
	  goto do_exit;
	}

#ifdef WITH_AGENT
      if (use_agent)
	{
	  rc = agent_set_pinentry_options (crypto->agent);
	  if (rc)
	    {
	      estatus = EXIT_FAILURE;
	      goto do_exit;
	    }
	}
#endif

      for (opt = 0; cache_push[opt]; opt++)
	{
	  if (!do_cache_push (cache_push[opt], crypto) && !force)
	    {
	      strv_free (cache_push);
	      startup_failure ();
	      estatus = EXIT_FAILURE;
	      cleanup_crypto (&crypto);
	      goto do_exit;
	    }

	  cleanup_crypto_stage1 (crypto);
	}

#ifdef WITH_AGENT
      if (use_agent)
	(void) kill_scd (crypto->agent);
#endif

      cleanup_crypto (&crypto);
      strv_free (cache_push);
      log_write (!nofork ? _("Done. Daemonizing...") :
		 _("Done. Waiting for connections..."));
    }

  config_clear_keys ();

  if (listen (sockfd, 0) == -1)
    {
      log_write ("listen(): %s", pwmd_strerror (gpg_error_from_errno (errno)));
      goto do_exit;
    }

  if (!nofork)
    {
      switch (fork ())
	{
	case -1:
	  log_write ("fork(): %s",
		     pwmd_strerror (gpg_error_from_errno (errno)));
	  goto do_exit;
	case 0:
	  close (0);
	  close (1);
	  close (2);
	  setsid ();
	  break;
	default:
	  _exit (EXIT_SUCCESS);
	}
    }

  (void)test_pidfile (socketpath, socketname, buf, sizeof(buf), &pidfile, 1,
		      mode, 0);
  cmdline = 0;
  pthread_key_create (&thread_name_key, free_key);
  pthread_setspecific (thread_name_key, str_dup ("main"));
  estatus = server_loop (sockfd, &socketpath);

do_exit:
  if (socketpath && do_unlink)
    {
      unlink (socketpath);
      xfree (socketpath);
    }

  xfree (socketarg);
#ifdef WITH_GNUTLS
  gnutls_global_deinit ();
#endif
  if (rcfile_tid)
    {
#ifdef HAVE_PTHREAD_CANCEL
      pthread_cancel (rcfile_tid);
#else
      pthread_kill (rcfile_tid, SIGUSR2);
      pthread_cond_signal (&rcfile_cond);
#endif
      pthread_join (rcfile_tid, NULL);
    }

  pthread_cond_destroy (&rcfile_cond);
  pthread_mutex_destroy (&rcfile_mutex);
  pthread_key_delete (last_error_key);
#ifndef HAVE_PTHREAD_CANCEL
  pthread_key_delete (signal_thread_key);
#endif

  if (global_config)
    config_free (global_config);

  free_invoking_users (invoking_users);
  xfree (rcfile);
  xfree (home_directory);
  xfree (homedir);
  xmlCleanupParser ();
  xmlCleanupGlobals ();

  if (pidfile)
    unlink (pidfile);
  xfree (pidfile);

  if (estatus == EXIT_SUCCESS && !terminate)
    log_write (_("pwmd exiting normally"));

  pthread_key_delete (thread_name_key);
  closelog ();
#if defined(DEBUG) && !defined(MEM_DEBUG)
  xdump ();
#endif

  if (log_fd != -1)
    close (log_fd);

  exit (estatus);
}

gpg_error_t lock_flock (assuan_context_t ctx, const char *filename,
                        int type, int *fd)
{
  gpg_error_t rc = 0;

#ifdef HAVE_FLOCK
  *fd = open (filename, O_RDONLY);
  if (*fd == -1)
    return gpg_error_from_syserror ();

  TRY_FLOCK (ctx, *fd, type, rc);
  if (rc)
    {
      close (*fd);
      *fd = -1;
    }
#endif

  return rc;
}

void unlock_flock (int *fd)
{
#ifdef HAVE_FLOCK
  if (*fd != -1)
    close (*fd);

  *fd = -1;
#endif
}
