/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>

#include "pwmd-error.h"
#include <gcrypt.h>
#include "util-misc.h"
#include "util-string.h"
#include "mutex.h"
#include "cache.h"
#include "mem.h"

extern void log_write (const char *fmt, ...);

int
valid_filename (const char *filename)
{
  const char *p;

  if (!filename || !*filename)
    return 0;

  if (filename[0] == '-' && filename[1] == 0)
    return 0;

  for (p = filename; *p; p++)
    {
      if (*p == '/' || isspace (*p))
	return 0;
    }

  return 1;
}

/* If 'user' is null then lookup entry for 'uid'. Caller must free 'buf'. */
struct passwd *
get_pwd_struct (const char *user, uid_t uid, struct passwd *pw, char **buf,
                gpg_error_t *rc)
{
  struct passwd *result;
#ifdef HAVE_GETPWNAM_R
  size_t len;
  int err = 0;

  len = sysconf (_SC_GETPW_R_SIZE_MAX);
  if (len == -1)
    len = 16384;

  *buf = xmalloc (len);
  if (!*buf)
    return NULL;

  if (user)
    {
      err = getpwnam_r (user, pw, *buf, len, &result);
      if (!err)
        return result;
    }
  else
    {
      err = getpwuid_r (uid, pw, *buf, len, &result);
      if (!err)
        return result;
    }

  errno = err;
  *rc = gpg_error_from_errno (err);
  return NULL;
#else
  if (buf)
    *buf = NULL;

  errno = 0;
  if (user)
    result = getpwnam (user);
  else
    result = getpwuid (uid);

  if (!result)
    *rc = gpg_error_from_syserror ();

  return result;
#endif
}

static char *
get_pwd_entry(uid_t uid, int which)
{
  gpg_error_t rc;
  char *buf;
  struct passwd pw;
  struct passwd *result = get_pwd_struct (NULL, uid, &pw, &buf, &rc);
  char *value = NULL;

  if (!result)
    return NULL;

  if (result)
    {
      if (!which)
        value = str_dup (result->pw_dir);
      else
        value = str_dup (result->pw_name);
    }

  xfree (buf);
  return value;
}

char *
get_username (uid_t uid)
{
  return get_pwd_entry (uid, 1);
}

char *
get_home_dir ()
{
  if (home_directory)
    return home_directory;

  home_directory = get_pwd_entry (getuid(), 0);
  return home_directory;
}

char *
expand_homedir (char *str)
{
  char *p = str;

  if (*p++ == '~')
    {
      char *dir = get_pwd_entry(getuid(), 0);
      char *s;

      if (!dir)
        return NULL;

      s = str_asprintf ("%s%s", dir, p);
      xfree (dir);
      return s;
    }

  return str_dup (str);
}

/* The 'more' parameter lets the calling command process remaining
 * non-option arguments. */
gpg_error_t
parse_options (char **line, struct argv_s * args[], void *data, int more)
{
  char *p = *line;
  gpg_error_t rc = 0;

  for (; p && *p; p++)
    {
      while (*p == ' ')
	p++;

      if (!*p)
	break;

      if (*p == '-' && *(p + 1) == '-')
	{
	  p += 2;
	  char opt[255] = { 0 }, value[255] =
	  {
	  0};
	  char *tp;
	  unsigned len;

	  if (!*p || *p == ' ')
	    {
	      if (*p)
		p++;

	      break;
	    }

	  for (tp = opt, len = 0; *p; p++, len++)
	    {
	      if (len + 1 == 255)
		return GPG_ERR_LINE_TOO_LONG;

	      if (*p == ' ' || *p == '=')
		{
		  *tp = 0;

		  if (*p == '=')
		    {
                      int inquote = 0;

		      p++;

		      for (tp = value, len = 0; *p; p++, len++)
			{
			  if (len + 1 == 255)
			    return GPG_ERR_LINE_TOO_LONG;

			  if (*p == '\"')
			    {
			      inquote = !inquote;
			      continue;
			    }
			  else if (*p == ' ' && !inquote)
			    break;

			  *tp++ = *p;
			}

		      *tp = 0;
		    }

		  break;
		}

	      *tp++ = *p;
	    }

	  *tp = 0;
	  int match = 0;

	  for (int i = 0; args[i]; i++)
	    {
	      if (!strcmp (args[i]->opt, opt))
		{
		  log_write1 ("param: name='%s' value='%s'", opt, value);
		  if (args[i]->type == OPTION_TYPE_NOARG && *value)
		    return GPG_ERR_SYNTAX;
		  else if (args[i]->type == OPTION_TYPE_ARG && !*value)
		    return GPG_ERR_SYNTAX;

		  rc = args[i]->func (data, value);

		  if (rc)
		    return rc;

		  match = 1;
		  break;
		}
	    }

	  if (!match)
	    return GPG_ERR_UNKNOWN_OPTION;

	  if (!*p)
	    break;

	  continue;
	}

      if (*(p+1) && !more)
        return GPG_ERR_SYNTAX;

      break;
    }

  *line = p;
  return rc;
}

char *
bin2hex (const unsigned char *data, size_t len)
{
  size_t size = len * 2 + 1;
  char buf[size];
  size_t n;

  buf[0] = 0;

  for (n = 0; n < len; n++)
    {
      char c[3];

      sprintf (c, "%02X", data[n]);
      strcat (buf, c);
    }

  return str_dup (buf);
}

char *
plus_escape (const char *fmt, ...)
{
  va_list ap;
  char *str;

  va_start (ap, fmt);

  if (str_vasprintf (&str, fmt, ap) > 0)
    {
      char *p;

      for (p = str; *p; p++)
	{
	  if (*p == ' ')
	    *p = '+';
	}

      va_end (ap);
      return str;
    }

  va_end (ap);
  return NULL;
}

static char *
strip_texi (const char *str)
{
  const char *p;
  char *s = str_dup (str);
  int c = 0;

  if (!s)
    return NULL;

  for (p = str; *p; p++)
    {
      if (*p == '@' && *(p + 1) != '@')
	{
	  if (!strncasecmp (p + 1, "table", 5)
	      || !strncasecmp (p + 1, "end ", 4))
	    {
	      while (*p++ != '\n');
	      p--;
	      continue;
	    }
	  else if (!strncasecmp (p + 1, "example", 7)
		   || !strncasecmp (p + 1, "end ", 4))
	    {
	      while (*p++ != '\n');
	      p--;
	      continue;
	    }
	  else if (!strncasecmp (p + 1, "sp ", 3))
	    {
	      p += 3;

	      while (*p && isdigit (*p++))
		p++;
	      continue;
	    }
	  else if (!strncasecmp (p + 1, "item", 4))
	    {
	      p += 5;
	      while (*p && *p != '\n')
		s[c++] = *p++;

              s[c++] = ' ';
	      continue;
	    }
	  else if (!strncasecmp (p + 1, "pxref", 5))
	    {
	      p += 6;
	      strcat (s, "see ");
	      c = strlen (s);
	      goto close;
	    }
	  else if (!strncasecmp (p + 1, "xref", 4))
	    {
	      p += 5;
	      strcat (s, "See ");
	      c = strlen (s);
	      goto close;
	    }

	  while (*p && *p != '{')
	    {
	      if (*++p == '*')
		{
		  p++;
		  goto append;
		}
	    }

	close:
	  if (*p)
	    {
	      p++;
	      s[c++] = '`';
	      while (*p && *p != '}')
		s[c++] = *p++;

	      if (*p)
		p++;

	      s[c++] = '\'';
	    }
	}
      else if (*p == '@' && *(p + 1) == '@')
	p++;

    append:
      if (*p)
	s[c++] = *p;

      s[c] = 0;
    }

  s[c] = 0;
  return s;
}


char *
strip_texi_and_wrap (const char *str)
{
  char *tmp = strip_texi (str);
  char *help = xmalloc (strlen (tmp) * 2 + 1);
  char *p, *ph;
  int i;

  for (p = tmp, ph = help, i = 0; *p; p++, i++)
    {
      if (i == 78 || *p == '\n')
	{
	  if (!isspace (*p))
	    {
	      char *t = ph;
	      int n = 0;

	      while (!(*--t == ' '))
		n++;

	      *t++ = '\n';
	      i = -1;
	      p -= n;
	      while (n--)
		{
		  *t++ = *p++;
		  i++;
		}
	    }
	  else
	    {
	      *ph++ = '\n';
	      i = -1;
	      while (*p != '\n' && *p == ' ')
		p++;
	    }
	}

      *ph++ = *p;
    }

  *ph = 0;
  xfree (tmp);
  return help;
}

void
free_key (void *data)
{
  xfree (data);
}

gpg_error_t
create_thread (void *(*cb) (void *), void *data,
	       pthread_t * tid, int detached)
{
  pthread_attr_t attr;
  int n;

  pthread_attr_init (&attr);

  if (detached)
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

  n = pthread_create (tid, &attr, cb, data);
  pthread_attr_destroy (&attr);
  return gpg_error_from_errno (n);
}

void
cleanup_mutex_cb (void *arg)
{
  pthread_mutex_t *m = (pthread_mutex_t *) arg;

  MUTEX_UNLOCK (m);
}

void
cleanup_fd_cb (void *arg)
{
  int fd = *(int *) arg;

  if (fd != -1)
    close (fd);
}

void
cleanup_unlink_cb (void *arg)
{
  char *file = arg;

  if (file && *file)
    unlink (file);
}

void
cleanup_cache_mutex (void *arg)
{
  cache_unlock ();
}

int
valid_keygrip (const unsigned char *data, size_t len)
{
  for (size_t i = 0; i < len; i++)
    {
      if (data[i])
	return 1;
    }

  return 0;
}

/* The advisory lock should be obtained before calling this function. */
gpg_error_t
get_checksum (const char *filename, unsigned char **r_crc, size_t * r_crclen)
{
  int fd;
  unsigned char *buf;
  gpg_error_t rc = 0;
  struct stat st;

  if (stat (filename, &st) == -1)
    return gpg_error_from_errno (errno);

  fd = open (filename, O_RDONLY);
  if (fd == -1)
    return gpg_error_from_errno (errno);

  pthread_cleanup_push (cleanup_fd_cb, &fd);
  buf = xmalloc (st.st_size);
  if (buf)
    {
      pthread_cleanup_push (xfree, buf);

      size_t len = read (fd, buf, st.st_size);
      if (len == st.st_size)
	{
	  if (buf)
	    {
	      unsigned char *crc;

	      len = gcry_md_get_algo_dlen (GCRY_MD_CRC32);
	      crc = xmalloc (len);
	      if (crc)
		{
		  pthread_cleanup_push (xfree, crc);
		  gcry_md_hash_buffer (GCRY_MD_CRC32, crc, buf, st.st_size);
		  *r_crc = crc;
		  *r_crclen = len;
		  pthread_cleanup_pop (0);
		}
	      else
		rc = GPG_ERR_ENOMEM;
	    }
	}
      else
	rc = GPG_ERR_TOO_SHORT;

      pthread_cleanup_pop (1);
    }
  else
    rc = GPG_ERR_ENOMEM;

  pthread_cleanup_pop (1);
  return rc;
}

wchar_t *str_to_wchar (const char *str)
{
    wchar_t *wc;
    mbstate_t ps;
    const char *p = str;
    size_t len;

    memset (&ps, 0, sizeof(mbstate_t));
    len = mbsrtowcs (NULL, &p, 0, &ps);
    if (len == -1)
      return NULL;

    wc = xcalloc (len+1, sizeof(wchar_t));
    p = str;
    memset (&ps, 0, sizeof(mbstate_t));
    len = mbsrtowcs (wc, &p, len, &ps);
    return wc;
}
